/*
 *                        ***** TA Area *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1 Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following disclaimer.
 *
 * 2 Redistributions in binary form must reproduce the
 *   above copyright notice, this list of conditions
 *   and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * 3 Neither the name of the General Energy Technologies
 *   nor the names of its contributors may be used to
 *   endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static TaAreaQos g_area_qos = {
	10000, 	// tp_cycle
	0,	// justify_frame: 
	100, 	// messages_array_size
	65536, 	// send_buffer_size
	65536,	// recv_buffer_size
	0,	// tp_phase_correct
	1000,	// max_poll_fails
	15,	// enable_mask
	32768};	// soap_buffer_size

/* ------------------------------------------------------- */
void taDefaultAreaQos(TaAreaQos *qos)
{
	*qos = g_area_qos;
}

/* ------------------------------------------------------- */
int taInitArea(TaLong aid, const TaAreaQos *qos, const char *name,
	TaArea *area)
{
	if (!area)
		return TA_ERR_ARGUMENT;
	memset(area, 0, sizeof(TaArea));
	area->aid = aid;
	if (qos)
		area->qos = *qos;
	else
		area->qos = g_area_qos;
	if (name)
		strncpy(area->name, name, TA_NAME_LENGTH);
	area->name[TA_NAME_LENGTH-1] = 0;
	taInitAreaBuffers(area);
	taInitAreaCtypes(area->qos.enable_mask, area);
	return 0;
}

/* ------------------------------------------------------- */
int taInitLibArea(const TaArea *area, TaLibStationData *lsd,
	TaArea *libarea)
{
	memset(libarea, 0, sizeof(TaArea));
	libarea->aid = area->aid;
	memcpy(libarea->name, area->name, TA_NAME_LENGTH);
	libarea->qos = area->qos;
	libarea->n_stations = 2;
	libarea->max_stations = 2;
	lsd->libstations[0] = *lsd->station;
	lsd->pstations[0] = &lsd->libstations[0];
	libarea->host_station = lsd->pstations[0];
	libarea->host_station->sid = 0;
	libarea->host_station->area = libarea;
	lsd->libstations[1] = *area->host_station;
	lsd->pstations[1] = &lsd->libstations[1];
	libarea->send_station = lsd->pstations[1];
	libarea->send_station->sid = 1;
	libarea->send_station->area = libarea;
	libarea->stations = lsd->pstations;
	libarea->n_ctypes = area->n_ctypes;
	memcpy(libarea->ctypes, area->ctypes, sizeof(int)*TA_N_CHANNELS);
	lsd->area = libarea;
	return 0;
}

/* ------------------------------------------------------- */
int taInitThreadData(TaAreaQos *qos, TaStation *station)
{
	TaArea *area, *tarea;
	if (station->ttd)
		return TA_NOT_FOUND;
	station->ttd = (TaThreadData *)
		malloc(sizeof(TaThreadData));
	memset(station->ttd, 0, sizeof(TaThreadData));
	station->ttd->station = station;

	area = station->area;
	tarea = &station->ttd->threadarea;
	tarea->aid = area->aid;
	memcpy(tarea->name, area->name, TA_NAME_LENGTH);
	if (qos)
		tarea->qos = *qos;
	else
		tarea->qos = area->qos;
	tarea->n_stations = area->n_stations;
	tarea->max_stations = area->max_stations;
	tarea->host_station = area->host_station;
	tarea->send_station = station;
	tarea->stations = area->stations;
	tarea->n_ctypes = area->n_ctypes;
	memcpy(tarea->ctypes, area->ctypes, sizeof(int)*TA_N_CHANNELS);
	station->ttd->area = tarea;
	taInitAreaBuffers(tarea);
	
	return 0;
}

/* ------------------------------------------------------- */
int taInitThreadsData(TaAreaQos *qos, TaArea *area)
{
	int i, rc = 0;
	for (i = 0; i < area->n_stations; i++)
		if (area->stations[i] != area->host_station)
			rc |= taInitThreadData(qos, area->stations[i]);
	return rc;
}

/* ------------------------------------------------------- */
void taInitAreaBuffers(TaArea *area)
{
	area->t_crt = area->t_crtu = area->t_crt0 = taGetTime();
	area->current_mid = (TaLong)(area->t_crt%USHRT_MAX);
	area->send_messages = (TaMessage *) malloc(
		area->qos.messages_array_size * sizeof(TaMessage));
	memset(area->send_messages, 0, 
		area->qos.messages_array_size * sizeof(TaMessage));
	area->recv_messages = (TaMessage *) malloc(
		area->qos.messages_array_size * sizeof(TaMessage));
	memset(area->recv_messages, 0, 
		area->qos.messages_array_size * sizeof(TaMessage));
	area->send_buffer = (char *) malloc(area->qos.send_buffer_size);
	memset(area->send_buffer, 0, area->qos.send_buffer_size);
	area->recv_buffer = (char *) malloc(area->qos.recv_buffer_size);
	memset(area->recv_buffer, 0, area->qos.recv_buffer_size);
}

/* ------------------------------------------------------- */
int taInitAreaCtypes(TaULong enable_mask, TaArea *area)
{
	int i, j;
	
	if (enable_mask & TA_ENABLE_TCP_SERVER)
		enable_mask |= TA_ENABLE_TCP_CLIENT;
	else if (enable_mask & TA_ENABLE_TCP_CLIENT)
		enable_mask |= TA_ENABLE_TCP_SERVER;
	area->qos.enable_mask = enable_mask;
	for (i = j = 0; i < TA_N_PROTOCOLS; i++)
		if (enable_mask & 1<<i) {
			area->ctypes[j] = i;
			++j;
			if (j >= TA_N_CHANNELS)
				break;
		}
	area->n_ctypes = j;
	return j;
}

/* ------------------------------------------------------- */
void taReallocStations(int max_stations, TaArea *area)
{
	TaStation **stations = NULL;
	if (!area)
		return;
	if (max_stations) {
		if (area->stations && area->max_stations >= max_stations)
			return;
		stations = (TaStation **)
			malloc(max_stations * sizeof(TaStation *));
		memset(stations, 0, max_stations * sizeof(TaStation *));
		if (area->stations && area->max_stations > 0)
			memcpy(stations, area->stations, 
				area->max_stations * sizeof(TaStation *));
	}
	if (area->stations) {
		free(area->stations);
		area->stations = NULL;
		area->max_stations = area->n_stations = 0;
	}
	area->stations = stations;
	area->max_stations = max_stations;
}

/* ------------------------------------------------------- */
TaStation *taSetStationHostByName(const char *name, TaArea *area)
{
	int i;
	if (!area || !name)
		return NULL;
	area->host_station = area->send_station = NULL;
	for (i = 0; i < area->n_stations; i++) {
		if (!area->stations[i])
			continue;
		if (!strcmp(area->stations[i]->name, name))
			area->host_station = area->stations[i];
		else if (!area->send_station)
			area->send_station = area->stations[i];
	}
	return area->host_station;
}

/* ------------------------------------------------------- */
TaStation *taSetStationHostBySid(TaShort sid, TaArea *area)
{
	int i;
	if (!area || sid < 0 || sid >= area->n_stations)
		return NULL;
	area->host_station = area->stations[sid];
	for (i = 0; i < area->n_stations; i++) {
		if (!area->stations[i])
			continue;
		if (area->stations[i] == area->host_station)
			continue;
		area->send_station = area->stations[i];
		break;
	}
	return area->host_station;
}

/* ------------------------------------------------------- */
#define TA_FIRST_CORRECT_DELAY	30000000
void taRun(TaArea *area, void *client_data)
{
	TaLongLong area_tp_cycle;
	if (!area)
		return;
	area_tp_cycle = TA_TP(area->qos.tp_cycle);
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		taDumpStationLinks(area);
	// pre-run time frame installation
	area->t_crt0 = area->t_crt = area->t_crtu = taGetTime();
	area->t_phase_last = area->t_crtu-TA_TP(area->qos.tp_phase_correct)
		+TA_FIRST_CORRECT_DELAY;
	area->n_skips = 0;
	if (area->qos.justify_frame && area->qos.tp_cycle) {
		TaLongLong tp_cdt;
		TaLong fr;
		tp_cdt = area_tp_cycle - 
			area->t_crtu % area_tp_cycle;
		if (area->qos.justify_frame > 0)
			fr = area->qos.justify_frame % TA_MAX_N_FRAMES;
		else if (area->host_station)
			fr = (area->host_station->sid+1) % TA_MAX_N_FRAMES;
		else
			fr = 0;
		tp_cdt += area_tp_cycle * fr / TA_MAX_N_FRAMES;
		taSleep(tp_cdt);
		area->t_crt = area->t_crtu = taGetTime();
		area->t_crt0 += tp_cdt;
	}
	// run itselt
	while (!area->finish_flag) {
		int j, rc;
		TaLongLong cst;
		for (j = 0; j < 5; j++) {
			area->t_crt0 += area_tp_cycle;
			if (area->t_crt0 > area->t_crt)
				break;
			++area->n_skips;
		}
		if (j == 5)
			continue;
		if (area->phase_correction) {
			area->t_crt0 += area_tp_cycle * 
				area->phase_correction/TA_MAX_N_FRAMES;
			area->phase_correction = 0;
		}
		cst = area->t_crt0 - area->t_crt;
		if (area->second_phase || area->with_second) {
			TaLongLong cst0;
			if (area->second_phase)
				cst0 = cst/TA_MAX_N_FRAMES*area->second_phase;
			else
				cst0 = cst/TA_MAX_N_FRAMES;	
			// previous frame as max
			taSleep(cst0);
			rc = taSecondUpdateData(area);
			// if changed if any delay before
			cst = area->t_crt0 - area->t_crt;
			if (ta_trace_level & TA_TRAL_RUNTIME)
				fprintf(ta_file,
				"%u.%03u second rc:%d sleep:%d+%d mcs ph:%d\n",
					TA_SEC(area->t_crtu), 
					TA_MCS(area->t_crtu)/1000, 
					rc, TA_MCS(cst0), TA_MCS(cst),
					area->second_phase);
		}
		taSleep(cst);
		rc = taUpdateData(area, client_data);
		if (ta_trace_level & TA_TRAL_RUNTIME)
			fprintf(ta_file, "%u.%03u updated rc:%d sleep:%u mcs "
			"update=%u mcs, prev=%u mcs\n",
				TA_SEC(area->t_crtu), 
				TA_MCS(area->t_crtu)/1000, rc,
				TA_MCS(cst), 
				TA_MCS(area->t_crt-area->t_crtu), 
				TA_MCS(area->tp_cdt));
		area->tp_cdt = area->t_crt - area->t_crtu;
		if (area->t_expired && area->t_crt >= area->t_expired)
			area->finish_flag = 2;
	}
}

/* ------------------------------------------------------- */
int taUpdateData(TaArea *area, void *client_data)
{
	int i, rc = 0;
	area->t_crtu = taGetTime();
	rc |=
	taPollChannels(area);
	rc |= 
	taRecvMessages(area, _False);
        for (i = 0; i < area->n_stations; i++) {
		TaStation *station;
		station = area->stations[i];
		if (!station)
			continue;
		if (!station->is_initialized) {
			station->t_crt0 = area->t_crtu;
			station->n_updates = 0;
			rc |= taInitStationData(area, station);
			station->is_initialized = _True;
		}
		if (!station->on_update || station->ttd)
			continue;
		if (station->t_crt0 > area->t_crtu+area->tp_deadzone)
			continue;
		rc |=
		taUpdateStationData(area, station, client_data);
		station->t_crt0 = area->t_crtu+TA_TP(station->qos.tp_cycle);
		++station->n_updates;
        }
	if (area->tablist)
		rc |= taUpdateAreaTables(area);
	if (area->on_update)
		rc |= (*area->on_update)(area, area->t_crtu, client_data);
	rc |= 
	taSendMessages(area, _False);
	rc |= 
	taSendPingMessages(area);
	rc |=
	taProcessPhaseMessages(area);
	area->t_crt = taGetTime();
	if (ta_trace_level && ta_file != stdout)
		fflush(ta_file);
	return rc;
}

/* ------------------------------------------------------- */
int taSecondUpdateData(TaArea *area)
{
	int rc = 0;
	area->t_crtu = taGetTime();
	if (area->second_phase)
		rc |= taSendPhaseMessages(area, area->second_phase);
	// if is_redundant
	if (area->with_second) {
		//rc |= taPollChannels(area);
		//rc |= taRecvMessages(area, _True);
		rc |= taSendMessages(area, _True);
	}
	area->t_crt = taGetTime();
	return rc;
}

/* ------------------------------------------------------- */
int taUpdateLibData(TaLibStationData *lsd)
{
	int i, rc = 0;
	TaArea *area = lsd->area;
	TaChannel *channel = &area->send_station->channels[lsd->cid];
	area->t_crtu = taGetTime();
	// clean message data first
	area->n_recv_messages = 0;
	area->n_recv_used = 0;
	area->n_send_messages = 0;
	area->n_send_used = 0;

	// instead of taRecvMessages
	rc |=
	taRecvChannelMessages (area, channel, area->send_station->sid);
	if (area->tablist)
		rc |= taUpdateAreaTables(area);
	if (area->on_update)
		rc |= (*area->on_update)(area, area->t_crtu, 
			lsd->client_data);
	// instead of taSendMessages
	for (i = 0; i < area->n_send_messages; i++) {
		TaMessage *message;
		message = &area->send_messages[i];
		if (message->n_refs-- <= 0)
			continue;
		rc |=
		taSendByChannel(area, message->buffer,
			message->buffer_length, channel);
	}

	area->t_crt = taGetTime();
	return rc;
}

/* ------------------------------------------------------- */
#ifndef _WINDOWS
void *taThreadRun(void *client_data)
{
	TaThreadData *ttd = (TaThreadData *) client_data;
	TaArea *area = ttd->area;
	if (!ttd || !area)
		return NULL;
#else
DWORD WINAPI  taThreadRun(LPVOID client_data)
{
	TaThreadData *ttd = (TaThreadData *) client_data;
	TaArea *area = ttd->area;
	if (!ttd || !area)
		return 1;
#endif
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		taDumpStationLinks(area);
	// pre-run time frame installation
	area->t_crt0 = area->t_crt = area->t_crtu = taGetTime();
	area->n_skips = 0;

	// run itselt
	while (!area->finish_flag) {
		int rc;
		TaLongLong cst;
		cst = TA_TP(area->qos.tp_cycle);
		area->t_crt0 = area->t_crt + cst;
		taSleep(cst);
		rc = taUpdateThreadData(ttd, ttd->client_data);
		if (ta_trace_level & TA_TRAL_RUNTIME)
			fprintf(ta_file, 
			"%u.%03u th#%d updated rc:%d sleep:%u mcs "
			"update=%u mcs, prev=%u mcs\n",
				TA_SEC(area->t_crtu), 
				TA_MCS(area->t_crtu)/1000, 
				ttd->station->sid, rc, TA_MCS(cst), 
				TA_MCS(area->t_crt-area->t_crtu), 
				TA_MCS(area->tp_cdt));
		area->tp_cdt = area->t_crt - area->t_crtu;
		if (area->t_expired && area->t_crt >= area->t_expired)
			area->finish_flag = 2;
	}
#ifndef _WINDOWS
	return NULL;
#else
	return 0;
#endif
}

/* ------------------------------------------------------- */
int taThreadStart(TaThreadData *ttd, void *client_data)
{
#ifndef _WINDOWS
	pthread_attr_t attr;
	if (!ttd || !ttd->area)
		return TA_ERR_ARGUMENT;
	if (pthread_attr_init(&attr) < 0)
		return TA_ERR_WRONG_DATA;
	if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) < 0)
		return TA_ERR_WRONG_DATA;
	pthread_create(&ttd->f_thread, &attr, taThreadRun, ttd);
	pthread_mutex_init(&ttd->send_mutex, NULL);
	pthread_mutex_init(&ttd->recv_mutex, NULL);
#else
	ttd->handle = (TaLong)CreateThread( 
		NULL,                   // default security attributes
		0,                      // use default stack size  
		taThreadRun,			// thread function name
		ttd,					// argument to thread function 
		0,                      // use default creation flags 
		&ttd->f_thread);		// returns the thread identifier 
	if (!ttd->handle)
		return TA_ERR_OUT_OF_MEMORY;
	if (!InitializeCriticalSectionAndSpinCount(&ttd->send_mutex, 0x00000400))
		return TA_ERR_WRONG_DATA;
	if (!InitializeCriticalSectionAndSpinCount(&ttd->recv_mutex, 0x00000800))
		return TA_ERR_WRONG_DATA;
#endif
	ttd->client_data = client_data;
	return 0;
}

/* ------------------------------------------------------- */
int taThreadsStart(TaArea *area, void *client_data)
{
	int i, rc = 0;
	for (i = 0; i < area->n_stations; i++) {
		TaStation *station = area->stations[i];
		if (station == area->host_station)
			continue;
		rc |=
		taThreadStart(station->ttd, client_data);
	}
	return rc;
}

/* ------------------------------------------------------- */
int taUpdateThreadData(TaThreadData *ttd, void *client_data)
{
	TaArea *area = ttd->area;
	TaStation *station = ttd->station;
	int rc = 0;
	area->t_crtu = taGetTime();

	// update only this station
	if (station->is_initialized && station->on_update && station->ttd) {
		if (station->t_crt0 <= area->t_crtu+area->tp_deadzone) {
			rc |=
			taUpdateStationData(area, station, client_data);
			station->t_crt0 = area->t_crtu+
				TA_TP(station->qos.tp_cycle);
			++station->n_updates;
		}
	}

	// suppose RecvMessages fills the proper arrays in area
	if (area->tablist)
		rc |= taUpdateAreaTables(area);
	if (area->on_update)
		rc |= (*area->on_update)(area, area->t_crtu, client_data);
	// clear messages from main thread
	pthread_mutex_lock(&ttd->recv_mutex);
	area->n_recv_messages = 0;
	area->n_recv_used = 0;
	pthread_mutex_unlock(&ttd->recv_mutex);

	area->t_crt = taGetTime();
	return rc;
}

/* ------------------------------------------------------- */
int taInitStationsData(TaArea *area)
{
	int i, rc = 0;
        for (i = 0; i < area->n_stations; i++) {
		TaStation *station;
		station = area->stations[i];
		if (!station)
			continue;
		if (!station->is_initialized) {
			station->n_updates = 0;
			station->is_initialized = _True;
			rc |= taInitStationData(area, station);
		}
        }
	return rc;
}

/* ------------------------------------------------------- */
void taCloseStationsData(TaArea *area)
{
	int i;
        for (i = 0; i < area->n_stations; i++) {
		TaStation *station;
		station = area->stations[i];
		if (!station)
			continue;
		if (station->is_initialized)
			taCloseStationData(station);
        }
}

/* ------------------------------------------------------- */
int taProcessPhaseMessages(TaArea *area)
{
	int rc = 0;
	if (!area->qos.tp_phase_correct)
		return 0;	// correction disabled
	// maybe the smallest sid instead of sid==0 in the future
	if (!area->host_station || area->host_station->sid != 0) {
		area->second_phase = 0;
		// phase correction turned on
		if (area->n_phase_update) {
			if (area->n_phase_update != area->phase_recv) {
				// here
				taPhaseCorrect(area, area->n_phase_update);
				area->n_phase_update = 0;
			} else {
				++area->n_phase_update;
				if (area->n_phase_update > TA_MAX_N_FRAMES)
					area->n_phase_update = 0;
			}
		}
	} else if (area->t_crtu >= area->t_phase_last
	+ TA_TP(area->qos.tp_phase_correct)) {
		// start messaging second phase 1..N_FRAMES
		area->second_phase = 1;
		area->t_phase_last = area->t_crtu;
	} else if (area->second_phase >= TA_MAX_N_FRAMES-1)
		area->second_phase = 0;
	else if (area->second_phase > 0)
		++area->second_phase;
	if (area->second_phase == 1)
		rc |= taSendPhaseMessages(area, 0);
	return rc;
}

/* ------------------------------------------------------- */
void taPhaseRecvCorrect(TaArea *area, TaUShort phase)
{
	if (!area->n_phase_update && phase == 0)
		area->n_phase_update = 1;
	area->phase_recv = phase;
}

/* ------------------------------------------------------- */
void taPhaseCorrect(TaArea *area, TaUShort crt_phase)
{
	int dnst, frame;
	if (!area->host_station)
		return;
	dnst = TA_MAX_N_FRAMES/area->n_stations;
	frame = area->host_station->sid*dnst;
	//area->phase_correction = (TaShort)(frame-crt_phase);
	area->phase_correction = (TaShort)(frame-crt_phase)*7/10;
	if (ta_trace_level & TA_TRAL_MESSAGE)
		fprintf(ta_file, "crt_phase:%d, correction:%d\n", 
			crt_phase, area->phase_correction);
}
