/*
 *                       ***** TA Channel *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>

#ifndef _WINDOWS
        #include <fcntl.h>
        #include <unistd.h>
        #include <netdb.h>
        #include <arpa/inet.h>
        #include <sys/un.h>
        #include <sys/poll.h>
        #include <netinet/tcp.h>
#else
        #define MSG_NOSIGNAL    0
#endif


static TaChannelQos g_channel_qos = {
	1, 		// n_repeats
	32768, 		// mem_bufsize
	_False,		// is_less_be
	1,		// max_send_polls
	_False,		// tcp_nodelay
	_False,		// so_keepalive
	0,		// so_reuseaddr
	1000000,	// tp_connect_timeout
	60000000,	// tp_disconnect_timeout
	_True,		// shm_rm_onexit
	_True,		// ping_set_err
	_False,		// assign_sids
	_True};		// check_host_ip

int taInitSocket(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel);
void taCloseSocket(TaChannel *channel);
void taShutdownSocket(TaChannel *channel);
int taTryConnectTcpChannel(TaArea *area, TaStation *dstation, int cid);
int taSendByUDP(const TaArea *area, const char *buffer, int buffer_length,
	TaChannel *channel);
int taSendByTCP(const TaArea *area, const char *buffer, int buffer_length,
	TaChannel *channel);
int taRecvByUDP(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);
int taRecvByTCP(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);
#ifndef _WINDOWS
int taInitUnixSocket(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel);
#endif
const char *taPrintErrno(int errnum);

// standart list of protocols
TaChannelProtocol g_protocols[TA_N_PROTOCOLS] = {
	{"UDP_MCAST", TA_CHANNEL_UDP_MCAST, TA_FLAGS_DGRAM|TA_FLAGS_MCAST, 
		taInitSocket, taCloseSocket, NULL, NULL, 
                taSendByUDP, taRecvByUDP, taPrintErrno},
	{"UDP", TA_CHANNEL_UDP, TA_FLAGS_DGRAM, 
		taInitSocket, taCloseSocket, NULL, NULL,
                taSendByUDP, taRecvByUDP, taPrintErrno},
	{"TCP_SERVER", TA_CHANNEL_TCP_SERVER, TA_FLAGS_STREAM|TA_FLAGS_LISTEN, 
		taInitSocket, taCloseSocket, taShutdownSocket, NULL,
                taSendByTCP, taRecvByTCP, taPrintErrno},
	{"TCP_CLIENT", TA_CHANNEL_TCP_CLIENT, TA_FLAGS_STREAM, 
		taInitSocket, taCloseSocket, taShutdownSocket, 
		taTryConnectTcpChannel, taSendByTCP, taRecvByTCP, 
                taPrintErrno}
#ifndef _WINDOWS
		,
	{"UNIX_DOMAIN", TA_CHANNEL_UNIX_DOMAIN, 
		TA_FLAGS_DGRAM|TA_FLAGS_UNIX_DOMAIN, 
		taInitUnixSocket, taCloseSocket, NULL, 
                NULL, taSendByUDP, taRecvByUDP, taPrintErrno}
#endif
};

/* ------------------------------------------------------- */
int taAddChannelProtocol(const TaChannelProtocol *protocol)
{
	if (!protocol || protocol->ctype < 0 || 
	protocol->ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	g_protocols[protocol->ctype] = *protocol;
	return 0;
}

/* ------------------------------------------------------- */
const TaChannelProtocol *taChannelProtocol(int ctype)
{
	if (ctype < 0 || ctype >= TA_N_PROTOCOLS)
		return NULL;
	return &g_protocols[ctype];
}

/* ------------------------------------------------------- */
void taDefaultChannelQos(TaChannelQos *qos)
{
	*qos = g_channel_qos;
}

/* ------------------------------------------------------- */
int taInitChannel(int ctype, const TaChannelQos *qos,
	TaChannel *channel)
{
	if (!channel || ctype < 0 || ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	if (ctype != g_protocols[ctype].ctype)
		return TA_NOT_FOUND;	// no such protocol
	memset(channel, 0, sizeof(TaChannel));
	channel->ctype = ctype;
	channel->flags = g_protocols[ctype].flags;
	if (qos)
		channel->qos = *qos;
	else
		channel->qos = g_channel_qos;
	channel->sid_acked = _True;
	return 0;
}

/* ------------------------------------------------------- */
int taInitChannelData(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	int rc = 0;
	if (channel->ctype < 0 || channel->ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	channel->recv_rc = channel->send_rc = channel->complete_recv_rc =
		channel->complete_send_rc = channel->recv_errno =
		channel->send_errno = 0;
	if (g_protocols[channel->ctype].init_proc)
		rc = (*g_protocols[channel->ctype].init_proc)(is_datahost, 
			is_send_station, station, channel);
	return rc;
}

/* ------------------------------------------------------- */
int taInitSocket(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
        int rc, my_errno;
#ifndef _WINDOWS
        int fl;
#else
        u_long v;
#endif
	struct addrinfo hints, *local;
	char buffer[32];
	struct in_addr inaddr;
	TaBool listening = _False;
	char *hostname;
	int valtrue = 1;

	if (!channel)
		return TA_ERR_ARGUMENT;
	if (channel->sd)
		taCloseChannelData(channel);

	// create addrs
	hostname = channel->hostname;
#ifdef _WINDOWS
        //rc = inet_pton(AF_INET, channel->hostname, &inaddr);
		memset(&inaddr, 0, sizeof(struct in_addr));
		inaddr.s_addr = inet_addr(channel->hostname);
		if (inaddr.s_addr == INADDR_NONE)
			rc = 0;
		else
			rc = 1;
#else
        rc = inet_aton(channel->hostname, &inaddr);
#endif
	if (!rc && isalpha(channel->hostname[0])) {
		struct hostent *h;
		h=gethostbyname(channel->hostname);
		if (h) {
			memcpy(&inaddr, h->h_addr, sizeof(struct in_addr));
			rc = 1;
		}
	}
	if (!rc) {
		if (ta_trace_level & TA_TRAL_TRANSPORT)
			fprintf(ta_file, "*** inet_aton for '%s'\n",
				channel->hostname);
		return TA_ERR_WRONG_DATA;
	}
        channel->s_addr0 = inaddr.s_addr;
	memset(&channel->sa_from, 0, sizeof(struct sockaddr_in));
	channel->sa_from.sin_addr.s_addr =
                (unsigned long int) channel->s_addr0;
	channel->sa_from.sin_family = AF_INET;
	channel->sa_from.sin_port = htons(channel->port);
	channel->sa_len = sizeof(struct sockaddr_in);

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_INET;
	switch (channel->ctype) {
	case TA_CHANNEL_UDP_MCAST:
		if (!is_datahost && !is_send_station)
			return 0;
		hints.ai_socktype = SOCK_DGRAM;
		break;
	case TA_CHANNEL_UDP:
		if (!is_datahost && !is_send_station)
			return 0;
		hints.ai_socktype = SOCK_DGRAM;
		hostname = NULL;
		break;
	case TA_CHANNEL_TCP_SERVER:
		if (!is_datahost)
			return 0;
		hints.ai_socktype = SOCK_STREAM;
		hostname = NULL;
		channel->sid_acked = _False;
		break;
	case TA_CHANNEL_TCP_CLIENT:
		if (is_datahost)
			return 0;
		hints.ai_socktype = SOCK_STREAM;
		hostname = NULL;
		break;
	}

	// is service possible?
	sprintf(buffer, "%d", channel->port);
	if ((rc = getaddrinfo(hostname, buffer, &hints, &local)) != 0) {
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "*** getaddrinfo :%d host=%s port=%s rc=%d %s\n",
				channel->ctype, channel->hostname, buffer, rc,
                strerror(rc));
		return TA_ERR_WRONG_DATA;
	}

	channel->sd = socket(local->ai_family, local->ai_socktype, 
		local->ai_protocol);
        my_errno = errno;
	if (channel->sd <= 0) {
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "*** socket rc=%d, err=%d %s\n",
                                channel->sd, my_errno, strerror(my_errno));
		return TA_ERR_OUT_OF_MEMORY;
        }
#ifndef _WINDOWS
	fl = O_NONBLOCK;
	if (fcntl(channel->sd, F_SETFL, fl) < 0) {
#else
	v = 1;
        if (ioctlsocket(channel->sd, FIONBIO, &v) != 0) {
#endif
                my_errno = errno;
                if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
                        fprintf(ta_file, "*** NONBLOCK socket err:%d\n", my_errno);
		taCloseChannelData(channel);
		return TA_ERR_WRONG_DATA;
	}

	if (channel->qos.tcp_nodelay && 
	(channel->ctype == TA_CHANNEL_TCP_SERVER 
	|| channel->ctype == TA_CHANNEL_TCP_CLIENT)) {
                if (setsockopt(channel->sd, IPPROTO_TCP, TCP_NODELAY,
		(char*)&valtrue, sizeof(int)) < 0) {
			if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt NODELAY\n");
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}
	if (channel->qos.so_reuseaddr > 0
	|| (channel->port && is_datahost && channel->qos.so_reuseaddr == 0)) {
		if (setsockopt(channel->sd, SOL_SOCKET, SO_REUSEADDR,
		(char*)&valtrue, sizeof(int)) < 0) {
			if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt REUSEADDR\n");
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}
	if (channel->qos.so_keepalive) {
		if (setsockopt(channel->sd, SOL_SOCKET, SO_KEEPALIVE,
		(char*)&valtrue, sizeof(int)) < 0) {
			if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt KEEPALIVE\n");
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}
	if (channel->port && is_datahost) {
		// bind
#ifdef _WINDOWS
		if (channel->ctype == TA_CHANNEL_UDP_MCAST) {
			struct sockaddr_in *addr = (struct sockaddr_in *) local->ai_addr;
			addr->sin_addr.s_addr = htonl(INADDR_ANY);
		}
#endif
		rc = bind(channel->sd, local->ai_addr, local->ai_addrlen);
                my_errno = errno;
		if (rc < 0) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file, "*** bind socket :%d port=%d err:%d '%s'\n",
					channel->ctype, channel->port, my_errno,
                                        strerror(my_errno));
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
		// multicast after bind
		if (channel->ctype == TA_CHANNEL_UDP_MCAST) {
			struct ip_mreq mreq;
			unsigned char ttl = 32;
			/*
			ownid instead of inaddr
			if (setsockopt(channel->sd, IPPROTO_IP, IP_MULTICAST_IF,
			(char*)&inaddr, sizeof(struct in_addr)) < 0) {
				printf("*** setsockopt MULTICAST_IF\n");
				taCloseChannelData(channel);
				return TA_ERR_WRONG_DATA;
			}
			*/
			memset(&mreq, 0, sizeof(struct ip_mreq));
			mreq.imr_multiaddr = inaddr;
			if (setsockopt(channel->sd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			(char*)&mreq, sizeof(struct ip_mreq)) < 0) {
				if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt ADD_MEMBERSHIP, "
					"check multicast tuning for %s\n", 
					station->qos.group_addr);
				taCloseChannelData(channel);
				return TA_ERR_WRONG_DATA;
			}
			if (setsockopt(channel->sd, IPPROTO_IP, IP_MULTICAST_TTL,
			(char*)&ttl, sizeof(unsigned char)) < 0) {
				if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt MULTICAST_TTL\n");
				taCloseChannelData(channel);
				return TA_ERR_WRONG_DATA;
			}
			if (setsockopt(channel->sd, IPPROTO_IP, IP_MULTICAST_LOOP,
			(char*)&valtrue, sizeof(int)) < 0) {
				if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file, "*** setsockopt MULTICAST_LOOP\n");
				taCloseChannelData(channel);
				return TA_ERR_WRONG_DATA;
			}
		} else if (channel->ctype == TA_CHANNEL_TCP_SERVER) {
                        if (listen(channel->sd, 5) < 0) {
                                my_errno = errno;
				if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
                                    fprintf(ta_file, "*** listen socket port=%d err:%d '%s'\n",
                                                channel->port, my_errno,
                                                strerror(my_errno));
				taCloseChannelData(channel);
				return TA_ERR_WRONG_DATA;
			}
			listening = _True;
		}
	}

	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "%s socket sd=%d host=%s %s port=%d ctype=%d %s\n",
			g_protocols[channel->ctype].name,
			channel->sd, channel->hostname, 
			(is_datahost ? "bind" : "nobind"), channel->port,
			channel->ctype, (listening ? "listening" : ""));
	if (listening)
		channel->link_stat = TA_LINK_LISTEN;
	else if (channel->flags & TA_FLAGS_DGRAM)
		channel->link_stat = TA_LINK_OK;
	else
		channel->link_stat = TA_LINK_INIT;
	return 1;
}

/* ------------------------------------------------------- */
#ifndef _WINDOWS
int taInitUnixSocket(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	int rc, my_errno;
        int fl;
	int valtrue = 1;
	struct sockaddr_un unaddr;

	if (!channel)
		return TA_ERR_ARGUMENT;
	if (channel->sd)
		taCloseChannelData(channel);

	// create addrs
	if (!is_datahost && !is_send_station)
		return 0;
	unaddr.sun_family = AF_UNIX;
	memset(unaddr.sun_path, 0, sizeof(unaddr.sun_path));
	memcpy(&unaddr.sun_path[1], channel->hostname,
		strlen(channel->hostname));

	// is service possible?
	channel->sd = socket(PF_UNIX, SOCK_DGRAM, 0);
        my_errno = errno;
	if (channel->sd <= 0) {
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "*** unix socket rc=%d, err=%d %s\n",
                                channel->sd, my_errno, strerror(my_errno));
		return TA_ERR_OUT_OF_MEMORY;
        }
	fl = O_NONBLOCK;
	if (fcntl(channel->sd, F_SETFL, fl) < 0) {
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "*** NONBLOCK socket \n");
		taCloseChannelData(channel);
		return TA_ERR_WRONG_DATA;
	}

	if (channel->qos.so_reuseaddr > 0
	|| (channel->port && is_datahost && channel->qos.so_reuseaddr == 0)) {
		if (setsockopt(channel->sd, SOL_SOCKET, SO_REUSEADDR,
		(char*)&valtrue, sizeof(int)) < 0) {
			if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file, "*** setsockopt REUSEADDR\n");
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}
	if (channel->qos.so_keepalive) {
		if (setsockopt(channel->sd, SOL_SOCKET, SO_KEEPALIVE,
		(char*)&valtrue, sizeof(int)) < 0) {
			if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file, "*** setsockopt KEEPALIVE\n");
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}

		// bind
	if (is_datahost) {
		rc = bind(channel->sd, (struct sockaddr *)&unaddr, 
			sizeof(unaddr));
                my_errno = errno;
		if (rc < 0) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
                            fprintf(ta_file, "*** bind socket :%d port=%d err:%d '%s'\n",
					channel->ctype, channel->port, 
                                        my_errno, strerror(my_errno));
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
	}

	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "%s unix socket bind:%d sd=%d %s port=%d ctype=%d\n",
			g_protocols[channel->ctype].name, is_datahost,
			channel->sd, channel->hostname, 
			channel->port, channel->ctype);
	channel->link_stat = TA_LINK_OK;
	return 1;
}
#endif

/* ------------------------------------------------------- */
void taCloseChannelData(TaChannel *channel)
{
	if (channel->ctype < 0 || channel->ctype >= TA_N_PROTOCOLS)
		return;
	if (g_protocols[channel->ctype].close_proc)
		(*g_protocols[channel->ctype].close_proc)(channel);
}

/* ------------------------------------------------------- */
void taCloseSocket(TaChannel *channel)
{
	if (channel->sd == 0)
		return;
#ifndef _WINDOWS
	close(channel->sd);
#else
        closesocket(channel->sd);
#endif
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "? taCloseChannelData %d\n", channel->sd);
	channel->sd = 0;
	channel->link_stat = TA_LINK_CLOSED;
}

/* ------------------------------------------------------- */
void taShutdownChannelData(TaChannel *channel)
{
	if (channel->ctype < 0 || channel->ctype >= TA_N_PROTOCOLS)
		return;
	if (g_protocols[channel->ctype].shutdown_proc)
		(*g_protocols[channel->ctype].shutdown_proc)(channel);
}

/* ------------------------------------------------------- */
void taShutdownSocket(TaChannel *channel)
{
	if (channel->sd == 0)
		return;
	if (channel->ctype == TA_CHANNEL_TCP_CLIENT) {
		shutdown(channel->sd, 2);
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "? taShutdownChannelData %d\n", 
				channel->sd);
		taCloseSocket(channel);
		taInitChannelData(_False, _False, 
			NULL, channel);
		channel->link_stat = TA_LINK_INIT;
	} else if (channel->ctype == TA_CHANNEL_TCP_SERVER)
		taCloseSocket(channel);
}

/* ------------------------------------------------------- */
//#define WITH_POSIX_POLL
#ifndef _WINDOWS
#ifndef WITH_POSIX_POLL
#define WITH_POSIX_POLL
#endif
#endif

#ifdef WITH_POSIX_POLL
#define SET_SD_EVENTS(n,sd,re,we,xe); ufds[n].fd = sd; ufds[n].events = re | we | xe;
#define GET_SD_EVENTS(n,channel); channel->poll_events = ufds[n].revents;
#else
#define SET_SD_EVENTS(n,sd,re,we,xe); if(re) {rf=&rfds;FD_SET(sd,rf);} if(we) {wf=&wfds;FD_SET(sd,wf);} if(xe) {xf=&xfds;FD_SET(sd,xf);} if(sd>=nmax) nmax=sd+1;
#define GET_SD_EVENTS(n,channel); channel->poll_events = 0; if(FD_ISSET(channel->sd,&rfds)) channel->poll_events |= POLLIN; if(FD_ISSET(channel->sd,&wfds)) channel->poll_events |= POLLOUT; if(FD_ISSET(channel->sd,&xfds)) channel->poll_events |= (POLLERR|POLLHUP);
#endif
int taPollChannels(TaArea *area)
{
        int i, k, rc, n, my_errno;
	TaStation *station;
	TaChannel *channel;
#ifdef WITH_POSIX_POLL
	struct pollfd ufds[TA_MAX_POLL_CHANNELS];

	memset(ufds, 0, TA_MAX_POLL_CHANNELS * sizeof(struct pollfd));
#else
	fd_set rfds, wfds, xfds, *rf = NULL, *wf = NULL, *xf = NULL;
	int nmax = 0;
	struct timeval tv;
	FD_ZERO(&rfds); FD_ZERO(&wfds); FD_ZERO(&xfds);
#endif
	n = 0;
	for (i = 0; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!station || station->is_user_locked)
			continue;
		station->n_ready_for_sending = 0;
		station->n_available_for_sending = 0;
		for (k = 0; k < station->n_channels; k++) {
			channel = &station->channels[k];
			channel->poll_events = 0;
			if (!channel->is_enabled
			|| channel->link_stat <= TA_LINK_CLOSED 
			|| channel->trans_flags == 0)
				continue;
			switch (channel->ctype) {
			case TA_CHANNEL_UDP_MCAST:
			case TA_CHANNEL_UDP:
				if (station == area->host_station) {
					// recv 
					SET_SD_EVENTS(n, channel->sd, POLLIN, 0, POLLERR);
					++n;
				} else if (station == area->send_station) {
					// send
					SET_SD_EVENTS(n, channel->sd, 0, POLLOUT, POLLERR);
					++n;
					++station->n_available_for_sending;
				}
				break;
			case TA_CHANNEL_TCP_SERVER:
				if (station == area->host_station) {
					// accept
					SET_SD_EVENTS(n, channel->sd, POLLIN, 0, POLLERR);
				} else {
					// accept
					SET_SD_EVENTS(n, channel->sd, POLLIN, POLLOUT,
						(channel->sid_acked ? POLLERR | POLLHUP : POLLHUP));
					++station->n_available_for_sending;
				}
				++n;
				break;
			case TA_CHANNEL_TCP_CLIENT:
				if (station == area->host_station)
					break;
				SET_SD_EVENTS(n, channel->sd, POLLIN, POLLOUT,
					POLLERR | POLLHUP);
				++n;
				++station->n_available_for_sending;
				break;
			case TA_CHANNEL_ICMP:
				if (station == area->send_station) {
					SET_SD_EVENTS(n, channel->sd, POLLIN, POLLOUT,
						POLLERR);
					++n;
					++station->n_ready_for_sending;
					++station->n_available_for_sending;
				}
				break;
			default:
				if (channel->link_stat != TA_LINK_OK)
					break;
				channel->poll_events = 0;
				if (channel->trans_flags & TA_TRANS_SEND) {
					channel->poll_events |= POLLOUT;
					++station->n_ready_for_sending;
					++station->n_available_for_sending;
				}
				if (channel->trans_flags & TA_TRANS_RECV)
					channel->poll_events |= POLLIN;
				break;
			}
#ifdef _WINDOWS
#ifdef WITH_POSIX_POLL
                        if (n>0) ufds[n-1].events &= (POLLIN|POLLOUT);
#endif
#endif
		}
	}

	if (n <= 0)
		return 0;
	// poll it
#ifndef WITH_POSIX_POLL
	tv.tv_sec = 0; tv.tv_usec = 0;
	rc = select(nmax, rf, wf, xf, &tv);
#else
#ifndef _WINDOWS
	rc = poll(ufds, n, 0);
#else
        rc = WSAPoll(ufds, n, 0);
#endif
#endif
        my_errno = errno;
	if (rc > 0)
		area->n_poll_fails = 0;
	else {
		++area->n_poll_fails;
                if (area->qos.max_poll_fails
		&& area->n_poll_fails >= area->qos.max_poll_fails) {
			n = 0;
			rc = 0;
			for (i = 0; i < area->n_stations; i++) {
				if (area->stations[i]->qos.ping_mask == 0)
					continue;
				if (area->stations[i]) {
					rc |= taInitStationData(area, 
						area->stations[i]);
					++n;
				}
			}
			area->n_poll_fails = 0;
			if (n > 0) {
				if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
					fprintf(ta_file, "* on poll err: reinit all channels\n");
			}
			return 0;
		}
	}
	if (rc < 0) {
		if (ta_trace_level & TA_TRAL_TRANSPORT)
                    fprintf(ta_file, "? poll rc:%d err:%d %s\n", rc, my_errno,
                           strerror(my_errno));
		return TA_ERR_TRANSPORT;
	}
	//if (rc == 0 && (ta_trace_level & TA_TRAL_TRANSPORT))
	//	fprintf(ta_file, "? poll timeout rc=0\n");

	n = 0;
	for (i = 0; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!station || station->is_user_locked)
			continue;
		for (k = 0; k < station->n_channels; k++) {
			channel = &station->channels[k];
			if (!channel->is_enabled || channel->sd <= 0
			|| channel->trans_flags == 0)
				continue;
			switch (channel->ctype) {
			case TA_CHANNEL_UDP_MCAST:
			case TA_CHANNEL_UDP:
				if (station == area->host_station
				||  station == area->send_station) {
					GET_SD_EVENTS(n, channel);
					++n;
				}
				if (channel->poll_events & POLLOUT)
					++station->n_ready_for_sending;
				break;
			case TA_CHANNEL_TCP_SERVER:
				GET_SD_EVENTS(n, channel);
				++n;
				if (station == area->host_station)
					break;
				if (channel->poll_events & POLLHUP) {
					if (ta_trace_level & TA_TRAL_TRANSPORT)
						fprintf(ta_file,
					"close socket sd:%d on hup poll:%d\n", 
							channel->sd, 
							channel->poll_events);
					taCloseChannelData(channel);
				}
				if (channel->poll_events & POLLOUT)
					++station->n_ready_for_sending;
				break;
			case TA_CHANNEL_TCP_CLIENT:
				if (station == area->host_station)
					break;
				GET_SD_EVENTS(n, channel);
				++n;
				if (channel->poll_events & POLLHUP 
				&& channel->link_stat == TA_LINK_OK) {
					if (ta_trace_level & TA_TRAL_TRANSPORT)
						fprintf(ta_file,
				"shutdown connect socket sd:%d on hup poll:%d\n", 
							channel->sd, 
							channel->poll_events);
					taShutdownChannelData(channel);
				}
				if (channel->poll_events & POLLOUT)
					++station->n_ready_for_sending;
				break;
			case TA_CHANNEL_ICMP:
				if (station == area->send_station) {
					GET_SD_EVENTS(n, channel);
					++n;
				}
				break;
			default:
				break;
			}
		}
	}

	return rc;	// number of
}

/* ------------------------------------------------------- */
int taSendByChannel(const TaArea *area, const char *buffer, int buffer_length,
	TaChannel *channel)
{
	int rc = 0;
	if (channel->ctype < 0 || channel->ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	if (channel->link_stat != TA_LINK_OK)
		return TA_ERR_LINK;
	if (g_protocols[channel->ctype].send_proc)
		channel->send_rc = rc = 
		(*g_protocols[channel->ctype].send_proc)(area, buffer,
			buffer_length, channel);
	if (rc)
		channel->complete_send_rc = rc;
	if (rc > 0) {
		channel->send_errno = 0;
		channel->t_csend = area->t_crtu;
	}
	return rc;
}

/* ------------------------------------------------------- */
int taSendByUDP(const TaArea *area, const char *buffer, int buffer_length,
	TaChannel *channel)
{
        int rc = 0, my_errno;
	TaChannel *dstchannel;
	struct sockaddr *sa_from;
	socklen_t sa_len;
#ifndef _WINDOWS
	struct sockaddr_un unaddr;
#endif
	if (!area->send_station)
		return TA_ERR_ARGUMENT;
	dstchannel = &area->send_station->channels[channel->cid];
	if (ta_trace_level & TA_TRAL_POLLING)
		fprintf(ta_file, "sendto.poll=%d sd:%d poll_events:%d sid:%d\n",
			!(!(dstchannel->poll_events & POLLOUT) 
			|| dstchannel->sd <= 0), dstchannel->sd,
			dstchannel->poll_events, (int)area->send_station->sid);
	if (dstchannel->sd <= 0)
		return 0;
	if (dstchannel->poll_events & POLLOUT) {
		dstchannel->n_send_polls = 0;
	} else {
		++dstchannel->n_send_polls;
		if (dstchannel->qos.max_send_polls
		&& dstchannel->n_send_polls 
		>= dstchannel->qos.max_send_polls)
			return 0;
	}
	if (channel->ctype == TA_CHANNEL_UDP_MCAST) {
		sa_from = (struct sockaddr *)&dstchannel->sa_from;
		sa_len = dstchannel->sa_len;
#ifndef _WINDOWS
	} else if (channel->ctype == TA_CHANNEL_UNIX_DOMAIN) {
		sa_from = (struct sockaddr *)&unaddr;
		sa_len = sizeof(unaddr);
		unaddr.sun_family = AF_UNIX;
		memset(unaddr.sun_path, 0, sizeof(unaddr.sun_path));
		memcpy(&unaddr.sun_path[1], channel->hostname,
			strlen(channel->hostname));
#endif
	} else {
		sa_from = (struct sockaddr *)&channel->sa_from;
		sa_len = channel->sa_len;
	}
	rc =
	sendto(dstchannel->sd, buffer, buffer_length, 
		MSG_NOSIGNAL, (const struct sockaddr *)sa_from, 
			sa_len);
        channel->send_errno = my_errno = errno;
	if (ta_trace_level & TA_TRAL_TRANSPORT 
	|| rc < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		char clienthost[NI_MAXHOST];
		char clientservice[NI_MAXSERV];
		clienthost[0] = clientservice[0] = 0;
		getnameinfo((const struct sockaddr *)sa_from, sa_len, 
			clienthost, sizeof(clienthost), clientservice, 
			sizeof(clientservice), 
			NI_NUMERICHOST|NI_NUMERICSERV);
		if (rc > 0)
			fprintf(ta_file,
			"sendto %s port=%s sd=%d rc=%d len=%d\n",
				clienthost, clientservice, dstchannel->sd, 
					rc, buffer_length);
		else 
			fprintf(ta_file,
			"* sendto %s port=%s sd=%d rc=%d len=%d"
                        " err:%d '%s'\n",
				clienthost, clientservice, dstchannel->sd, 
                                rc, buffer_length, my_errno,
                                strerror(my_errno));
	}
	if (rc < 0)
		return TA_ERR_TRANSPORT;
	return rc;
}

/* ------------------------------------------------------- */
int taSendByTCP(const TaArea *area, const char *buffer, int buffer_length,
	TaChannel *channel)
{
        int rc = 0, my_errno;
	if (ta_trace_level & TA_TRAL_POLLING)
		fprintf(ta_file, "send.poll=%d sd:%d poll_events:%d\n",
			!(!(channel->poll_events & POLLOUT) 
			|| channel->sd <= 0), channel->sd,
			channel->poll_events);
	if (channel->sd <= 0)
		return 0;
	if (channel->poll_events & POLLOUT) {
		channel->n_send_polls = 0;
	} else {
		++channel->n_send_polls;
		if (channel->qos.max_send_polls &&
		channel->n_send_polls >= channel->qos.max_send_polls)
			return 0;
	}
	rc =
	send(channel->sd, buffer, buffer_length, MSG_NOSIGNAL);
        channel->send_errno = my_errno = errno;
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| (rc < 0 || !(channel->poll_events & POLLOUT)) && (ta_trace_level == TA_TRAL_ERRORS)) {
		if (rc > 0) {
			fprintf(ta_file,
			"send port=%d sd=%d rc=%d len=%d\n",
				channel->port, channel->sd, 
					rc, buffer_length);
		} else 
			fprintf(ta_file,
			"* send port=%d sd=%d rc=%d len=%d"
                        " err:%d '%s'\n",
				channel->port, channel->sd, rc, 
                                buffer_length, my_errno,
                                strerror(my_errno));
	}
#ifndef _WINDOWS
	if (rc < 0 && channel->send_errno == EPIPE) {
#else
	if (rc < 0 && channel->send_errno == WSAECONNRESET) {
#endif
		taShutdownChannelData(channel);
		return TA_ERR_TRANSPORT;
	}
	if (rc < 0)
		return TA_ERR_TRANSPORT;
	return rc;
}

/* ------------------------------------------------------- */
int taTryAcceptChannels(TaArea *area, int cid)
{
	TaChannel *channel = NULL;
        int i, sd, n_accepted, my_errno;
	if (!area || !area->host_station)
		return TA_ERR_ARGUMENT;
	channel = &area->host_station->channels[cid];
	if (channel->ctype != TA_CHANNEL_TCP_SERVER
	|| !channel->is_enabled || channel->sd <= 0)
		return TA_ERR_WRONG_DATA;

	if (ta_trace_level & TA_TRAL_POLLING)
		fprintf(ta_file, "accept.poll=%d poll_events:%d\n",
			(channel->poll_events & POLLIN), 
			channel->poll_events);
	if (!(channel->poll_events & POLLIN))
		return 0;	// nothing to accept

	n_accepted = 0;
	while (1) {
		TaStation *dstation;
		TaChannel *dstchannel;
		struct sockaddr_in sa_from;
		socklen_t sa_len = sizeof(struct sockaddr_in);
		TaBool is_accepted = _False;
		memset(&sa_from, 0, sizeof(struct sockaddr_in));
		sa_from.sin_addr.s_addr = INADDR_ANY;
		sa_from.sin_family = AF_INET;
		sa_from.sin_port = htons(channel->port);
		sd =
		accept(channel->sd, (struct sockaddr *)&sa_from, &sa_len);
                my_errno = errno;
		if (sd <= 0) {
                        if (my_errno == TA_EC(EWOULDBLOCK))
				break;
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
                            fprintf(ta_file, "*** accept socket sd:%d port=%d err:%d '%s'\n",
                                        channel->sd, channel->port, my_errno,
                                        strerror(my_errno));
			return TA_ERR_TRANSPORT;
		} else if (ta_trace_level & TA_TRAL_TRANSPORT) {
			char clienthost[NI_MAXHOST];
			char clientservice[NI_MAXSERV];
			getnameinfo((const struct sockaddr *)&sa_from,
				sa_len, clienthost, 
				sizeof(clienthost), clientservice, 
				sizeof(clientservice), 
				NI_NUMERICHOST|NI_NUMERICSERV);
			fprintf(ta_file, "accepted sd:%d host:%s port:%s\n",
				sd, clienthost, clientservice);
		}
		// try to find proper addr 
		for (i = 0; i < area->n_stations; i++) {
			dstation = area->stations[i];
			if (!dstation || dstation == area->host_station)
				continue;
			dstchannel = &dstation->channels[cid];
			if (!dstchannel->is_enabled)
				continue;
			/*if (memcmp(&sa_from, dstchannel->sa_from, 
			sizeof(sockaddr_in)))
				continue;*/
			if (dstchannel->qos.check_host_ip && sa_from.sin_addr.s_addr 
			!= dstchannel->sa_from.sin_addr.s_addr)
				continue;
			if (dstchannel->sd > 0) {
				// accepted, only one channel for station
				continue;
				//close(dstchannel->sd);
				//dstchannel->sd = 0;
			}
			dstchannel->sd = sd;
			dstchannel->sid_acked = _False;
			++n_accepted;
			is_accepted = _True;
			dstchannel->t_crecv = area->t_crtu;
			dstchannel->is_receiving = dstchannel->is_receiving2 
				= _True;
			dstchannel->link_stat = TA_LINK_OK;
			break;
		}
		if (is_accepted) {
#ifndef _WINDOWS
			int fl;
			fl = O_NONBLOCK;
			if (fcntl(sd, F_SETFL, fl) < 0)
#else
			u_long v;
			v = 1;
			if (ioctlsocket(sd, FIONBIO, &v) != 0)
#endif
			{
                                my_errno = errno;
				if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
                                    fprintf(ta_file, "*** NONBLOCK socket err:%d\n", my_errno);
				taCloseChannelData(dstchannel);
				return TA_ERR_WRONG_DATA;
			}
		} else {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file, "*** unexpected host accepted at sd:%d\n",
					channel->sd);
#ifdef _WINDOWS
                        closesocket(sd);
#else
			close(sd);
#endif
		}
	}
	return 1;
}

/* ------------------------------------------------------- */
int taTryConnectChannel(TaArea *area, TaStation *dstation, int cid)
{
	int rc = 0;
	TaChannel *dstchannel;
	if (!dstation || cid < 0 || cid >= dstation->n_channels)
		return TA_ERR_ARGUMENT;
	dstchannel = &dstation->channels[cid];
	if (dstchannel->ctype < 0 || dstchannel->ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	if (g_protocols[dstchannel->ctype].connect_proc)
		rc = (*g_protocols[dstchannel->ctype].connect_proc)(area, 
			dstation, cid);
	return rc;
}

/* ------------------------------------------------------- */
int taTryConnectTcpChannel(TaArea *area, TaStation *dstation, int cid)
{
	TaChannel *dstchannel;
        int rc, my_errno;
	if (!area || !dstation)
		return TA_ERR_ARGUMENT;
	dstchannel = &dstation->channels[cid];
	if (!dstchannel->is_enabled || dstchannel->sd <= 0)
		return TA_ERR_WRONG_DATA;
	switch (dstchannel->link_stat) {
	case TA_LINK_INIT:
		//if (!(dstchannel->poll_events & POLLOUT))
		//	return 0;
		rc = connect(dstchannel->sd, (const struct sockaddr *)
			&dstchannel->sa_from, sizeof(struct sockaddr_in));
                my_errno = errno;
		if (ta_trace_level & TA_TRAL_TRANSPORT) {
				char clienthost[NI_MAXHOST];
				char clientservice[NI_MAXSERV];
				getnameinfo((const struct sockaddr *)
					&dstchannel->sa_from,
					dstchannel->sa_len, clienthost, 
					sizeof(clienthost), clientservice, 
					sizeof(clientservice), 
					NI_NUMERICHOST|NI_NUMERICSERV);
				fprintf(ta_file,
				"connecting sid:%d host:%s port:%s sd:%d\n",
					dstation->sid, clienthost, 
					clientservice, dstchannel->sd);
		}
		if (ta_trace_level & TA_TRAL_TRANSPORT)
			fprintf(ta_file,
		"connect sd:%d rc:%d errno:%d/inprogress:%d ipoll:%d %s\n",
                                dstchannel->sd, rc, my_errno, TA_EC(EINPROGRESS),
				dstchannel->poll_events,
                                (my_errno && my_errno != TA_EC(EINPROGRESS)
                                        ? strerror(my_errno) : ""));
		if (rc < 0) {
                        if (my_errno == TA_EC(EINPROGRESS)
                        || my_errno == TA_EC(EWOULDBLOCK))
				dstchannel->link_stat = TA_LINK_WAIT;
                        if (my_errno == TA_EC(EISCONN)) {
#ifdef CLOSE_ALREADY_CONNECTED
				if (ta_trace_level & TA_TRAL_TRANSPORT)
					fprintf(ta_file,
				"close already connect socket sd:%d\n", 
						dstchannel->sd);
				taShutdownChannelData(dstchannel);
#else
				dstchannel->link_stat = TA_LINK_OK;
#endif
			}
		}
		if (rc > 0) {
			// seems impossible
			dstchannel->link_stat = TA_LINK_OK;
		}
		break;
	case TA_LINK_WAIT:
		if (ta_trace_level & TA_TRAL_POLLING)
			fprintf(ta_file, "connect.poll=%d poll_events:%d\n",
				(dstchannel->poll_events & POLLOUT),
				dstchannel->poll_events);
		if (dstchannel->poll_events & POLLHUP) {
			dstchannel->link_stat = TA_LINK_INIT;
		} else if (dstchannel->poll_events & POLLOUT) {
#define SECOND_BY_GETPEERNAME
#ifdef SECOND_BY_GETPEERNAME
			socklen_t so_length;
			struct sockaddr_in sa_peer;
			so_length = sizeof(struct sockaddr_in);
			rc = getpeername(dstchannel->sd, (struct sockaddr *)
				&sa_peer, &so_length);
                        my_errno = errno;
			if (rc < 0) {
				if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
					fprintf(ta_file,
			"* connect sd:%d rc:%d errno:%d/inprogress:%d %s\n",
                                dstchannel->sd, rc, my_errno, TA_EC(EINPROGRESS),
                                (my_errno && my_errno != TA_EC(EINPROGRESS)
                                ? strerror(my_errno) : ""));
				dstchannel->link_stat = TA_LINK_INIT;
				break;
			}
#endif
#ifdef SECOND_CONNECT
			rc = connect(dstchannel->sd, (const struct sockaddr *)
				&dstchannel->sa_from, 
				sizeof(struct sockaddr_in));
                        my_errno = errno;
			if (rc < 0) {
                                if (my_errno == TA_EC(EISCONN)) {
					if (ta_trace_level & TA_TRAL_TRANSPORT)
						fprintf(ta_file,
					"already connected sd:%d\n",
							dstchannel->sd);
				} else {
					if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
						fprintf(ta_file,
			"* connect sd:%d rc:%d errno:%d/inprogress:%d %s\n",
                                        dstchannel->sd, rc, my_errno, TA_EC(EINPROGRESS),
                                        (my_errno && my_errno != TA_EC(EINPROGRESS)
                                        ? strerror(my_errno) : ""));
					dstchannel->link_stat = TA_LINK_INIT;
					break;
				}
			}
#endif
			dstchannel->link_stat = TA_LINK_OK;
			if (ta_trace_level & TA_TRAL_TRANSPORT) {
				char clienthost[NI_MAXHOST];
				char clientservice[NI_MAXSERV];
				getnameinfo((const struct sockaddr *)
					&dstchannel->sa_from,
					dstchannel->sa_len, clienthost, 
					sizeof(clienthost), clientservice, 
					sizeof(clientservice), 
					NI_NUMERICHOST|NI_NUMERICSERV);
				fprintf(ta_file, "connected host:%s port:%s sd:%d\n",
					clienthost, clientservice, 
					dstchannel->sd);
			}
		} else if (area->t_crtu > TA_TP(dstation->qos.tp_ping_timeout)
		+ dstation->t_send)
			dstchannel->link_stat = TA_LINK_INIT;
		break;
	case TA_LINK_OK:
		//dstchannel->link_stat = TA_CONNECT_NONE;
		return 0;
	default:
		break;
	}
	//dstchannel->link_stat = TA_CONNECT_DONE;
	dstchannel->t_crecv = area->t_crtu;
	dstchannel->is_receiving = dstchannel->is_receiving2 = _True;
	return 1;
}

/* ------------------------------------------------------- */
int taRecvByChannel(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel)
{
	int rc = 0;
	if (channel->ctype < 0 || channel->ctype >= TA_N_PROTOCOLS)
		return TA_ERR_ARGUMENT;
	if (channel->link_stat != TA_LINK_OK)
		return TA_ERR_LINK;
	if (g_protocols[channel->ctype].recv_proc)
		rc = channel->recv_rc =
		(*g_protocols[channel->ctype].recv_proc)(area, buffer,
			buffer_length, channel);
	if (rc < 0)
		channel->complete_recv_rc = rc;
	else if (rc > 1) {
		channel->complete_recv_rc = rc;
		channel->recv_errno = 0;
		channel->t_crecv = area->t_crtu;
	}
	return rc;
}

/* ------------------------------------------------------- */
int taRecvByUDP(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel)
{
        int rc, my_errno;
	struct sockaddr *sa_msg;
#ifndef _WINDOWS
	struct sockaddr_un unaddr;
#endif
	if (ta_trace_level & TA_TRAL_POLLING)
		fprintf(ta_file, "recv.poll=%d sd:%d poll_events:%d\n",
			!(channel->sd <= 0 || !(channel->poll_events & POLLIN)),
			channel->sd, channel->poll_events);
	if (channel->sd <= 0 || !(channel->poll_events & POLLIN))
		return 0;	// no data available
	channel->sa_len = buffer_length;
	if (channel->ctype == TA_CHANNEL_UNIX_DOMAIN) {
#ifndef _WINDOWS
		sa_msg = (struct sockaddr *)&unaddr;
		channel->sa_msg_len = sizeof(unaddr);
		unaddr.sun_family = AF_UNIX;
		memset(unaddr.sun_path, 0, sizeof(unaddr.sun_path));
		memcpy(&unaddr.sun_path[1], channel->hostname,
			strlen(channel->hostname));
#endif
	} else {
		sa_msg = (struct sockaddr *)&channel->sa_msg;
		channel->sa_msg_len = sizeof(channel->sa_msg);
	}
	rc = recvfrom(channel->sd, buffer, buffer_length, MSG_NOSIGNAL,
		sa_msg, &channel->sa_msg_len);
        channel->recv_errno = my_errno = errno;
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| rc < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		char clienthost[NI_MAXHOST];
		char clientservice[NI_MAXSERV];
		clienthost[0] = clientservice[0] = 0;
		getnameinfo((struct sockaddr *)sa_msg, 
			channel->sa_msg_len, clienthost, 
			sizeof(clienthost), clientservice, 
			sizeof(clientservice), 
			NI_NUMERICHOST|NI_NUMERICSERV);
		if (rc > 0)
			fprintf(ta_file, "recvfrom %s port=%s sd=%d rc=%d\n", 
				clienthost, clientservice, channel->sd, 
				rc);
                else if (rc < 0 && my_errno != TA_EC(EWOULDBLOCK))
			fprintf(ta_file,
                                "* recvfrom %s port=%s sd=%d rc=%d err:%d '%s'\n",
				clienthost, clientservice, channel->sd, 
                                rc, my_errno, strerror(my_errno));
	}
        if (rc < 0 && my_errno == TA_EC(EWOULDBLOCK))
		channel->recv_errno = rc = 0;
	if (rc < 0)
		return TA_ERR_TRANSPORT;
	return rc;
}

/* ------------------------------------------------------- */
int taRecvByTCP(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel)
{
        int rc, my_errno;
	if (ta_trace_level & TA_TRAL_POLLING)
		fprintf(ta_file, "recv.poll=%d sd:%d poll_events:%d\n",
			!(channel->sd <= 0 || !(channel->poll_events & POLLIN)),
			channel->sd, channel->poll_events);
	if (channel->sd <= 0 || !(channel->poll_events & POLLIN))
		return 0;	// no data available
	channel->sa_len = buffer_length;
	rc = recv(channel->sd, buffer, buffer_length, MSG_NOSIGNAL);
        channel->recv_errno = my_errno = errno;
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| rc < 0 && (ta_trace_level  &TA_TRAL_ERRORS)) {
		if (rc > 0)
			fprintf(ta_file, "recv port=%d sd=%d rc=%d\n",
				channel->port, channel->sd, rc);
                else if (rc < 0 && my_errno != TA_EC(EWOULDBLOCK))
                    fprintf(ta_file, "* recv port=%d sd=%d rc=%d err:%d '%s'\n",
				channel->port, channel->sd, rc,
                                my_errno, strerror(my_errno));
	}
	channel->sa_msg_len = 0;
	if (rc == 0) {
		if (ta_trace_level & TA_TRAL_TRANSPORT)
			fprintf(ta_file, "safe disconnect sd:%d\n", channel->sd);
		taShutdownChannelData(channel);
		return 0;
	} else if (rc == 1) {
		if (ta_trace_level & TA_TRAL_TRANSPORT)
			fprintf(ta_file, "? unsafe disconnect sd:%d\n", channel->sd);
		taShutdownChannelData(channel);
		return 0;
	}
        if (rc < 0 && my_errno == TA_EC(EWOULDBLOCK))
		channel->recv_errno = rc = 0;
#ifndef _WINDOWS
	if (rc < 0 && my_errno == EPIPE) {
#else
	if (rc < 0 && my_errno == WSAECONNRESET) {
#endif
		taShutdownChannelData(channel);
		return TA_ERR_TRANSPORT;
	}
	if (rc < 0)
		return TA_ERR_TRANSPORT;
	return rc;
}

/* ------------------------------------------------------- */
TaBool taCheckChannels(TaChannel *msg_channel, TaChannel *sid_channel)
{
	/*if (ta_trace_level & TA_TRAL_TRANSPORT) {
		char clienthost[NI_MAXHOST], sidhost[NI_MAXHOST];
		char clientservice[NI_MAXSERV], sidservice[NI_MAXSERV];
		getnameinfo((sockaddr *)&msg_channel->sa_msg, 
			msg_channel->sa_msg_len, clienthost, 
			sizeof(clienthost), clientservice, 
			sizeof(clientservice), 
			NI_NUMERICHOST|NI_NUMERICSERV);
		getnameinfo((sockaddr *)&sid_channel->sa_from, 
			sid_channel->sa_len, sidhost, 
			sizeof(clienthost), sidservice, 
			sizeof(clientservice), 
			NI_NUMERICHOST|NI_NUMERICSERV);
		printf("chk %s:%s <> %s:%s len:%d\n", clienthost, clientservice, 
			sidhost, sidservice, msg_channel->sa_msg_len);
	}*/
	if (1||msg_channel->ctype == TA_CHANNEL_UDP_MCAST)
		return _True;
	if (msg_channel->sa_msg_len == 0)
		return _True;	// TCP - addr is inside connect
	if (msg_channel->sa_msg.sin_addr.s_addr 
	== sid_channel->sa_from.sin_addr.s_addr)
		return _True;
	return _False;
	//return _True;	// allways ok, skips nothing
}

/* ------------------------------------------------------- */
const char * taReadableChannelErrno(int ierrno, TaChannel *channel)
{
	if (!channel || channel->ctype < 0 
	|| channel->ctype >= TA_N_PROTOCOLS)
		return NULL;
	if (g_protocols[channel->ctype].readable_errno)
		return (*g_protocols[channel->ctype].readable_errno)(ierrno);
	return NULL;
}

/* ------------------------------------------------------- */
const char *taPrintErrno(int errnum)
{
    return strerror(errnum);
}

/* ------------------------------------------------------- */
void taClearChannelRc(TaArea *area, int sid, int cid)
{
	if (sid < 0 || sid >= area->n_stations || cid < 0 || cid >= TA_N_CHANNELS)
		return;
	area->stations[sid]->channels[cid].complete_send_rc =
	area->stations[sid]->channels[cid].complete_recv_rc = 0;
}
