/*
 *                       ***** TA Channel *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>

#define isalnum_(c) (isalnum(c)||c=='_')

typedef struct {
	char name[64];
	int ta_type;
	int ta_size;
	int ta_packed_size;
	const TaRTTI *rtti;
} TypeNames;
#define TA_MAX_TYPE_NAMES	512
static TypeNames g_type_names[TA_MAX_TYPE_NAMES] = {
	{"", TaNo_TYPE, 0, 0},
	// ta basic types
	{"TaBool", TaBool_TYPE, 1, 1},
	{"TaChar", TaChar_TYPE, 1, 1},
	{"TaOctet", TaOctet_TYPE, 1, 1},
	{"TaWChar", TaWChar_TYPE, 2, 2},
	{"TaShort", TaShort_TYPE, 2, 2},
	{"TaLong", TaLong_TYPE, 4, 4},
	{"TaLongLong", TaLongLong_TYPE, 8, 8},
	{"TaFloat", TaFloat_TYPE, 4, 4},
	{"TaDouble", TaDouble_TYPE, 8, 8},
	{"TaLongDouble", TaLongDouble_TYPE, 16, 16},
	{"ARRAY", TaArr_TYPE, 0, 0},
	{"RECORD", TaRec_TYPE, 0, 0},
	{"POINTER", TaPtr_TYPE, 0, 0},
	//
	{"TaUShort", TaShort_TYPE, 2, 2},
	{"TaULong", TaLong_TYPE, 4, 4},
	{"TaULongLong", TaLongLong_TYPE, 8, 8},
	// basic oberon types
	/*
	{"BOOLEAN", TaBool_TYPE, 1, 1},
	{"SHORTCHAR", TaChar_TYPE, 1, 1},
	{"BYTE", TaOctet_TYPE, 1, 1},
	{"CHAR", TaWChar_TYPE, 2, 2},
	{"SHORT", TaShort_TYPE, 2, 2},
	{"INTEGER", TaLong_TYPE, 4, 4},
	{"LONGINT", TaLongLong_TYPE, 8, 8},
	{"SHORTREAL", TaFloat_TYPE, 4, 4},
	{"REAL", TaDouble_TYPE, 8, 8},
	{"SHORTINT", TaShort_TYPE, 2, 2},
	*/
	// idl types
	{"boolean", TaBool_TYPE, 1, 1},
	{"char", TaChar_TYPE, 1, 1},
	{"octet", TaOctet_TYPE, 1, 1},
	{"wchar", TaWChar_TYPE, 2, 2},
	{"short", TaShort_TYPE, 2, 2},
	// int, long are multi-size
	{"longlong", TaLongLong_TYPE, 8, 8},
	{"float", TaFloat_TYPE, 4, 4},
	{"double", TaDouble_TYPE, 8, 8}
};

static TaRTTI g_rtti_TaErrorData;
static const char *g_TaErrorData = "typedef struct { int errcode; } TaErrorData;";

static char *ParseCouple(char *p, char *buf_type, char *buf_name, 
	char *buf_dim);
static void AddItemCrc(TaRTTItem *item, TaLong *names_crc, TaShort *types_crc);
static TaRTTItem *AddCouple(TaRTTI **ritem, char *buf_type, char *buf_name, 
	char *buf_dim, TaRTTI *rtti, TaRTTItem *rttitem, TaBool add_crc, 
	int kdim);
static int FindAlias(const char *buf_type, const char *iface_name);
static int AppendAlias(const TaRTTI *rtti);
static void DoRead(TaTable *t, TaBool *changed);
static void ReadItem(TaRTTItem *ritem, const char *item_addr, char *recent_data, 
	int *col, int *offset_p, int basic_length, int buffer_length, TaBool *changed,
	TaUShort realloc_mode);
TaBool BoolVal(const char *addr);
void PutBoolVal(char *addr, TaBool b);
TaOctet I1Val(const char *addr);
void PutI1Val(char *addr, TaOctet i1);
TaChar CharVal(const char *addr);
void PutCharVal(char *addr, TaChar c);
TaShort I2Val(const char *addr);
void PutI2Val(char *addr, TaShort i2);
TaWChar WCharVal(const char *addr);
void PutWCharVal(char *addr, TaWChar wc);
TaLong I4Val(const char *addr);
void PutI4Val(char *addr, TaLong i4);
TaLongLong I8Val(const char *addr);
void PutI8Val(char *addr, TaLongLong i8);
TaFloat R4Val(const char *addr);
void PutR4Val(char *addr, TaFloat r4);
TaDouble R8Val(const char *addr);
void PutR8Val(char *addr, TaDouble r8);
#define readOp	-10
#define clearOp -11
#define tcallOp -18
#define writeOp -19
#define waitOp -20
static TaBool Synch(TaTable *t, int op);
static void SetWaitingExpiration(TaElem *e, TaLongLong *t_waiting_expired);
static void Put(TaTable *t, int op, TaRTTItem *item, TaTableCommand *tcmd,
	void *par, TaShort *write_sids, int n_write_sids, TaShort waiting, 
	TaLongLong t_waiting_expired);
#define TA_MAX_OPERATIONS	20
static void Process(TaTable *t, TaBool got_message, TaLongLong *waiting);
static int DoSend(TaTable *t, TaShort *write_sids, int n_write_sids);
static int BeginWrite(const TaRTTItem *item, const void *data, TaTable *t, 
	TaShort *write_sids, int n_write_sids);
static void WriteItem(const TaRTTItem *item, TaTable *t, const char *data, 
	char *buffer, int buffer_size, int *col, int *offset_p, int dyn_dim);
static int ItemPackedSize(const char *addr, const TaRTTItem *item, TaBool dyn);

/* ------------------------------------------------------- */
void taRTTI_Init()
{
	taParseToRTTI(g_TaErrorData, "Ta.ErrorData", &g_rtti_TaErrorData, 0);
}

/* ------------------------------------------------------- */
char *ParseCouple(char *p, char *buf_type, char *buf_name, char *buf_dim)
{
	// advance to first TaLongLong t_changed
	while (!isalnum_(*p)) {
		if (*p == '}')
			return p;
		if (*p == 0)
			return NULL;
		++p;
	}
	// get TaLongLong
	while (isalnum_(*p))
		*buf_type++ = *p++;
	*buf_type = 0;
	while (!isalnum_(*p)) {
		if (*p == '*') {
			*buf_name++ = *p++;
			break;
		}
		if (*p != ' ' && *p != '\t')
			return NULL;
		++p;
	}
	// get t_changed
	while (isalnum_(*p))
		*buf_name++ = *p++;
	*buf_name = 0;
	// get dim=12 from arr[12]
	if (*p == '[') {
		++p;
		while (isalnum_(*p))
			*buf_dim++ = *p++;
		if (*p == ']')
			++p;
		else
			return NULL;
	}
	*buf_dim = 0;

	while (*p) {
		if (*p == ' ' || *p == '\t')
			++p;
		if (*p == ';' || *p == '}')
			return p;
		break;
	}
	return NULL;
}

/* ------------------------------------------------------- */
int FindAlias(const char *name, const char *iface_name)
{
	int i;
	for (i = 1; (i < TA_MAX_TYPE_NAMES) && g_type_names[i].ta_type; i++)
		if (!strcmp(name, g_type_names[i].name)) {
			if (!g_type_names[i].rtti)
				return i;
			if (!strcmp(iface_name, 
			g_type_names[i].rtti->iface_name))
				return i;
		}
	return -1;
}

/* ------------------------------------------------------- */
int AppendAlias(const TaRTTI *rtti)
{
	int i;
	for (i = 1; i < TA_MAX_TYPE_NAMES; i++)
		if (!g_type_names[i].ta_type) {
			g_type_names[i].ta_type = TaRec_TYPE;
			g_type_names[i].rtti = rtti;
			strncpy(g_type_names[i].name, rtti->mainitem.type_name,
				64);
			g_type_names[i].name[63] = 0;
			g_type_names[i].ta_size = rtti->sizeof_struct;
			g_type_names[i].ta_packed_size = rtti->sizeof_packed;
			return i;
		}
	return -1;
}

/* ------------------------------------------------------- */
void AddItemCrc(TaRTTItem *item, TaLong *names_crc, TaShort *types_crc)
{
	TaRTTItem *item1;
	const char *type_name;
	TaLong old_names_crc;
	TaShort old_types_crc;
//printf("AddItemCrc %s tp:%d\n", item->item_name, item->item_type_id);
	switch (item->item_type_id) {
	case TaArr_TYPE:
		AddItemCrc(item->contains, names_crc, types_crc);
		return;
	case TaRec_TYPE:
		item1 = item->contains;
		while (item1) {
			AddItemCrc(item1, names_crc, types_crc);
			item1 = item1->next;
		}
		type_name = item->type_name;
		break;
	case TaPtr_TYPE:
		//AddItemCrc(item->contains, names_crc, types_crc);
		type_name = "POINTER";
		break;
	default:
		type_name = g_type_names[item->item_type_id].name;
	}
	old_names_crc = *names_crc;
	old_types_crc = *types_crc;
	*names_crc = (TaLong)taCrc32(item->item_name,
		(TaULong)*names_crc, 32);
	*types_crc = (TaShort)taCrc32(type_name, (TaULong)*types_crc, 16);
	/*if (ta_trace_level & TA_TRAL_MESSAGE)
		fprintf(ta_file, "\t'%s':%d '%s':%d old:%d %d o:%d tp:%d %x\n", 
		type_name, *types_crc, item->item_name, *names_crc, 
		old_types_crc, old_names_crc,
		item->offset, item->item_type_id, item);*/
}

/* ------------------------------------------------------- */
TaRTTItem *AddCouple(TaRTTI **ritem, char *buf_type, char *buf_name, 
	char *buf_dim, TaRTTI *rtti, TaRTTItem *rttitem, TaBool add_crc,
	int kdim)
{
	int i, align_size, dim = 1;
	TaRTTItem *item;
	const char *type_name;
	TaBool used_packed = _False;
	TaLong old_names_crc;
	TaShort old_types_crc;

	// incapsulated structures?
//printf("AddCouple(%s,%s)\n", buf_type, buf_name);
	i = FindAlias(buf_type, rtti->iface_name);
	if (i < 0)
		return NULL;
	*ritem = (TaRTTItem *) malloc(sizeof(TaRTTItem));
	if (!*ritem)
		return NULL;
	item = *ritem;
	memset(item, 0, sizeof(TaRTTItem));
	strncpy(item->type_name, buf_type, TA_NAME_LENGTH);
	item->type_name[TA_NAME_LENGTH-1] = 0;
	strncpy(item->item_name, buf_name, TA_NAME_LENGTH);
	item->item_name[TA_NAME_LENGTH-1] = 0;
	item->item_type_id = g_type_names[i].ta_type;
	item->item_type_size = g_type_names[i].ta_size;
	align_size = item->item_type_size;
	type_name = g_type_names[g_type_names[i].ta_type].name;
	if (!buf_dim)
		item->dim = 0;
	else {
		item->dim = atoi(buf_dim);
		if (item->dim > 1)
			dim = item->dim;
	}
	item->size = g_type_names[i].ta_size * dim;
	item->size_packed = g_type_names[i].ta_packed_size * dim;
	if (*buf_name == '*') {
		// POINTER
		TaRTTItem dummy;
		memset(&dummy, 0, sizeof(TaRTTItem));
		strncpy(item->item_name, buf_name+1, TA_NAME_LENGTH);
		item->item_name[TA_NAME_LENGTH-1] = 0;
		item->item_type_id = TaPtr_TYPE;
		item->item_type_size = item->size = sizeof(void *);
		align_size = item->item_type_size;
		item->size_packed = 16;
		i = TaPtr_TYPE;
		type_name = g_type_names[TaPtr_TYPE].name;
		// dim = 1 for array or record
		if (!AddCouple(&item->contains, buf_type, buf_name+1,
		"1", rtti, &dummy, _False, 1))
			return NULL;
		AddItemCrc(item->contains, &item->names_crc, &item->types_crc);
//printf("ptr%x %s nx:%x :%d '%s':%d o:%d\n", item, item->item_name, item->next,
//		item->types_crc, item->contains->item_name, item->names_crc,
//		item->offset);
	} else if (item->dim > 0) {
		// ARRAY contains
		if (!AddCouple(&item->contains, buf_type, buf_name,
		NULL, rtti, rttitem, add_crc, item->dim))
			return NULL;
		add_crc = _False;
		item->item_type_id = TaArr_TYPE;
		item->item_type_size = item->contains->item_type_size;
		item->size = item->contains->item_type_size
			* item->dim;
		item->size_packed = item->contains->size_packed * dim;
	} else if (item->item_type_id == TaRec_TYPE) {
		// RECORD
		TaRTTItem *recitem, *insitem, *previtem;
		char loc_dim[256], loc_name[256];
		int prevsize;

		recitem = g_type_names[i].rtti->mainitem.next;
		if (!recitem)
			return item;

		// go inside the record
		item->contains = NULL;
		previtem = item;
		item->offset = rttitem->offset + rttitem->size;
		item->size = 0;
		align_size = g_type_names[i].rtti->max_alignment;
		if ((item->offset % align_size) != 0)
			item->offset = 
				(item->offset/align_size+1)*align_size;
		type_name = item->type_name;
		prevsize = rtti->sizeof_packed;
		while (recitem) {
			sprintf(loc_dim, "%d", recitem->dim);
			if (recitem->item_type_id == TaPtr_TYPE) {
				sprintf(loc_name, "*%s", recitem->item_name);
				if (!AddCouple(&insitem, recitem->type_name, 
				loc_name, loc_dim, rtti, previtem, 
				add_crc, 1))
					return NULL;
			} else
				if (!AddCouple(&insitem, recitem->type_name, 
				recitem->item_name, loc_dim, rtti, previtem, 
				add_crc, 1))
					return NULL;
			if (!item->contains)
				item->contains = insitem;
			else
				previtem->next = insitem;
			previtem = insitem;
			recitem = recitem->next;
		}
		item->size = g_type_names[i].rtti->sizeof_struct;
		item->size_packed = g_type_names[i].rtti->sizeof_packed;
		used_packed = _True;
		if (kdim > 1)
			rtti->sizeof_packed = prevsize + kdim 
				* (rtti->sizeof_packed - prevsize);
	}
//printf("%x %s (%s) tp:%d dim:%d size_p:%d add_crc:%d\n", 
//item, item->item_name, rttitem->item_name, item->item_type_id, 
//item->dim, item->size_packed, add_crc);

	if (align_size > rtti->pack_alignment)
		align_size = rtti->pack_alignment;
	item->offset = rttitem->offset + rttitem->size;
	if ((item->offset % align_size) != 0)
		item->offset = 
			(item->offset/align_size+1)*align_size;

	if (!add_crc)
		return item;

	++rtti->columns;
	item->number = rttitem->number + 1;
	if (item->item_type_size > rtti->max_alignment)
		rtti->max_alignment = item->item_type_size;
	if (rtti->max_alignment > rtti->pack_alignment)
		rtti->max_alignment = rtti->pack_alignment;
	if (!used_packed)
		rtti->sizeof_packed += item->size_packed*kdim;
	rtti->sizeof_struct = item->offset+item->size*kdim;
	/*if ((rtti->sizeof_struct % rtti->max_alignment) != 0)
		rtti->sizeof_struct =
			(rtti->sizeof_struct/rtti->max_alignment+1)
				* rtti->max_alignment;
	*/

	old_names_crc = rtti->names_crc;
	old_types_crc = rtti->types_crc;
	rtti->names_crc = (TaLong)taCrc32(item->item_name,
		(TaULong)rtti->names_crc, 32);
	rtti->types_crc = (TaShort)taCrc32(
		type_name, (TaULong)rtti->types_crc, 16);
	if (ta_trace_level & TA_TRAL_MESSAGE)
		fprintf(ta_file, 
		"'%s'/'%s':%d '%s':%d old:%d %d o:%d s:%d/%d p:%d/%d i:%d\n", 
		g_type_names[g_type_names[i].ta_type].name, buf_type,
		rtti->types_crc, item->item_name, rtti->names_crc,
		old_types_crc, old_names_crc,
		item->offset, rtti->sizeof_struct, item->size,
		rtti->sizeof_packed, item->size_packed,
		item->item_type_id);
	return item;
}

/* ------------------------------------------------------- */
int taParseToRTTI(const char *struct_descr, const char *iface_name, 
	TaRTTI *rtti, int pack_alignment)
{
	char *p, *p1, *pi, buf_type[256], buf_name[256], buf_dim[256];
	int i, len;
	TaRTTItem *ritem;
	TaLong old_names_crc;
	TaShort old_types_crc;

	if (!struct_descr || !*struct_descr || !rtti)
		return TA_ERR_ARGUMENT;
	memset(rtti, 0, sizeof(TaRTTI));
	// find interface name
	if (iface_name) {
		strncpy(rtti->iface_name, iface_name, TA_NAME_LENGTH);
		p = struct_descr;
	} else {
		p = strstr(struct_descr, "interface");
		if (!p)
			return TA_PARSING;
		p += strlen("interface") + 1;
		while (*p && !isalnum_(*p)) ++p;
		if (!isalnum_(*p))
			return TA_PARSING;
		for (i = 0; i < TA_NAME_LENGTH-1; i++) {
			if (isalnum_(p[i]))
				rtti->iface_name[i] = p[i];
			else {
				rtti->iface_name[i] = 0;
				break;
			}
		}
	}
	rtti->iface_name[TA_NAME_LENGTH-1] = 0;
	// find struct type name
	p = strstr(p, "struct");
	if (!p)
		return TA_PARSING;
	p += strlen("struct") + 1;
	p1 = strchr(p, '{');
	if (!p1)
		return TA_PARSING;
	// parse struct types
	++p1;
	ritem = &rtti->mainitem;
	ritem->offset = 0;
	ritem->item_type_size = 0;
	rtti->max_alignment = 1;
	rtti->columns = 0;
	if (pack_alignment > 0)
		rtti->pack_alignment = pack_alignment;
	else
		rtti->pack_alignment = TA_PACK_ALIGN;
	while (1) {
		p1 = ParseCouple(p1, buf_type, buf_name, buf_dim);
		if (!p1)
			return TA_PARSING;
		if (*p1 == '}')
			break;
		ritem = AddCouple(&ritem->next, buf_type, buf_name, buf_dim, 
			rtti, ritem, _True, 1);
		if (!ritem)
			return TA_PARSING;
	}
	// end of parse struct types
	p = strchr(p1, '}');
	if (!p)
		return TA_PARSING;
	while (*p && !isalnum_(*p)) ++p;
	if (!isalnum_(*p))
		return TA_PARSING;
	for (i = 0; i < TA_NAME_LENGTH-1; i++) {
		if (isalnum_(p[i]))
			rtti->mainitem.type_name[i] = p[i];
		else {
			rtti->mainitem.type_name[i] = 0;
			break;
		}
	}
	old_names_crc = rtti->names_crc;
	old_types_crc = rtti->types_crc;
	rtti->names_crc = (TaLong)taCrc32(rtti->iface_name,
		(TaULong)rtti->names_crc, 32);
	rtti->types_crc = (TaShort)taCrc32(rtti->mainitem.type_name,
		(TaULong)rtti->types_crc, 16);
	if (ta_trace_level & TA_TRAL_MESSAGE)
	fprintf(ta_file, "'%s':%d '%s':%d old: %d %d size_p:%d\n", 
		rtti->iface_name, rtti->names_crc,
		rtti->mainitem.type_name, rtti->types_crc, 
		old_types_crc, old_names_crc, rtti->sizeof_packed);
	rtti->mainitem.item_type_id = TaRec_TYPE;
	rtti->mainitem.size = rtti->sizeof_struct;
	AppendAlias(rtti);

//ritem = &rtti->mainitem;
//while(ritem) {
//printf("==%s %x %s %s tp:%d contains:%x %s next:%x %s\n", 
//rtti->mainitem.type_name, ritem, ritem->type_name, 
//ritem->item_name, ritem->item_type_id,
//ritem->contains, (ritem->contains ? ritem->contains->item_name : ""),
//ritem->next, (ritem->next ? ritem->next->item_name : ""));
//ritem = ritem->next;
//}
	return 0;
}

/* ------------------------------------------------------- */
void taClearRTTI(TaRTTI *rtti)
{
	// to be continued
}

/* ------------------------------------------------------- */
void taNewTable(TaArea *area, TaTable *table)
{
	if (!area || !table)
		return;
	memset(table, 0, sizeof(TaTable));
	// NEW(t.prealloced[i])
	table->area = area;
	if (!area->tablist)
		area->tablist = table;
	else {
		TaTable *t = area->tablist;
		while (t->next) t = t->next;
		t->next = table;
	}
	taDefaultMessageQos(&table->mqos);
}

/* ------------------------------------------------------- */
int taAttachTable(TaTable *table, const TaRTTI *rtti, TaUShort dst_oid, 
	TaUShort src_oid, TaUShort match_mode, TaULong keep_history, TaOnDataHandler *on_data,
	TaUShort realloc_mode)
{
	int i;
	if (!table || keep_history > TA_MAX_KEEP_HISTORY)
		return TA_ERR_ARGUMENT;
	table->rows = table->columns = table->res = 0;
	table->rtti = NULL;
	table->changed = _False;
	if (table->recent_data)
		free(table->recent_data);
	table->recent_data = NULL;
	for (i = 0; i < TA_MAX_KEEP_HISTORY; i++) {
		if (table->history_data[i])
			free(table->history_data[i]);
		table->history_data[i] = NULL;
	}
	if (!rtti || keep_history == 0)
		return 0;	// detach

	table->rtti = rtti;
	table->columns = rtti->columns;
	table->dst_oid = dst_oid;
	table->src_oid = src_oid;
	table->match_mode = match_mode;
	table->realloc_mode = realloc_mode;
	switch (match_mode) {
	case TA_MSG_SOAP_MATCH_INTERFACE:
		table->msg_type = (TaULong) taCrc32(rtti->iface_name,
			0, 32);
		table->iface_code = (TaUShort) taCrc32(
			rtti->mainitem.type_name, 0, 16);
		table->buffer_size = table->area->qos.soap_buffer_size;
		break;
	case TA_SOAP_GET_CLIENT:
	case TA_SOAP_POST_CLIENT:
		table->msg_type = TA_SOAP_MID;
		table->iface_code = TA_SOAP_POST_CLIENT;
		table->buffer_size = table->area->qos.soap_buffer_size;
		break;
	case TA_SOAP_11_CLIENT:
	case TA_SOAP_12_CLIENT:
		table->msg_type = TA_SOAP_MID;
		table->iface_code = (TaUShort) taCrc32(
			rtti->mainitem.type_name, 0, 16);
		table->buffer_size = table->area->qos.soap_buffer_size;
		break;
	default:
		table->msg_type = (TaULong) rtti->names_crc;
		table->iface_code = (TaUShort) rtti->types_crc;
		table->buffer_size = rtti->sizeof_packed;	// sizeof_packed
		break;
	}
	table->keep_history = keep_history;
	table->on_data = on_data;
	table->sendto_sids[0] = TA_SIDS_ALL;
	table->n_sendto_sids = TA_N_SIDS_DEFAULT;

	table->recent_data = malloc(rtti->sizeof_struct);
	memset(table->recent_data, 0, rtti->sizeof_struct);
	for (i = 0; i < keep_history; i++) {
		table->history_data[i] = malloc(rtti->sizeof_struct);
		memset(table->history_data[i], 0, rtti->sizeof_struct);
	}

	table->buffer = (char *) malloc(table->buffer_size);
	memset(table->buffer, 0, table->buffer_size);
	table->dyn_buffer = NULL;
	table->dyn_buffer_size = 0;
	return 1;
}

/* ------------------------------------------------------- */
int taSendtoSids(TaTable *table, TaShort *sendto_sids, int n_sendto_sids)
{
	int i;
	if (sendto_sids == NULL || sendto_sids[0] == TA_SIDS_ALL) {
		table->sendto_sids[0] = TA_SIDS_ALL;
		table->n_sendto_sids = (n_sendto_sids > 0 ? 1 : -1);
	} else {
		table->n_sendto_sids = n_sendto_sids;
		if (n_sendto_sids < 0)
			n_sendto_sids = -n_sendto_sids;
		for (i = 0; i < n_sendto_sids; i++)
			table->sendto_sids[i] = sendto_sids[i];
	}
}

/* ------------------------------------------------------- */
int taUpdateAreaTables(TaArea *area)
{
	TaMessage message;
	int msg_ind, rc, j;
	TaTable *t;
	TaBool found, match;
	TaErrorData edp;
	TaLongLong t_waiting_expired;
	msg_ind = 0;
	area->n_msgs_not_found = 0;
	while ((rc = taGetMessage(area, msg_ind++, &message)) >= 0) {
		t = area->tablist;
		found = _False;
		while (!found && t != NULL) {
			edp.errcode = 0;
			switch (t->match_mode) {
			case TA_MSG_SRPS_ACCEPT_ANY:
				match = (message.basic_length - sizeof(TaSRPSHeader)
				== t->rtti->sizeof_packed &&
				message.msg_type == t->msg_type &&
                        	message.iface_code == t->iface_code);
				break;
			case TA_MSG_SRPS_MATCH_DST:
				match = (message.dst_oid == t->dst_oid &&
				message.basic_length - sizeof(TaSRPSHeader) 
				== t->rtti->sizeof_packed &&
				message.msg_type == t->msg_type &&
				message.iface_code == t->iface_code);
				break;
			case TA_MSG_SRPS_MATCH_SID:
				match = (message.src_sid == t->dst_oid &&
				message.basic_length - sizeof(TaSRPSHeader)
				== t->rtti->sizeof_packed &&
				message.msg_type == t->msg_type &&
				message.iface_code == t->iface_code);
				break;
			case TA_MSG_SOAP_MATCH_INTERFACE:
				match = (message.msg_type == t->msg_type &&
				message.iface_code == t->iface_code);
				break;
			case TA_SOAP_11_CLIENT:
			case TA_SOAP_12_CLIENT:
			case TA_SOAP_GET_CLIENT:
			case TA_SOAP_POST_CLIENT:
				match = (message.src_sid == t->dst_oid &&
				message.msg_type == t->msg_type &&
				message.msg_type == (TaULong)t->src_mid);
				break;
			case TA_SOAP_MID:
				match = (message.src_sid == t->dst_oid &&
					SOAP_MSG(message.msg_type));
				break;
			default:
				match = _False;
			}
			if (ta_trace_level & TA_TRAL_MESSAGE)
				printf("msg_match:%d mid:%d dst:%d msg_type:%d#%d "
				"dst_oid:%d#%d iface_code:%d#%d l:%d#%d\n", 
				match, message.mid, message.dst_sid, message.msg_type, 
				t->msg_type, message.dst_oid, t->dst_oid, 
				message.iface_code, t->iface_code, 
				message.basic_length - sizeof(TaSRPSHeader),
				t->rtti->sizeof_packed);
			if (match) {
				found = _True;
				t->changed = _False;
				t->message = message;
				DoRead(t, &t->changed);
				if (t->res == TA_TO_BE_CONTINUED)
					continue;
				if (t->changed) {
					if (t->rows < t->keep_history)
						++t->rows;
					for (j = t->keep_history-1; j >= 1; j--)
						memcpy(t->history_data[j],
							t->history_data[j-1],
							t->rtti->sizeof_struct);
					memcpy(t->history_data[0],
						t->recent_data,
						t->rtti->sizeof_struct);
				}
				if (t->res < 0)
					edp.errcode = t->res;
				else
					edp.errcode = 0;
				Process(t, _True, &t_waiting_expired);
				if (t->on_data) {
					if (edp.errcode)
						(*t->on_data)(&edp,  (TaULong)
							g_rtti_TaErrorData.names_crc, t);
					else
						(*t->on_data)(t->recent_data, t->msg_type, t);
				}
			}
			t = t->next;
		}
		if (!found)
			++area->n_msgs_not_found;
//		if (!found)
//		printf("*** %s src:%d msg_type:%d iface_code%d src_sid:%d dst_sid:%d\n", 
//		area->host_station->name, area->host_station->sid, message.msg_type,
//		message.iface_code, message.src_sid, message.dst_sid);
	}

	if (area->t_waiting_expired > 0) {
		t = area->tablist;
		t_waiting_expired = 0;
		while (t != NULL) {
			Process(t, _False, &t_waiting_expired);
			t = t->next;
		}
		area->t_waiting_expired = t_waiting_expired;
	}
	return 0;
}

/* ------------------------------------------------------- */
TaBool BoolVal(const char *addr)
{
	return *(const TaBool *) addr;
}
void PutBoolVal(char *addr, TaBool b)
{
	*((TaBool *)addr) = b;
}
TaOctet I1Val(const char *addr)
{
	return *(const TaOctet *) addr;
}
void PutI1Val(char *addr, TaOctet i1)
{
	*((TaOctet *)addr) = i1;
}
TaChar CharVal(const char *addr)
{
	return *(const TaChar *) addr;
}
void PutCharVal(char *addr, TaChar c)
{
	*((TaChar *)addr) = c;
}
TaShort I2Val(const char *addr)
{
	return *(const TaShort *) addr;
}
void PutI2Val(char *addr, TaShort i2)
{
	*((TaShort *)addr) = i2;
}
TaWChar WCharVal(const char *addr)
{
	return *(const TaWChar *) addr;
}
void PutWCharVal(char *addr, TaWChar wc)
{
	*((TaWChar *)addr) = wc;
}
TaLong I4Val(const char *addr)
{
	return *(const TaLong *) addr;
}
void PutI4Val(char *addr, TaLong i4)
{
	*((TaLong *)addr) = i4;
}
TaLongLong I8Val(const char *addr)
{
	return *(const TaLongLong *) addr;
}
void PutI8Val(char *addr, TaLongLong i8)
{
	*((TaLongLong *)addr) = i8;
}
TaFloat R4Val(const char *addr)
{
	return *(const TaFloat *) addr;
}
void PutR4Val(char *addr, TaFloat r4)
{
	*((TaFloat *)addr) = r4;
}
TaDouble R8Val(const char *addr)
{
	return *(const TaDouble *) addr;
}
void PutR8Val(char *addr, TaDouble r8)
{
	*((TaDouble *)addr) = r8;
}

/* ------------------------------------------------------- */
void ReadItem(TaRTTItem *ritem, const char *item_addr, char *recent_data,
	int *col, int *offset_p, int basic_length, int buffer_length, TaBool *changed,
	TaUShort realloc_mode)
{
	TaOctet i1;
	TaBool b;
	TaChar c;
	TaShort i2;
	TaWChar wc;
	TaLong i4;
	TaLongLong i8;
	TaFloat r4;
	TaDouble r8;
	TaRTTItem *item1;
	TaLong nb;

	switch (ritem->item_type_id) {
	case TaBool_TYPE:
		memcpy(&i1, item_addr+*offset_p, 1);
		b = ((i1 == 0) ? _False : _True);
		if (b != BoolVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutBoolVal(recent_data+ritem->offset, b);
		}
		++*offset_p;
		++*col;
		break;
	case TaOctet_TYPE:
		memcpy(&i1, item_addr+*offset_p, 1);
		if (i1 != I1Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI1Val(recent_data+ritem->offset, i1);
		}
		++*offset_p;
		++*col;
		break;
	case TaChar_TYPE:
		memcpy(&c, item_addr+*offset_p, 1);
		if (c != CharVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutCharVal(recent_data+ritem->offset, c);
		}
		++*offset_p;
		++*col;
		break;
	case TaShort_TYPE:
		memcpy(&i2, item_addr+*offset_p, 2);
		i2 = taN2H16(i2);
		if (i2 != I2Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI2Val(recent_data+ritem->offset, i2);
		}
		*offset_p += 2;
		++*col;
		break;
	case TaWChar_TYPE:
		memcpy(&i2, item_addr+*offset_p, 2);
		wc = (TaWChar) taN2H16(i2);
		if (wc != WCharVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutWCharVal(recent_data+ritem->offset, wc);
		}
		*offset_p += 2;
		++*col;
		break;
	case TaLong_TYPE:
		memcpy(&i4, item_addr+*offset_p, 4);
		i4 = taN2H32(i4);
		if (i4 != I4Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI4Val(recent_data+ritem->offset, i4);
		}
		*offset_p += 4;
		++*col;
		break;
	case TaLongLong_TYPE:
		memcpy(&i8, item_addr+*offset_p, 8);
		i8 = taN2H64(i8);
		if (i8 != I8Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI8Val(recent_data+ritem->offset, i8);
		}
		*offset_p += 8;
		++*col;
		break;
	case TaFloat_TYPE:
		memcpy(&r4, item_addr+*offset_p, 4);
		r4 = taN2Hr32(r4);
		if (r4 != R4Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutR4Val(recent_data+ritem->offset, r4);
		}
		*offset_p += 4;
		++*col;
		break;
	case TaDouble_TYPE:
		memcpy(&r8, item_addr+*offset_p, 8);
		r8 = taN2Hr64(r8);
		if (r8 != R8Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutR8Val(recent_data+ritem->offset, r8);
		}
		*offset_p += 8;
		++*col;
		break;
	case TaArr_TYPE:
		for (i4 = 0; i4 < ritem->dim; i4++) {
			int oldc = *col;
			ReadItem(ritem->contains, item_addr, 
			recent_data + i4 * ritem->contains->size,
				col, offset_p, basic_length, buffer_length, 
				changed, realloc_mode);
			if (i4 > 0)
				*col = oldc;
		}
		break;
	case TaRec_TYPE:
		if (ritem->contains)
			item1 = ritem->contains;
		else
			item1 = ritem->next;
		while (item1) {
			ReadItem(item1, item_addr, recent_data, 
				col, offset_p, basic_length, buffer_length, 
				changed, realloc_mode);
			item1 = item1->next;
		}
		if (ritem->contains)
			++*col;
		break;
	case TaPtr_TYPE:
		// to be continued
		memcpy(&nb, item_addr+*offset_p, 4);
		memcpy(&i4, item_addr+*offset_p+8, 4);
		memcpy(&i2, item_addr+*offset_p+12, 2);
		nb = taN2H32(nb);
		i4 = taN2H32(i4);	// names_crc
		i2 = taN2H16(i2);	// types_crc
		/*printf("=%d %s %x names:%d<>%d types:%d<>%d nb:%d\n",
			*offset_p, ritem->item_name, ritem, i4, 
			ritem->names_crc, i2, ritem->types_crc, nb);*/
		if (i4 != ritem->names_crc || i2 != ritem->types_crc) {
//int offs;
//short dim;
			if (ta_trace_level & TA_TRAL_ERRORS)
			printf("* %s %x names:%d<>%d types:%d<>%d\n",
				ritem->item_name, ritem, i4, ritem->names_crc,
				i2, ritem->types_crc);
//memcpy(&offs, item_addr+*offset_p+4, 4);
//memcpy(&dim, item_addr+*offset_p+14, 2);
//printf("nb:%d offs:%d dim:%d ~%d\n",nb,offs,dim,*offset_p);
			*col = -1;	/* incompatible */
		} else {
			TaLong dim, i;
			char **addr;
			TaRTTIadata *adt;
			int cl = 0, offs = 0;
			/* offset, dim */
			memcpy(&i4, item_addr+*offset_p+4, 4);
			i4 = taN2H32(i4);	// offset
			i4 += basic_length;
			if (i4 > buffer_length) {
				*col = -1;
				return;
			}
			memcpy(&i2, item_addr+*offset_p+14, 2);
			i2 = taN2H16(i2);	// dim

			addr = (const char **) (recent_data+ritem->offset);
			if (!*addr) {
				adt = NULL;
				dim = 0;
			} else {
				adt = (TaRTTIadata *)(*addr-sizeof(TaRTTIadata));
				dim = adt->dim;
			}
			item1 = ritem->contains;
			if (i2 != dim) {
				if (realloc_mode == TA_DONT_REALLOC) {
					*col = TA_ERR_OUT_OF_MEMORY;
					return;
				}
				if (!adt) {
					*addr = (char *) taAllocWithRTTI(
						item1->size, i2);
					adt = (TaRTTIadata *)
						(*addr-sizeof(TaRTTIadata));
				} else
					*addr = (char *) taReallocWithRTTI(
						*addr, i2);
				*changed = _True;
			}
//printf("ReadItem %s %x nb:%d offs:%d dim:%d adt->dim:%d adt->size:%d\n", 
//item1->item_name, *addr, nb, i4-basic_length, i2, (adt ? adt->dim : 0), 
//(adt ? adt->size : 0));
//for (i = 0; i < nb; i++) printf("%d.", item_addr[i4+i]);printf("\n");
			offs = i4;
			for (i = 0; i < (TaLong) i2; i++) {
				ReadItem(item1->contains, item_addr, 
					*addr+i*item1->contains->size, 
					&cl, &offs, basic_length, buffer_length, 
					changed, realloc_mode);
//if (0&&item1->contains->item_type_id==TaChar_TYPE)
//printf("[%d] offs:%d dd:%d ritem_offset:%d %c %s\n", i, i*item1->contains->size
//+item1->contains->offset, i*item1->contains->size, item1->contains->offset,
//*(*addr+i*item1->contains->size+item1->contains->offset), item1->contains->item_name);
			}
			++*col;
		}
		*offset_p += 16;
		break;
	case TaProc_TYPE:
		break;
	default:
		return;
	}
}

/* ------------------------------------------------------- */
void DoRead(TaTable *t, TaBool *changed)
{
	int col = 0, offset_p = 0;
	char *item_addr;
	*changed = _False;
	t->res = 0;
	if (t->match_mode < TA_MSG_SOAP_MATCH_INTERFACE) {
		item_addr = t->message.buffer + sizeof(TaSRPSHeader);
//for(i=t->message.basic_length; i < t->message.buffer_length; i++)
//printf("%d.", t->message.buffer[i]);
//printf("w %d %d %x\n", t->message.basic_length-sizeof(TaSRPSHeader), t->message.buffer_length-t->message.basic_length, item_addr);
		ReadItem(&t->rtti->mainitem, item_addr, t->recent_data, 
			&col, &offset_p, t->message.basic_length-sizeof(TaSRPSHeader), 
			t->message.buffer_length-sizeof(TaSRPSHeader), changed, t->realloc_mode);
		if (t->rtti->sizeof_packed != t->message.basic_length - sizeof(TaSRPSHeader)
		|| col != t->columns) {
//printf("sizeof_packed:%d t->message.basic_length:%d sizeof(TaSRPSHeader):%d col:%d t->columns:%d\n",
//t->rtti->sizeof_packed,t->message.basic_length,sizeof(TaSRPSHeader),col,t->columns);
			t->res = TA_MSG_INCOMPATIBLE;
			*changed = _False;
		}
	} else {
		TaOnParseSoap ops;
		int rc;
		memset(t->recent_data, 0, t->rtti->sizeof_struct);
		ops.message_buffer = t->message.buffer;
		ops.message_mid = t->message.mid;
		ops.type_name = (const char *) t->rtti->mainitem.type_name;
		ops.table_buffer = (char *) t->buffer;
		ops.soap_buffer_size = t->area->qos.soap_buffer_size;
		ops.on_read_soap = taOnReadSoap;
		ops.client_data = t;
		rc =
		taReadSoap(&ops);
		if (rc < 0)
			t->res = rc;
		else if (rc > 0)
			*changed = _True;
	}
}

/* ------------------------------------------------------- */
void DoClear(TaTable *t)
{
	t->rows = 0;
	t->mid_resp = 0;
}

/* ------------------------------------------------------- */
TaBool Synch(TaTable *t, int op)
{
	TaElem *h;
	h = t->queue;
	if (h == t->tail)
		return _True;
	else if (op >= clearOp) {
		while (h != t->tail && h->table != t)
			h = h->next;
		return (TaBool)(h == t->tail);
	}
	return _False;
}

/* ------------------------------------------------------- */
void SetWaitingExpiration(TaElem *e, TaLongLong *t_waiting_expired)
{
	if (e->t_waiting_expired == 0) {
		if (e->op == writeOp) {
			if (e->table->mqos.tp_send_timeout < 0)
				e->t_waiting_expired = -1;	// infinite
			else
				e->t_waiting_expired = e->table->area->t_crtu 
					+ TA_TP(e->table->mqos.tp_send_timeout);
		} else {
			if (e->table->mqos.tp_timeout < 0)
				e->t_waiting_expired = -1;	// infinite
			else
				e->t_waiting_expired = e->table->area->t_crtu 
					+ TA_TP(e->table->mqos.tp_timeout);
		}
	}
	//if (*t_waiting_expired == 0
	//|| e->t_waiting_expired < *t_waiting_expired)
	*t_waiting_expired = e->t_waiting_expired;
}

/* ------------------------------------------------------- */
// without freeing memory
void Put(TaTable *t, int op, TaRTTItem *item, TaTableCommand *tcmd,
	void *par, TaShort *write_sids, int n_write_sids, TaShort waiting, 
	TaLongLong t_waiting_expired)
{
	TaElem *e, *h, *k;
	int j;
	TaBool found;
	found = _False; j = 0;
	while (j < TA_MAX_PREALLOCED && !found) {
		if (t->prealloced[j].used == TA_ELEM_NOT_USED) {
			e = &t->prealloced[j];
			memset(e, 0, sizeof(TaElem));
			t->prealloced[j].used = TA_ELEM_PREALLOCED;
			found = _True;
		}
		++j;
	}
	if (!found) {
		e = malloc(sizeof(TaElem));
		memset(e, 0, sizeof(TaElem));
		e->used = TA_ELEM_MALLOCED;
	}
	e->table = t; e->op = op; e->item = item; e->tcmd = tcmd; e->par = par;
	e->waiting = waiting; e->t_waiting_expired = t_waiting_expired;
	e->n_write_sids = n_write_sids;
	if (write_sids && n_write_sids) {
		if (n_write_sids < 0)
			n_write_sids = -n_write_sids;
		for (j = 0; j < n_write_sids; j++)
			e->write_sids[j] = write_sids[j];
	}
	if (t->queue == NULL)	// start now
		SetWaitingExpiration(e, &t->area->t_waiting_expired);
//printf("put e:%x op:%d waiting:%d\n", e, op, waiting);
	h = t->queue;
	if (h == t->tail) {
		e->next = h; t->queue = e;
	} else if (op >= clearOp) {
		/* clear, read: insert after last op on same table */
		k = NULL;
		while (h != t->tail) {
			if (h->table == t)
				k = h;
			h = h->next;
		}
		if (k == NULL) {
			e->next = t->queue; t->queue = e;
		} else {
			e->next = k->next; k->next = e;
		}
	} else {
		/* append to the end of queue */
		while (h->next != t->tail)
			h = h->next; 
		e->next = h->next; h->next = e;
	}
//printf("Put: t:%x t->queue:%x op:%d\n", t, t->queue, op);
}

/* ------------------------------------------------------- */
void Process(TaTable *t, TaBool got_message, TaLongLong *t_waiting_expired)
{
	TaElem *e;
	int operations;
	TaBool freeze_process, waitOp_done, is_ready, time_expired, broken_chain, time_to_call;
	TaErrorData edp;

	operations = 0;
	freeze_process = _False;
	waitOp_done = _False;
	broken_chain = _False;
	t->got_message = got_message;
//printf("process got:%d queue:%x\n", got_message, t->queue);
	while (t->queue != NULL && operations < TA_MAX_OPERATIONS
	&& !freeze_process) {
		e = t->queue; t->queue = e->next; t->tail = t->queue; 
		++operations;
		if (!broken_chain) {
			// check expiration date
			time_expired = _False;
			time_to_call = _False;
			if (e->waiting != TA_NO_WAIT && e->t_waiting_expired > 0) {
				if (t->area->t_crtu < e->t_waiting_expired)
					/* still waiting */
					freeze_process = _True;
				else {
					e->t_waiting_expired = 0; /* to be reset in expiration */
					edp.auxdata = e->waiting;
					switch (e->waiting) {
					case TA_WAIT_UNTIL_TIMEOUT:
						/* waiting expired - 
						fire response timeout event */
						time_expired = _True;
						break;
					case TA_SUBSCRIBE:
					case TA_REPEAT_FOREACH_RESPONSE:
						edp.errcode = TA_RESPONSE_TIMEOUT;
						if (t->on_data)
							(*t->on_data)(&edp,  (TaULong)
								g_rtti_TaErrorData.names_crc, t);
						break;
					case TA_PUBLISH:
						time_to_call = _True;
						break;
					default:
						/* timeout errors */
						if (e->op == writeOp) {
							edp.errcode = TA_SEND_TIMEOUT;
							broken_chain = _True;
					/*	} else if (t->last_send_sids != 0 &&
						t->send_attempts_counter < t->mqos.n_send_attempts) {
							TaBool is_ready;
							++t->send_attempts_counter;
							taReadyForSendMessage(t->area, e->sids, 
								t->buffer_size+t->dyn_buffer_len, 
								&t->mqos, &is_ready);
							if (is_ready)
								DoSend(t, t->last_send_sids);
							edp.errcode = TA_RESPONSE_RESEND;
							e->t_waiting_expired =
								t->area->t_crtu + 
									t->mqos.tp_timeout;
							freeze_process = _True; */
						} else {
							edp.errcode = TA_RESPONSE_TIMEOUT;
							broken_chain = _True;
						}
						if (t->on_data)
							(*t->on_data)(&edp,  (TaULong)
								g_rtti_TaErrorData.names_crc, t);
						break;
					}
				}
			}
		}
			if (!time_expired && !broken_chain)
		switch (e->op) {
		case clearOp:
			DoClear(e->table);
			t->src_mid = 0;
			t->send_attempts_counter = 0;
			break;
		case tcallOp:
			switch (e->waiting) {
			case TA_PUBLISH:
				if (e->n_write_sids == TA_PERMANENT_MID) {
					if (t->message.mid == t->message.mid_resp)
						t->got_message = _False;	/* ignore stop */
					t->mid_resp = TA_PERMANENT_MID;
					e->current_mid = TA_PERMANENT_MID;
				} else if (t->got_message) {
					if (t->message.mid == t->message.mid_resp) {
						DoClear(e->table);	/* stop publish */
						t->got_message = _False;
						time_to_call = _False;
					} else
						t->mid_resp = t->message.mid;	/* start publish */
					e->current_mid = t->message.mid;
				}
				if (t->mid_resp == 0)
					time_to_call = _False;
				if (t->got_message || time_to_call)
					(*e->tcmd)(e->table, e->par);
				freeze_process = _True;
				break;
			case TA_SUBSCRIBE:
				freeze_process = _True;
				if (t->got_message) {
					if (e->current_mid == 0) e->current_mid = t->src_mid;
					if (e->n_write_sids == TA_PERMANENT_MID) {
						t->src_mid = TA_PERMANENT_MID;
						e->current_mid = TA_PERMANENT_MID;
						(*e->tcmd)(e->table, e->par);
					} else if (t->src_mid != e->current_mid)
						freeze_process = _False; // subscription changed
					else if (t->message.mid_resp != t->src_mid) {
						if (t->on_data) {
							edp.errcode = TA_RESPONSE_MISMATCH;
							edp.auxdata = t->message.mid_resp;
							(*t->on_data)(&edp,  (TaULong)
								g_rtti_TaErrorData.names_crc, t);
						}
					} else
						(*e->tcmd)(e->table, e->par);
					e->t_waiting_expired = 0;
				}
				break;
			case TA_REPEAT_FOREACH_RESPONSE:
				if (t->got_message) {
					t->mid_resp = t->message.mid;	/* fix request mid */
					(*e->tcmd)(e->table, e->par);
					e->t_waiting_expired = 0;
				}
				freeze_process = _True;
				break;
			case TA_REQUEST:
				freeze_process = _True;
				if (t->got_message) {
//printf("current_mid:%d src_mid:%d mid_resp:%d got:%d\n", e->current_mid, t->src_mid, t->message.mid_resp, t->got_message);
					if (e->current_mid == 0) e->current_mid = t->src_mid;
					if (t->src_mid != e->current_mid) {
					} else if (t->message.mid_resp != t->src_mid) {
					} else {
						freeze_process = _False;
						t->src_mid = 0;	/* only once */
						(*e->tcmd)(e->table, e->par);
					}
				}
				break;
			default:
				(*e->tcmd)(e->table, e->par);
				break;
			}
			break;
		case writeOp:
			if (e->waiting == TA_WAIT_WRITE_PERMISSION)
				taReadyForSendMessage(t->area, e->write_sids, 
					e->n_write_sids,
					t->buffer_size + t->dyn_buffer_len,
					&t->mqos, &is_ready);
			else
				is_ready = _True;
			if (is_ready) {
				BeginWrite(&e->table->rtti->mainitem, e->par, 
					e->table, e->write_sids, e->n_write_sids);
				freeze_process = _False;
			} else
				freeze_process = _True;
			break;
		case waitOp:
			if (e->waiting < TA_WAIT_UNTIL_TIMEOUT) {
				if (waitOp_done || !got_message)
					freeze_process = _True;
				else {
					waitOp_done = _True;
					freeze_process = _False;
					if (e->waiting <= TA_WAIT_FOR_RESPONSE) {
						if (t->message.mid_resp != t->src_mid) {
							freeze_process = _True;
							if (t->on_data) {
								edp.errcode = TA_RESPONSE_MISMATCH;
								edp.auxdata = t->message.mid_resp;
								(*t->on_data)(&edp,  (TaULong)
									g_rtti_TaErrorData.names_crc, t);
							}
						}
					}
				}
			}
			break;
		default:
			// ReadHistory
			break;
		}
//printf("e:%x e.op:%d e.waiting:%d freeze:%d timeout:%d broken:%d got:%d\n", 
//e, e->op, e->waiting, freeze_process, time_expired, broken_chain, got_message);
		if (freeze_process)
			t->queue = e;
		if (e != t->queue) {
			if (e->used == TA_ELEM_MALLOCED)
				free(e);
			else
				e->used = TA_ELEM_NOT_USED;
		}
		t->tail = NULL;
	}
	if (t->queue && t->queue->waiting != TA_NO_WAIT)
		SetWaitingExpiration(t->queue, t_waiting_expired);
}

/* ------------------------------------------------------- */
TaULong taErrorDataMsgType()
{
	return (TaULong) g_rtti_TaErrorData.names_crc;
}

/* ------------------------------------------------------- */
TaBool taAvailable(TaTable *t)
{
	return Synch(t, readOp);
}

/* ------------------------------------------------------- */
void taCallTable(TaTable *t, TaTableCommand *command,
	void *par)
{
	if (!command)
		return;
	if (Synch(t, tcallOp))
		(*command)(t, par);
	else
		Put(t, tcallOp, NULL, command, par, NULL, 0, TA_NO_WAIT, 0);
}

/* ------------------------------------------------------- */
void taWriteTable(TaTable *t, void *data, TaULong sids, TaBool wait_write_permission)
{
	TaBool is_ready;
	TaLongLong t_waiting_expired;
	TaShort write_sids[TA_MAX_MASK_ADDRS];
	int i, n_write_sids, n_sendto_sids;
	n_sendto_sids = t->n_sendto_sids;
	if (n_sendto_sids < 0)
		n_sendto_sids = -n_sendto_sids;
	if (sids == TA_ADDR_ALL) {
		if (t->sendto_sids[0] == TA_SIDS_ALL)
			write_sids[0] = TA_SIDS_ALL;
		else
			/* if n_sendto_sids < 0 - multicast else unicast */
			for (i = 0; i < n_sendto_sids; i++)
				write_sids[i] = t->sendto_sids[i];
		n_write_sids = t->n_sendto_sids;
	} else {
		if (t->sendto_sids[0] == TA_SIDS_ALL) {
			for (i = n_write_sids = 0; i < t->area->n_stations
			&& i < TA_MAX_MASK_ADDRS; i++) {
				if (sids & TA_TO_ADDR(i))
					write_sids[n_write_sids++] = (TaShort) i;
			}
		} else {
			for (i = n_write_sids = 0; i < n_sendto_sids; i++) {
				if (sids & TA_TO_ADDR(i))
					write_sids[n_write_sids++] = t->sendto_sids[i];
			}
		}
	}
	if (n_write_sids == 0)
		return;
	/* is_updating, i.e. inside update cycle: t_crtu > t_crt */
	if ((t->area->t_crtu > t->area->t_crt) && Synch(t, writeOp)) {
		if (wait_write_permission)
			taReadyForSendMessage(t->area, write_sids, n_write_sids,
				t->buffer_size+ t->dyn_buffer_len, &t->mqos, &is_ready);
		else
			is_ready = _True;
		if (is_ready)
			BeginWrite(&t->rtti->mainitem, data, t, write_sids, n_write_sids);
		else {
			if (t->mqos.tp_send_timeout < 0)
				t_waiting_expired = -1;	// infinite
			else
				t_waiting_expired = t->area->t_crtu 
					+ TA_TP(t->mqos.tp_send_timeout);
			Put(t, writeOp, &t->rtti->mainitem, NULL, data,
				write_sids, n_write_sids, TA_WAIT_WRITE_PERMISSION, 
				t_waiting_expired);
		}
	} else
		Put(t, writeOp, &t->rtti->mainitem, NULL, data,
			write_sids, n_write_sids,
			(wait_write_permission ? TA_WAIT_WRITE_PERMISSION : TA_NO_WAIT),
			0);
}

/* ------------------------------------------------------- */
void taWaitTable(TaTable *t, TaLongLong t_waiting_expired)
{
	TaShort waiting;
	if (t_waiting_expired > 0)
		waiting = TA_WAIT_UNTIL_TIMEOUT;
	else if (t_waiting_expired >= TA_REPEAT_FOREACH_RESPONSE)
		waiting = (TaShort) t_waiting_expired;
	if (waiting > TA_WAIT_UNTIL_TIMEOUT || waiting == TA_REPEAT_FOREACH_RESPONSE)
		return;
	if (waiting <= TA_WAIT_ANY_MESSAGE)
		t_waiting_expired = 0;		// should be set after
	else
		t_waiting_expired = taGetTime()+TA_TP(t_waiting_expired);
	Put(t, waitOp, NULL, NULL, NULL, NULL, 0, waiting, t_waiting_expired);
}

/* ------------------------------------------------------- */
void taRepeatTable(TaTable *t, TaTableCommand *command,
	void *par)
{
	taClearTable(t);
	Put(t, tcallOp, NULL, command, par, NULL, 0, 
		TA_REPEAT_FOREACH_RESPONSE, 0);
}

/* ------------------------------------------------------- */
void taClearTable(TaTable *t)
{
	if (Synch(t, clearOp))
		DoClear(t);
	else
		Put(t, clearOp, NULL, NULL, NULL, NULL, 0, TA_NO_WAIT, 0);
}

/* ------------------------------------------------------- */
void taRequestTable(TaTable *t, void *data, TaULong sids, 
	TaBool wait_write_permission, TaTableCommand *command, void *par)
{
	taClearTable(t);
	taWriteTable(t, data, sids, wait_write_permission);
	Put(t, tcallOp, NULL, command, par, NULL, 0, 
		TA_REQUEST, 0);
}

/* ------------------------------------------------------- */
void taPublishTable(TaTable *t, TaBool *permanent, TaTableCommand *command,
	void *par)
{
	TaUShort mid_resp;
	if (permanent)
		mid_resp = TA_PERMANENT_MID;
	else
		mid_resp = 0;
	Put(t, tcallOp, NULL, command, par, NULL, mid_resp, 
		TA_PUBLISH, 0);
}

/* ------------------------------------------------------- */
void taSubscribeTable(TaTable *t, TaBool *controlled, TaTableCommand *command,
	void *par)
{
	TaUShort src_mid;
	if (controlled)
		src_mid = 0;
	else
		src_mid = TA_PERMANENT_MID;
	Put(t, tcallOp, NULL, command, par, NULL, src_mid, 
		TA_SUBSCRIBE, 0);
}

/* ------------------------------------------------------- */
void taUnsubscribeTable(TaTable *t, TaULong sids)
{
	TaShort write_sids[TA_MAX_MASK_ADDRS];
	int i, n_write_sids, n_sendto_sids;
	n_sendto_sids = t->n_sendto_sids;
	if (n_sendto_sids < 0)
		n_sendto_sids = -n_sendto_sids;
	if (sids == TA_ADDR_ALL) {
		if (t->sendto_sids[0] == TA_SIDS_ALL)
			write_sids[0] = TA_SIDS_ALL;
		else
			/* if n_sendto_sids < 0 - multicast else unicast */
			for (i = 0; i < n_sendto_sids; i++)
				write_sids[i] = t->sendto_sids[i];
		n_write_sids = t->n_sendto_sids;
	} else {
		if (t->sendto_sids[0] == TA_SIDS_ALL) {
			for (i = n_write_sids = 0; i < t->area->n_stations
			&& i < TA_MAX_MASK_ADDRS; i++) {
				if (sids & TA_TO_ADDR(i))
					write_sids[n_write_sids++] = (TaShort) i;
			}
		} else {
			for (i = n_write_sids = 0; i < n_sendto_sids; i++) {
				if (sids & TA_TO_ADDR(i))
					write_sids[n_write_sids++] = t->sendto_sids[i];
			}
		}
	}
	if (n_write_sids == 0)
		++t->src_mid;
	else {
		if (t->area->current_mid == TA_MAX_MID)
			t->mid_resp = 1;
		else
			t->mid_resp = t->area->current_mid+1;
		DoSend(t, write_sids, n_write_sids);
	}
}

/* ------------------------------------------------------- */
int DoSend(TaTable *t, TaShort *write_sids, int n_write_sids)
{
	int rc;
	rc = taSetMessage(t->area, t->buffer, t->buffer_size,
		t->msg_type, t->src_oid, t->iface_code,
		write_sids, n_write_sids, t->mid_resp, &t->mqos);

	if (rc >= 0) {
		t->src_mid = (TaUShort) rc;
		if (t->dyn_buffer_len > 0)
			rc = taAddToMessage(t->area, t->dyn_buffer, 
				t->dyn_buffer_len, t->msg_type, 
				t->src_oid, t->iface_code);
	} else
		t->src_mid = 0;
	if (t->on_data && rc < 0) {
		TaErrorData edp;
		edp.errcode = rc;
		(*t->on_data)(&edp,  (TaULong)
			g_rtti_TaErrorData.names_crc, t);
	}
	return rc;
}

/* ------------------------------------------------------- */
int BeginWrite(const TaRTTItem *item, const void *data, TaTable *t,
	TaShort *write_sids, int n_write_sids)
{
	int rc, col = 0, offset_p = 0;
	t->res = 0;
	t->dyn_buffer_len = 0;
	t->send_attempts_counter = 0;
	if (t->match_mode < TA_MSG_SOAP_MATCH_INTERFACE) {
		memset(t->buffer, 0, t->buffer_size);
		WriteItem(item, t, (const char *) data, t->buffer, 
			t->buffer_size, &col, &offset_p, 0);
	} else {
		memset(t->buffer, 0, t->area->qos.soap_buffer_size);
		if (t->match_mode > TA_MSG_SOAP_MATCH_INTERFACE)
			t->mid_resp = t->match_mode;
		rc = taWriteSoapItem(item, t, (const char *) data, t->buffer, 
			t->buffer_size);
		if (rc < 0)
			return rc;
		t->buffer_size = rc;
	}
	rc = DoSend(t, write_sids, n_write_sids);
	if (t->match_mode > TA_MSG_SOAP_MATCH_INTERFACE)
		t->src_mid = (TaUShort) t->msg_type;
	return rc;
}

/* ------------------------------------------------------- */
void WriteItem(const TaRTTItem *ritem, TaTable *t, const char *data,
	char *item_addr, int buffer_size, int *col, int *offset_p, 
	int dyn_dim)
{
	TaOctet i1;
	TaBool b;
	TaChar c;
	TaShort i2;
	TaWChar wc;
	TaLong i4;
	TaLongLong i8;
	TaFloat r4;
	TaDouble r8;
	TaRTTItem *item1;
	int len;
	const char *addr;
	int offset, offset0;

	if (*offset_p > buffer_size) {
//printf("%s offset_p:%d buffer_size:%d\n", ritem->item_name, *offset_p,
//buffer_size);
		*offset_p = 0;
		return;
	}

	switch (ritem->item_type_id) {
	case TaBool_TYPE:
		b = BoolVal(data+ritem->offset);
		memcpy(item_addr+*offset_p, &b, 1);
		++*offset_p;
		++*col;
		break;
	case TaOctet_TYPE:
		i1 = I1Val(data+ritem->offset);
		memcpy(item_addr+*offset_p, &i1, 1);
		++*offset_p;
		++*col;
		break;
	case TaChar_TYPE:
		c = CharVal(data+ritem->offset);
		memcpy(item_addr+*offset_p, &c, 1);
		++*offset_p;
		++*col;
		break;
	case TaShort_TYPE:
		i2 = taH2N16(I2Val(data+ritem->offset));
		memcpy(item_addr+*offset_p, &i2, 2);
		*offset_p += 2;
		++*col;
		break;
	case TaWChar_TYPE:
		wc = taH2N16(WCharVal(data+ritem->offset));
		memcpy(item_addr+*offset_p, &wc, 2);
		*offset_p += 2;
		++*col;
		break;
	case TaLong_TYPE:
		i4 = taH2N32(I4Val(data+ritem->offset));
		memcpy(item_addr+*offset_p, &i4, 4);
		*offset_p += 4;
		++*col;
		break;
	case TaLongLong_TYPE:
		i8 = taH2N64(I8Val(data+ritem->offset));
		memcpy(item_addr+*offset_p, &i8, 8);
		*offset_p += 8;
		++*col;
		break;
	case TaFloat_TYPE:
		r4 = taH2Nr32(R4Val(data+ritem->offset));
		memcpy(item_addr+*offset_p, &r4, 4);
		*offset_p += 4;
		++*col;
		break;
	case TaDouble_TYPE:
		r8 = taH2Nr64(R8Val(data+ritem->offset));
		memcpy(item_addr+*offset_p, &r8, 8);
		*offset_p += 8;
		++*col;
		break;
	case TaArr_TYPE:
		if (dyn_dim == 0)
			dyn_dim = ritem->dim;
		for (i4 = 0; i4 < dyn_dim; i4++) {
			int oldc = *col;
//len= *offset_p;
			WriteItem(ritem->contains, t,
				data + i4 * ritem->contains->size,
				item_addr, buffer_size, col, offset_p, 0);
//printf("a %d>%d tp:%d bsize:%d i:%d dim:%d\n", len, *offset_p, ritem->contains->item_type_id, buffer_size, i4, dyn_dim);
			if (i4 > 0)
				*col = oldc;
		}
		break;
	case TaRec_TYPE:
		if (ritem->contains)
			item1 = ritem->contains;
		else
			item1 = ritem->next;
		while (item1) {
//i4= *offset_p;
			WriteItem(item1, t, data, item_addr, buffer_size, 
				col, offset_p, 0);
//printf("r %d>%d tp:%d\n", i4, *offset_p, item1->item_type_id);
			item1 = item1->next;
		}
		if (ritem->contains)
			++*col;
		break;
	case TaPtr_TYPE:
		addr = *(const char **)(data+ritem->offset);
		offset0 = taH2N32(t->dyn_buffer_len);
		if (addr == NULL) {
			i4 = 0; len = 0; i2 = 0;
			memcpy(item_addr+*offset_p, &len, 4);
			memcpy(item_addr+*offset_p+4, &offset0, 4);
			memcpy(item_addr+*offset_p+14, &i2, 2);
		} else {
			const TaRTTIadata *adt;
			adt = (const TaRTTIadata *) (addr-sizeof(TaRTTIadata));
			len = ItemPackedSize(addr, ritem->contains, _True);
//printf("ItemPackedSize:%d t->dyn_buffer_len:%d t->dyn_buffer_size:%d %s\n",
//len, t->dyn_buffer_len, t->dyn_buffer_size, ritem->contains->item_name);
			if (t->dyn_buffer_len+len > t->dyn_buffer_size) {
				if (t->dyn_buffer_len > 0) {
					char *bp = t->dyn_buffer;
					t->dyn_buffer_size = t->dyn_buffer_len
						+ len;
					t->dyn_buffer = (char *) 
						malloc(t->dyn_buffer_size);
					memset(t->dyn_buffer, 0, 
						t->dyn_buffer_size);
					for (i4 = 0; i4 < t->dyn_buffer_len; 
					i4++)
						t->dyn_buffer[i4] = bp[i4];
					free(bp);
				} else {
					t->dyn_buffer_size = len;
					t->dyn_buffer = (char *) 
						malloc(len);
					memset(t->dyn_buffer, 0, len);
				}
			}
			item1 = ritem->contains;
			// NULL ???
			if (!item1)
				return;
			i4 = 0;
			offset = t->dyn_buffer_len;
			t->dyn_buffer_len += len;
			if (t->dyn_buffer_len > 0)
				WriteItem(item1, t, addr, t->dyn_buffer, 
					t->dyn_buffer_size, &i4, 
					&offset, adt->dim);
			len = taH2N32(len);
			i2 = taH2N16((TaShort) adt->dim);
			memcpy(item_addr+*offset_p, &len, 4);
			memcpy(item_addr+*offset_p+4, &offset0, 4);
			memcpy(item_addr+*offset_p+14, &i2, 2);
		}
/*printf(
"wptr: nb:%d/%d offs:%d/%d/%d%c dim:%d mt:%d ic:%d :: %s %x dyn_l:%d\n", 
taN2H32(len), len, taN2H32(offset0), offset0, *offset_p, (item_addr == t->dyn_buffer ? 'd' : 's'),
taN2H16(i2), ritem->names_crc, ritem->types_crc,
ritem->item_name, addr, t->dyn_buffer_len);*/
		i4 = taH2N32(ritem->names_crc);
		i2 = taH2N16(ritem->types_crc);
		memcpy(item_addr+*offset_p+8, &i4, 4);
		memcpy(item_addr+*offset_p+12, &i2, 2);
		*offset_p += 16;
		++*col;
		break;
	case TaProc_TYPE:
		break;
	default:
		return;
	}
//if(t->dyn_buffer==item_addr && ritem->item_type_id==1)
//printf("%x>%x ~%d=>%d~%d %c:%d %c:%d tp:%d\n", data, item_addr,
//ofs, *offset_p, ritem->offset, item_addr[ofs], item_addr[ofs],
//	data[ritem->offset], data[ritem->offset], ritem->item_type_id);
}

/* ------------------------------------------------------- */
int ItemPackedSize(const char *addr, const TaRTTItem *item, TaBool dyn)
{
	const TaRTTItem *item1;
	int sum, dim, i;
	if (!item)
		return 0;
	switch (item->item_type_id) {
	case TaArr_TYPE:
		if (dyn) {
			const TaRTTIadata *adt;
			if (!addr)
				return 0;
			adt = (const TaRTTIadata *) (addr-sizeof(TaRTTIadata));
			dim = adt->dim;
		} else
			dim = item->dim;
		return item->contains->size_packed * dim;
		/*
		if (dim == 0)
			return 0;
		item1 = item->contains;
		if (item1->with_ptr) {
			sum = 0;
			for (i = 0; i < dim; i++) {
				sum += ItemPackedSize(addr + i * item->size,
					item1, _False);
			}
printf("sum:%d dim:%d size_packed:%d %s\n", sum, dim, item1->size_packed,
item1->item_name);
			return sum;
		} else {
			// all items of equal size
			return dim * ItemPackedSize(addr, item1, _False);
		}
		*/
	case TaRec_TYPE:
		sum = 0;
		item1 = item->contains;
		while (item1) {
			const char *addr1 = *(const char **)(addr+item->offset);
			sum += ItemPackedSize(addr1, item1, _False);
			item1 = item1->next;
		}
		return sum;
	case TaPtr_TYPE:
		return 16 + ItemPackedSize(addr, item->contains, _True);
	}
	return item->item_type_size;
}

/* ------------------------------------------------------- */
void *taAllocWithRTTI(TaLong itemsize, TaLong dim)
{
	TaRTTIadata *adt;
	adt = (TaRTTIadata *) malloc(itemsize * dim
		+ sizeof(TaRTTIadata));
	if (!adt)
		return NULL;
	memset(adt, 0, itemsize * dim + sizeof(TaRTTIadata));
	adt->size = itemsize * dim;
	adt->length = adt->size;
	adt->dim = dim;
	adt->elemsize = itemsize;
	return &adt[1];
}

/* ------------------------------------------------------- */
void taFreeWithRTTI(void *ptr)
{
	char *p = (char *) ptr;
	free ((void *)(p-sizeof(TaRTTIadata)));
}

/* ------------------------------------------------------- */
void *taReallocWithRTTI(void *ptr, TaLong newdim)
{
	char *p = (char *) ptr;
	TaRTTIadata *adt;
	if (!ptr || newdim < 0)
		return 0;
	adt = (TaRTTIadata *) (p-sizeof(TaRTTIadata));
	if (newdim > adt->size/adt->elemsize) {
		TaLong itemsize = adt->elemsize;
		void *newptr = taAllocWithRTTI(itemsize, newdim);
		if (newptr)
			memcpy(newptr, ptr, adt->size);
		free ((void *) adt);
		return newptr;
	}
	if (newdim != adt->dim) {
		adt->length = newdim * adt->elemsize;
		adt->dim = newdim;
	}
	return (void *) p;
}

/* ------------------------------------------------------- */
TaLong taLengthWithRTTI(void *ptr)
{
	char *p = (char *) ptr;
	TaRTTIadata *adt;
	if (!ptr)
		return 0;
	adt = (TaRTTIadata *) (p-sizeof(TaRTTIadata));
	return adt->dim;
}

/* ------------------------------------------------------- */
TaLong taLengthAllocatedRTTI(void *ptr)
{
	char *p = (char *) ptr;
	TaRTTIadata *adt;
	if (!ptr)
		return 0;
	adt = (TaRTTIadata *) (p-sizeof(TaRTTIadata));
	return adt->size/adt->elemsize;
}

/* ------------------------------------------------------- */
TaChar *taDupStringWithRTTI(const TaChar *src)
{
	int len;
	TaChar *dst;
	if (!src || !*src)
		return NULL;
	len = strlen(src);
	dst = (TaChar *)taAllocWithRTTI(1, len+1);
	if (!dst)
		return NULL;
	strncpy(dst, src, len);
	dst[len] = 0;
	return dst;
}

/* ------------------------------------------------------- */
TaWChar *taDupWStringWithRTTI(const TaWChar *src)
{
	int i, len;
	TaWChar *dst;
	if (!src || !*src)
		return NULL;
	len = strlen(src);
	dst = (TaWChar *)taAllocWithRTTI(sizeof(TaWChar), len+1);
	if (!dst)
		return NULL;
	for (i = 0; i < len; i++)
		dst[i] = src[i];
	dst[len] = 0;
	return dst;
}
