/*
 *                       ***** TA Memprot *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef _WINDOWS
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <dlfcn.h>
#endif


/* SHM */
int taInitShm(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel);
void taCloseShm(TaChannel *channel);
int taTryConnectShmChannel(TaArea *area, TaStation *dstation, int cid);
int taSendToShm(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel);
int taRecvFromShm(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);

/* LIB */
int taInitLib(TaBool is_datahost, TaBool is_send_station,
	const TaStation *station, TaChannel *channel);
void taCloseLib(TaChannel *channel);
int taTryConnectLibChannel(TaArea *area, TaStation *dstation, int cid);

/* MSGQ */
int taInitMsgq(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel);
void taCloseMsgq(TaChannel *channel);
int taTryConnectMsgqChannel(TaArea *area, TaStation *dstation, int cid);
int taSendToMsgq(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel);
int taRecvFromMsgq(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);

// memory additional list of protocols
static TaChannelProtocol g_memprots[] = {
	{"SHM_MCAST", TA_CHANNEL_SHM_MCAST, TA_FLAGS_MEMORY|TA_FLAGS_MCAST,
		taInitShm, taCloseShm, NULL, taTryConnectShmChannel,
		taSendToShm, taRecvFromShm, NULL},
        {"LIB", TA_CHANNEL_LIB, TA_FLAGS_MEMORY|TA_FLAGS_LIB,
                taInitLib, taCloseLib, NULL, taTryConnectLibChannel,
                taSendToShm, taRecvFromShm, NULL}
#ifndef _WINDOWS
	,
	{"MSGQ", TA_CHANNEL_MSGQ, TA_FLAGS_MEMORY|TA_FLAGS_MESSAGE,
		taInitMsgq, taCloseMsgq, NULL, taTryConnectMsgqChannel,
		taSendToMsgq, taRecvFromMsgq, NULL}
#endif
};

static int OnStationLibUpdate(struct TaStation *station,
	TaLongLong t_updating, void *client_data);
static TaULong readFromBuffer(char *dst_buf, const char *src_buf, 
	TaULong recv_offset, TaULong nbytes, TaULong dst_len);
static TaULong writeToBuffer(char *dst_buf, const char *src_buf, 
	TaULong *o_end, TaULong nbytes, TaULong dst_len);

/* ------------------------------------------------------- */
int taInitMemProts()
{
	unsigned i;
	for (i = 0; i < sizeof(g_memprots)/sizeof(TaChannelProtocol); i++)
	{
		if (taAddChannelProtocol(&g_memprots[i]) < 0)
			return TA_ERR_ARGUMENT;
	}
	return i;
}

#ifndef _WINDOWS
/* ------------------------------------------------------- */
int taInitShm(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	key_t keyt;
	int memid;
	char buf[32];
	struct stat st_file;
	TaBool connect_existing = _False;

	if (!channel)
		return TA_ERR_ARGUMENT;
	// I create my shm for receiving
	// and connect other for saving
	if (!is_send_station)
		return 0;

	sprintf(buf, "area%d.log", (int)station->area->aid);
	if (stat(buf, &st_file) < 0) {
		FILE *f;
		f = fopen(buf, "w");
		if (f) {
			fprintf(f, "%s\n", buf);
			fclose(f);
		}
	}
	keyt = ftok (buf, (int)station->area->host_station->sid+1);
	if (keyt < 0)
		return TA_ERR_ARGUMENT;
	if (channel->qos.mem_bufsize < TA_SHMMIN)
		return TA_ERR_OUT_OF_MEMORY;
	// try to get already created
	memid = shmget(keyt, channel->qos.mem_bufsize, 0777);
	if (memid < 0)
		memid = shmget(keyt, channel->qos.mem_bufsize, 
			IPC_CREAT | 0777);
	else
		connect_existing = _True;
	if (memid < 0)
		return TA_NOT_FOUND;
	channel->send_buffer = (char *)shmat(memid, NULL, 0);
	if (!channel->send_buffer)
		return TA_ERR_OUT_OF_MEMORY;
	if (!connect_existing)
		memset(channel->send_buffer, 0, channel->qos.mem_bufsize);
	channel->send_id = memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file,
		"SHM existing:%d ftok:%u send_id:%d addr:%x memsize:%d\n", 
			(int)connect_existing, (unsigned)keyt, 
			channel->send_id, (unsigned)channel->send_buffer, 
			(unsigned)channel->qos.mem_bufsize);
	channel->link_stat = TA_LINK_INIT;
	return 1;
}

/* ------------------------------------------------------- */
void taCloseShm(TaChannel *channel)
{
	struct shmid_ds sds;
	int rc;
	if (channel->send_buffer) {
		rc = shmdt(channel->send_buffer);
		if (ta_trace_level & TA_TRAL_TRANSPORT) {
			if (rc < 0)
				fprintf(ta_file, "* shmdt:%x send rc:%d errno:%d %s\n", 
					(unsigned)channel->send_buffer,
					rc, errno, strerror(errno));
			else
				fprintf(ta_file, "shmdt:%x send done\n",
					(unsigned)channel->send_buffer);
		}
		channel->send_buffer = NULL;
	}
	if (channel->recv_buffer) {
		rc = shmdt(channel->recv_buffer);
		if (ta_trace_level & TA_TRAL_TRANSPORT) {
			if (rc < 0)
				fprintf(ta_file, "* shmdt:%x recv rc:%d errno:%d %s\n", 
					(unsigned)channel->recv_buffer,
					rc, errno, strerror(errno));
			else
				fprintf(ta_file, "shmdt:%x recv done\n",
					(unsigned)channel->recv_buffer);
		}
		channel->recv_buffer = NULL;
	}
	if (channel->qos.shm_rm_onexit && channel->send_id > 0) {
		rc = shmctl(channel->send_id, IPC_RMID, &sds);
		if (ta_trace_level & TA_TRAL_TRANSPORT) {
			if (rc < 0)
				fprintf(ta_file, "* shmctl RMID:%d rc:%d errno:%d %s\n", 
					channel->send_id, rc, errno, 
					strerror(errno));
			else
				fprintf(ta_file, "shmctl RMID:%d done\n",
					channel->send_id);
		}
		channel->send_id = 0;
	}
	channel->link_stat = TA_LINK_CLOSED;
}

/* ------------------------------------------------------- */
int taTryConnectShmChannel(TaArea *area, TaStation *dstation, int cid)
{
	TaChannel *dstchannel;
	key_t keyt;
	int memid;
	char buf[32];
	const TaULong *o_end;
	struct shmid_ds sds;
	size_t shm_segsz;

	dstchannel = &dstation->channels[cid];
	if (!dstchannel->is_enabled)
		return TA_ERR_WRONG_DATA;
	// I create my shm for receiving
	// and connect other for saving
	if (dstation == area->host_station)
		return 0;

	sprintf(buf, "area%d.log", (int)dstation->area->aid);
	keyt = ftok (buf, dstation->sid+1);
	if (keyt < 0)
		return TA_ERR_ARGUMENT;
	// get bufsize
	// try to get already created
	//memid = shmget(keyt, dstchannel->qos.mem_bufsize, 0777);
	memid = shmget(keyt, TA_SHMMIN, 0777);
	if (memid < 0)
		return TA_NOT_FOUND;
	if (shmctl(memid, IPC_STAT, &sds) < 0)
		return TA_NOT_FOUND;
	shm_segsz = sds.shm_segsz;
	memid = shmget(keyt, shm_segsz, 0777);
	if (memid < 0)
		return TA_NOT_FOUND;
	dstchannel->recv_buffer = (char *)shmat(memid, NULL, 0);
	if (!dstchannel->recv_buffer)
		return TA_ERR_OUT_OF_MEMORY;
	dstchannel->recv_id = memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "SHM connect ftok:%u recv_id:%d addr:%x memsize:%d\n", 
			(unsigned)keyt, dstchannel->recv_id, 
			(unsigned)dstchannel->recv_buffer, 
			(unsigned)dstchannel->qos.mem_bufsize);
	o_end = (const TaULong *) &dstchannel->recv_buffer[4];
	dstchannel->recv_offset = *o_end;
	dstchannel->link_stat = TA_LINK_OK;
	return 1;
}
#else
/* ------------------------------------------------------- */
int taInitShm(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	char buf[32];
	struct stat st_file;
	int i, flen;
	HANDLE memid;
	TaBool connect_existing = _True;
	if (!channel)
		return TA_ERR_ARGUMENT;
	// I create my shm for receiving
	// and connect other for saving
	if (!is_send_station)
		return 0;

	sprintf(buf, "area%d_st%d.log", (int)station->area->aid,
		(int)station->area->host_station->sid+1);
	if (stat(buf, &st_file) < 0)
		flen = -1;
	else
		flen = st_file.st_size;
	if (flen != channel->qos.mem_bufsize) {
		FILE *f;
		f = fopen(buf, "w");
		if (f) {
			flen = channel->qos.mem_bufsize;
			for (i = 0; i < flen-1; i++)
				fprintf(f, "%c", ' ');
			fprintf(f, "%c", '\0');
			fclose(f);
		}
		connect_existing = _False;
	}
	memid = CreateFileMapping(
                INVALID_HANDLE_VALUE,   // use paging file
                NULL,                   // default security attributes
                PAGE_READWRITE,         // read/write access
                0,                      // size: high 32-bits
                channel->qos.mem_bufsize, // size: low 32-bits
                TEXT(buf)); 
	channel->send_buffer = (char *) MapViewOfFile( 
                memid,          // object to map view of
                FILE_MAP_WRITE, // read/write access
                0,              // high offset:  map from
                0,              // low offset:   beginning
                0);             // default: map entire file

	if (!channel->send_buffer)
		return TA_ERR_OUT_OF_MEMORY;
	if (!connect_existing)
		memset(channel->send_buffer, 0, channel->qos.mem_bufsize);
	channel->send_id = (int) memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file,
		"SHM existing:%d send_id:%d addr:%x memsize:%d\n", 
			(int)connect_existing, channel->send_id, 
			(unsigned)channel->send_buffer, (unsigned)channel->qos.mem_bufsize);
	channel->link_stat = TA_LINK_INIT;
	return 1;
}

/* ------------------------------------------------------- */
void taCloseShm(TaChannel *channel)
{
	int rc;

	if (channel->send_buffer) {
		rc = UnmapViewOfFile(channel->send_buffer);
		channel->send_buffer = NULL;
	}
	if (channel->recv_buffer) {
		rc = UnmapViewOfFile(channel->recv_buffer);
		channel->recv_buffer = NULL;
	}
    // Close the process's handle to the file-mapping object
	if (channel->qos.shm_rm_onexit && channel->send_id > 0) {
		rc = CloseHandle((HANDLE) channel->send_id);
		channel->send_id = 0;
	}
	channel->link_stat = TA_LINK_CLOSED;
}

/* ------------------------------------------------------- */
int taTryConnectShmChannel(TaArea *area, TaStation *dstation, int cid)
{
	TaChannel *dstchannel;
	HANDLE memid;
	char buf[32];
	const TaULong *o_end;
	struct stat st_file;

	dstchannel = &dstation->channels[cid];
	if (!dstchannel->is_enabled)
		return TA_ERR_WRONG_DATA;
	// I create my shm for receiving
	// and connect other for saving
	if (dstation == area->host_station)
		return 0;

	sprintf(buf, "area%d_st%d.log", (int)dstation->area->aid,
		(int)dstation->sid+1);
	if (stat(buf, &st_file) < 0)
		return TA_NOT_FOUND;
	// get bufsize
	// try to get already created
	memid = CreateFileMapping( 
                INVALID_HANDLE_VALUE,   // use paging file
                NULL,                   // default security attributes
                PAGE_READWRITE,         // read/write access
                0,                      // size: high 32-bits
				st_file.st_size,		// size: low 32-bits
                TEXT(buf)); 
	if ((int)memid < 0)
		return TA_NOT_FOUND;
	dstchannel->recv_buffer = (char *) MapViewOfFile(
                memid,          // object to map view of
                FILE_MAP_WRITE, // read/write access
                0,              // high offset:  map from
                0,              // low offset:   beginning
                0);             // default: map entire file
	if (!dstchannel->recv_buffer)
		return TA_ERR_OUT_OF_MEMORY;
	dstchannel->recv_id = memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "SHM connect recv_id:%d addr:%x memsize:%d\n", 
			dstchannel->recv_id, (unsigned)dstchannel->recv_buffer, 
			(unsigned)st_file.st_size);
	o_end = (const TaULong *) &dstchannel->recv_buffer[4];
	dstchannel->recv_offset = *o_end;
	dstchannel->link_stat = TA_LINK_OK;
	return 1;
}
#endif

/* ------------------------------------------------------- */
int taInitLib(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	if (!is_send_station)
		return 0;
	channel->lib = (TaLibStationData *)
		malloc(sizeof(TaLibStationData));
	memset(channel->lib, 0, sizeof(TaLibStationData));
	channel->send_buffer = (char *)
		malloc(channel->qos.mem_bufsize);
	memset(channel->send_buffer, 0, channel->qos.mem_bufsize);
	channel->send_id = station->area->host_station->sid;
        channel->link_stat = TA_LINK_INIT;
        return 1;
}

/* ------------------------------------------------------- */
void taCloseLib(TaChannel *channel)
{
	if (!channel->lib)
		return;
	if (channel->lib->plib) {
#ifdef _WINDOWS
		FreeLibrary((HINSTANCE)channel->lib->plib);
#else
		//dlclose(channel->lib->plib);
#endif
		channel->lib->plib = 0;
        	channel->link_stat = TA_LINK_CLOSED;
	}
}

/* ------------------------------------------------------- */
int OnStationLibUpdate(struct TaStation *station,
	TaLongLong t_updating, void *client_data)
{
	TaChannel *channel;
	if (station->lib_cid < 0 || station->lib_cid >= station->n_channels)
		return TA_ERR_ARGUMENT;
	channel = &station->channels[station->lib_cid];
	if (!channel->lib)
		return TA_NOT_FOUND;
	return
	taUpdateLibData(channel->lib);
	// actual client data
	//return (*channel->lib->on_update)(station, t_updating, channel->lib);
}

/* ------------------------------------------------------- */
int taTryConnectLibChannel(TaArea *area, TaStation *dstation, int cid)
{
	TaChannel *dstchannel, *libchannel;
        void *plib;
	int rc;
        TaLibStationData *libStationData;
        TaGetLibStationDataProc *getLibStationData;
	dstchannel = &dstation->channels[cid];
	if (!dstchannel->is_enabled || !dstchannel->lib)
		return TA_ERR_WRONG_DATA;
#ifdef _WINDOWS
	plib = (void *)LoadLibrary((const WCHAR *)dstation->name);
	if (ta_trace_level & TA_TRAL_TRANSPORT 
	|| !plib && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (plib)
			fprintf(ta_file, "LoadLibrary %s returns:%x\n", 
				dstation->name, (unsigned)plib);
		else
			fprintf(ta_file, "* LoadLibrary %s: error 0x%x\n",
				dstation->name,
				GetLastError());
	}
	if (!plib)
		return TA_NOT_FOUND;
        getLibStationData = (TaGetLibStationDataProc *)
                GetProcAddress((HMODULE)plib, "taGetLibStationData");
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| !getLibStationData && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (getLibStationData)
			fprintf(ta_file,
			"GetProcAddress taGetLibStationData returns:%x\n", 
				getLibStationData);
		else
			fprintf(ta_file,
			"* GetProcAddress taGetLibStationData: error 0x%x\n",
				GetLastError());
	}
#else
	plib = dlopen(dstation->name, RTLD_NOW);
	if (!plib && !strstr(dstation->name, ".so")) {
		char buffer[128];
		sprintf(buffer, "%s.so", dstation->name);
		plib = dlopen(buffer, RTLD_NOW);
	}
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| !plib && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (plib)
			fprintf(ta_file, "dlopen %s returns:%x\n", 
				dstation->name, 
				(unsigned)plib);
		else
			fprintf(ta_file, "* dlopen %s: %s\n",
				dstation->name, dlerror());
	}
	if (!plib)
		return TA_NOT_FOUND;
        getLibStationData = (TaGetLibStationDataProc *)
		dlsym(plib, "taGetLibStationData");
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| !getLibStationData && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (getLibStationData)
			fprintf(ta_file, "dlsym taGetLibStationData returns:%x\n",
				(unsigned)getLibStationData);
		else
			fprintf(ta_file, "* dlsym taGetLibStationData: %s\n",
				dlerror());
	}
#endif
	if (!getLibStationData)
		return TA_NOT_FOUND;
	taCloseLib(dstchannel);
	dstchannel->lib->station = dstation;
	dstchannel->lib->cid = cid;
	taInitLibArea(area, dstchannel->lib, &dstchannel->lib->libarea);
	rc = (*getLibStationData)(dstchannel->lib);
	if (rc < 0) {
		if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
			fprintf(ta_file, "* taGetLibStationData rc:%d\n", rc);
		return TA_ERR_WRONG_DATA;
	}
	taInitAreaBuffers(dstchannel->lib->area);
	dstchannel->lib->plib = plib;
	dstation->lib_cid = cid;
	dstation->on_update = OnStationLibUpdate;

	libchannel = &dstchannel->lib->area->send_station->channels[cid];
	libchannel->send_buffer = (char *)
		malloc(libchannel->qos.mem_bufsize);
	memset(libchannel->send_buffer, 0, libchannel->qos.mem_bufsize);
	libchannel->send_id = dstchannel->lib->area->host_station->sid;
	libchannel->qos.assign_sids = _True;
	dstchannel->qos.assign_sids = _True;

	dstchannel->recv_buffer = libchannel->send_buffer;
	libchannel->recv_buffer = dstchannel->send_buffer;

	dstchannel->link_stat = TA_LINK_OK;
        libchannel->link_stat = TA_LINK_OK;

	return 1;
}

/* ------------------------------------------------------- */
TaULong readFromBuffer(char *dst_buf, const char *src_buf, 
	TaULong recv_offset, TaULong nbytes, TaULong dst_len)
{
	TaULong i, j;
	if (nbytes == 0 || dst_len == 0)
		return 0;	// nothing done
	if (nbytes > dst_len)
		nbytes = dst_len;
	for (i = 0; i < nbytes; i++) {
		j = (i+recv_offset) % dst_len;
		dst_buf[i] = src_buf[j];
	}
	return nbytes;
}

/* ------------------------------------------------------- */
TaULong writeToBuffer(char *dst_buf, const char *src_buf, 
	TaULong *o_end, TaULong nbytes, TaULong dst_len)
{
	TaULong offset, i, j;
	if (nbytes == 0 || dst_len == 0)
		return 0;	// nothing done
	if (nbytes > dst_len)
		nbytes = dst_len;
	if (o_end)
		offset = *o_end;
	else
		offset = 0;
	for (i = 0; i < nbytes; i++) {
		//if (i+offset < dst_len)
		//	j = i+offset;
		//else
		//	j = 8+i+offset-dst_len;
		j = (i+offset) % dst_len;
		dst_buf[j] = src_buf[i];
	}
	if (o_end)
		*o_end = (nbytes+*o_end) % dst_len;
	return nbytes;
}

/* ------------------------------------------------------- */
int taSendToShm(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel)
{
	TaULong i, space_size, len;
	TaULong *o_begin;
	TaULong *o_end;
	char *send_buf;
	if (!channel->send_buffer)
		return TA_ERR_TRANSPORT;
	len = channel->qos.mem_bufsize-8;
	o_begin = (TaULong *) channel->send_buffer;
	o_end = (TaULong *) &channel->send_buffer[4];
	send_buf = &channel->send_buffer[8];
	// 1. check for space
	if ((TaULong)buffer_length > len)
		return TA_ERR_OUT_OF_MEMORY;
	if (*o_end >= *o_begin)
		space_size = *o_end-*o_begin;
	else
		space_size = len - (*o_begin-*o_end);
	// 2. clean old messages
	i = *o_begin;
	while (len-space_size < (TaULong)buffer_length) {
		TaSRPSHeader srps_header;
		TaULong urc =
		readFromBuffer((char *)&srps_header, send_buf, 
			i, sizeof(TaSRPSHeader), len);
		if (urc != sizeof(TaSRPSHeader)) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file,
		"* mem_send:%d clean error, begin:%d end:%d\n",
					(int)channel->send_id, 
					(int)*o_begin, (int)*o_end);
			i = *o_end;
			break;
		}
		srps_header.msg_type = 0;	// clean message data
		writeToBuffer(send_buf, (const char *)&srps_header, &i, 
			sizeof(TaSRPSHeader), len);
		space_size = space_size - taN2H32(srps_header.length);
	}
	// 3. change begin position
	*o_begin = i;
	// 4. copy new buffer
	writeToBuffer(send_buf, buffer, o_end, (TaULong) buffer_length, len);
	if (ta_trace_level & TA_TRAL_TRANSPORT) {
		TaSRPSHeader *sh = (TaSRPSHeader *) buffer;
		fprintf(ta_file, 
		"mem_send:%d %d bytes, mid:%d begin:%d end:%d ch:%x ctype:%d\n",
			(int)channel->send_id, buffer_length, (TaUShort)taN2H16(sh->mid),
			(int)*o_begin, (int)*o_end, channel, channel->ctype);
	}

	return buffer_length;
}

/* ------------------------------------------------------- */
int taRecvFromShm(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel)
{
	int rc;
	TaULong old_o_begin;
	const volatile TaULong *o_begin, *o_end;
	const char *recv_buf;
	TaULong len;

	if (!channel->recv_buffer)
		return TA_ERR_TRANSPORT;
	len = channel->qos.mem_bufsize-8;
	o_begin = (const TaULong *) channel->recv_buffer;
	o_end = (const TaULong *) &channel->recv_buffer[4];
	recv_buf = &channel->recv_buffer[8];
	if (*o_end == *o_begin)
		return 0;

	rc = 0;
	old_o_begin = *o_begin;
	while (channel->recv_offset != *o_end) {
		TaBool bwithin = _False;
		TaSRPSHeader srps_header;
		TaULong urc, length;
		// 1. is recv_offset within?
		if (old_o_begin != *o_begin)
			old_o_begin = *o_begin;
		if (*o_end > *o_begin) {
			if (channel->recv_offset >= *o_begin
			&&  channel->recv_offset < *o_end)
				bwithin = _True;
		} else {
			if (!(channel->recv_offset >= *o_end
			&&  channel->recv_offset < *o_begin))
				bwithin = _True;
		}
		if (0&& !bwithin) {
			channel->recv_offset = *o_begin;
			continue;
		}
	
		// 2. check if valid message
		urc =
		readFromBuffer((char *)&srps_header, recv_buf, 
			channel->recv_offset, sizeof(TaSRPSHeader), len);
		if (old_o_begin != *o_begin)
			continue;	// if changed - do it again
		if (urc != sizeof(TaSRPSHeader)) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file,
	"* mem_recv bad.len:%d recv_id:%d recv_offset:%d o_begin:%d o_end:%d\n",
					(int)urc, (int)channel->recv_id, 
					(int)channel->recv_offset, 
					(int)*o_begin, (int)*o_end);
			channel->recv_offset = *o_end;
			return TA_ERR_WRONG_DATA;
		}
		if (srps_header.major_version != TA_MAJOR_VERSION
		|| strncmp(srps_header.protname, "SRPS", 4)) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file,
	"* mem_recv prot, recv_id:%d recv_offset:%d o_begin:%d o_end:%d\n",
					(int)channel->recv_id, 
					(int)channel->recv_offset, 
					(int)*o_begin, (int)*o_end);
			channel->recv_offset = *o_end;
			return TA_ERR_WRONG_DATA;
		}

		length = (TaULong) taN2H32(srps_header.length);
		// 3. try to read message
		if (length < (TaULong) buffer_length)
			buffer_length = (int) length;
		urc =
		readFromBuffer(buffer, recv_buf, 
			channel->recv_offset, buffer_length, len);
		if (old_o_begin != *o_begin)
			continue;	// if changed - do it again
		channel->recv_mid = taN2H16(srps_header.mid);

		rc = (int) urc;
		// mid is not used
		if (ta_trace_level & TA_TRAL_TRANSPORT)
			fprintf(ta_file,
	"mem_recv id:%d %d bytes recv_offset:%d recv_mid:%d length:%d"
	" begin:%d end:%d\n",
				(int)channel->recv_id, rc,
				(int)channel->recv_offset,
				(int)channel->recv_mid, 
				(int)length,
				(int)*o_begin, (int)*o_end);
		channel->recv_offset = 
			(channel->recv_offset+length) % len;
		break;
	}
	return rc;
}

#ifndef _WINDOWS
/* ------------------------------------------------------- */
int taInitMsgq(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	key_t keyt;
	int memid;
	char buf[32];
	struct stat st_file;
	TaBool connect_existing = _False;

	if (!channel)
		return TA_ERR_ARGUMENT;
	if (!is_datahost)
		return 0;

	sprintf(buf, "area%d.txt", (int)station->area->aid);
	if (stat(buf, &st_file) < 0) {
		FILE *f;
		f = fopen(buf, "w");
		if (f) {
			fprintf(f, "%s\n", buf);
			fclose(f);
		}
	}
	keyt = ftok (buf, (int)station->sid+1);
	if (keyt < 0)
		return TA_ERR_ARGUMENT;
	// create msg queue for receiving
	memid = msgget(keyt, 0777);
	if (memid < 0)
		memid = msgget(keyt, IPC_CREAT | 0777);
	else
		connect_existing = _True;
	if (memid < 0)
		return TA_NOT_FOUND;
	channel->recv_id = memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file,
		"MSGQ existing:%d ftok:%u recv_id:%d\n", 
			(int)connect_existing, (unsigned)keyt, 
			channel->recv_id);
	channel->link_stat = TA_LINK_OK;
	return 1;
}

/* ------------------------------------------------------- */
void taCloseMsgq(TaChannel *channel)
{
	struct msqid_ds sds;
	int rc;
	if (channel->qos.shm_rm_onexit && channel->recv_id > 0) {
		rc = msgctl(channel->recv_id, IPC_RMID, &sds);
		if (ta_trace_level & TA_TRAL_TRANSPORT) {
			if (rc < 0)
				fprintf(ta_file, "* msgctl RMID:%d rc:%d errno:%d %s\n", 
					channel->recv_id, rc, errno, 
					strerror(errno));
			else
				fprintf(ta_file, "msgctl RMID:%d done\n",
					channel->recv_id);
		}
		channel->recv_id = 0;
	}
	channel->link_stat = TA_LINK_CLOSED;
}

/* ------------------------------------------------------- */
int taTryConnectMsgqChannel(TaArea *area, TaStation *dstation, int cid)
{
	TaChannel *dstchannel;
	key_t keyt;
	int memid;
	char buf[32];

	dstchannel = &dstation->channels[cid];
	if (!dstchannel->is_enabled)
		return TA_ERR_WRONG_DATA;
	// I create my shm for receiving
	// and connect other for saving
	if (dstation == area->host_station)
		return 0;

	sprintf(buf, "area%d.txt", (int)dstation->area->aid);
	keyt = ftok (buf, dstation->sid+1);
	if (keyt < 0)
		return TA_ERR_ARGUMENT;
	// get bufsize
	// try to get already created
	memid = msgget(keyt, 0777);
	if (memid < 0)
		return TA_NOT_FOUND;
	dstchannel->send_id = memid;
	if (ta_trace_level & TA_TRAL_TRANSPORT)
		fprintf(ta_file, "MSGQ connect ftok:%u send_id:%d\n", 
			(unsigned)keyt, dstchannel->send_id);
	dstchannel->link_stat = TA_LINK_OK;
	return 1;
}

/* ------------------------------------------------------- */
int taSendToMsgq(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel)
{
	int rc;
	rc = msgsnd(channel->send_id, (const struct msgbuf *)buffer, 
		buffer_length, IPC_NOWAIT);
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| rc < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (rc < 0)
			fprintf(ta_file, "* msg_send:%d %d bytes rc:%d errno:%d %s\n",
				(int)channel->send_id, buffer_length, rc,
				errno, strerror(errno));
		else
			fprintf(ta_file, "msg_send:%d %d bytes rc:%d\n",
				(int)channel->send_id, buffer_length, rc);
	}
	if (rc < 0)
		return rc;
	return buffer_length;
}

/* ------------------------------------------------------- */
int taRecvFromMsgq(TaArea *area, char *buffer, int buffer_length, TaChannel *channel)
{
	int rc;
	rc = msgrcv(channel->recv_id, (struct msgbuf *)buffer, 
		buffer_length, 0, IPC_NOWAIT);
	if (rc < 0 && errno == ENOMSG)
		return 0;	// no message in the queue
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| rc < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (rc < 0)
			fprintf(ta_file, "* msg_recv:%d rc:%d errno:%d %s\n",
				(int)channel->recv_id, rc,
				errno, strerror(errno));
		else
			fprintf(ta_file, "msg_recv:%d rc:%d\n",
				(int)channel->recv_id, rc);
	}
	return rc;
}
#endif
