/*
 *                       ***** TA Message *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

static TaMessageQos g_message_qos = {
	TA_PROT_ANY, 	// reliability
	_False, 	// is_redundant
	0, 		// send_mask
	3000000,	// tp_timeout
	2000000,	// tp_send_timeout
	0,		// n_send_attempts;
	_False,		// send_once
	TA_ACK_NONE,	// ack
	_False,		// all_channels_ready
	_False,		// all_sids_ready
	10};		// dim_chunk

static char g_ping_message[sizeof(TaSRPSHeader)];
static char g_phase_message[sizeof(TaSRPSHeader)];

static void dumpMessage(const TaMessage *message, char fs, int nbytes);

const TaShort ta_addr_all = (const TaShort) 0xFFFF;

/* ------------------------------------------------------- */
void taDefaultMessageQos(TaMessageQos *qos)
{
	*qos = g_message_qos;
}

/* ------------------------------------------------------- */
void taReadyForSendMessage(TaArea *area, const TaShort *sendto_sids,
	int n_sendto_sids, int buf_size, const TaMessageQos *qos, 
	TaBool *is_ready)
{
	int i, n, n_all;
	TaStation *station;
	TaBool sids_all, multi_cast;
	*is_ready = _True;
	if (area->n_send_messages >= area->qos.messages_array_size) {
		*is_ready = _False;
		return;
	}
	if (buf_size < TA_MIN_RECV_BUFFER_LENGTH)
		buf_size = TA_MIN_RECV_BUFFER_LENGTH;
	if (area->n_send_used + buf_size 
	>= area->qos.send_buffer_size) {
		*is_ready = _False;
		return;
	}
	n = n_all = 0;
	sids_all = (!sendto_sids || sendto_sids[0] == TA_SIDS_ALL);
	multi_cast = (n_sendto_sids <= 0);
	if (sids_all)
		n_sendto_sids = area->n_stations;
	else if (multi_cast)
		n_sendto_sids = -n_sendto_sids;
	for (i = 0; i < n_sendto_sids; i++) {
		if (sids_all)
			station = area->stations[i];
		else if (sendto_sids[i] < 0
		|| sendto_sids[i] >= area->n_stations)
			continue;
		else
			station = area->stations[sendto_sids[i]];
		if (station == area->host_station)
			continue;
		if (station->n_available_for_sending == 0)
			continue;
		++n_all;
		if (station->qos.connect_on_request != CONNECT_READY_NOCHECK) {
			int j;
			TaBool bconnected = _False;
			for (j = 0; j < station->n_channels; j++) {
				TaChannel *channel = &station->channels[j];
				if (!channel->is_enabled
				|| !(channel->trans_flags & TA_TRANS_SEND))
					continue;
				if (channel->link_stat == TA_LINK_OK)
					bconnected = _True;
				else if (channel->flags & TA_FLAGS_STREAM
				|| channel->flags & TA_FLAGS_MESSAGE)
					taTryConnectChannel(area, station, j);
			}
			if (!bconnected)
				continue;
		}
		if (station->qos.ping_mask == 0) {
			++n;	// seems all available
		} else if (qos->all_channels_ready) {
			if (station->n_ready_for_sending 
			== station->n_available_for_sending)
				++n;
		} else {
			if (station->n_ready_for_sending > 0)
				++n;
		}
		if (!qos->all_sids_ready) {
			if (n)
				return;
		}
	}
	if (n < n_all)
		*is_ready = _False;
}

/* ------------------------------------------------------- */
int taSetMessage(TaArea *area, const void *buffer, int buffer_length, 
	TaULong msg_type, TaUShort dst_oid, TaUShort iface_code,
	const TaShort *sendto_sids, int n_sendto_sids, 
	TaUShort mid_resp, const TaMessageQos *qos)
{
	TaSRPSHeader *srps_header;
	TaMessage *message;
	int i, blen;
	TaBool sids_all, multi_cast, with_lock;

	if (!area || !area->send_messages || !area->send_buffer 
	|| !area->host_station)
		return TA_ERR_ARGUMENT;
	
	if (area->n_send_messages >= area->qos.messages_array_size) {
		if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file,
			"* out of messages_array_size:%d, n_send:%d\n",
				area->qos.messages_array_size,
				area->n_send_messages);
		return TA_ERR_OUT_OF_MEMORY;
	}
	blen = (SOAP_MSG(mid_resp) ? buffer_length
		: sizeof(TaSRPSHeader) + buffer_length);
	if (area->n_send_used + blen >= area->qos.send_buffer_size) {
		if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file,
			"* out of send_buffer_size:%d, n_used:%d+buflen:%d\n",
				area->qos.send_buffer_size,
				area->n_send_used, blen);
		return TA_ERR_OUT_OF_MEMORY;
	}
	with_lock = (TaBool)(area->send_station->ttd != NULL
		&& area == area->send_station->ttd->area);
	if (with_lock)
		pthread_mutex_lock(&area->send_station->ttd->send_mutex);
	message = &area->send_messages[area->n_send_messages];
	message->buffer = &area->send_buffer[area->n_send_used];
	message->buffer_length = blen;
	message->basic_length = (TaUShort) message->buffer_length;
	if (area->current_mid == TA_MAX_MID)
		message->mid = area->current_mid = 1;
	else
		message->mid = ++area->current_mid;
	message->mid_resp = mid_resp;
	message->src_sid = area->host_station->sid;

	sids_all = (!sendto_sids || sendto_sids[0] == TA_SIDS_ALL);
	multi_cast = (n_sendto_sids <= 0);
	message->n_sendto_sids = n_sendto_sids;
	if (sids_all) {
		message->dst_sid = TA_SIDS_ALL;
		message->sendto_sids[0] = TA_SIDS_ALL;
	} else {
		if (multi_cast)
			n_sendto_sids = -n_sendto_sids;
		if (n_sendto_sids == 1)
			message->dst_sid = sendto_sids[0];
		else
			message->dst_sid = message->src_sid;
		
		for (i = 0; i < n_sendto_sids && i < TA_MAX_MASK_ADDRS; i++)
			message->sendto_sids[i] = sendto_sids[i];
	}

	message->msg_type = msg_type;
	message->dst_oid = dst_oid;
	message->iface_code = iface_code;
	if (qos)
		message->qos = *qos;
	else
		message->qos = g_message_qos;
	message->t_set = area->t_crtu;
	if (message->qos.is_redundant) {
		message->n_refs = 2;
		area->with_second = _True;
	} else
		message->n_refs = 1;

	if (SOAP_MSG(mid_resp)) {
		if (buffer_length > 0)
			memcpy(message->buffer, buffer, buffer_length);
	} else {
		srps_header = (TaSRPSHeader *) message->buffer;
		memset(srps_header, 0, sizeof(TaSRPSHeader));
		strncpy(srps_header->protname, "SRPS", 4);
		srps_header->major_version = TA_MAJOR_VERSION;
		srps_header->length = (TaULong)taH2N32(message->buffer_length);
		srps_header->basic_length = 
			(TaUShort)taH2N16(message->basic_length);
		srps_header->mid = (TaUShort)taH2N16(area->current_mid);
		srps_header->mid_resp = (TaUShort)taH2N16(mid_resp);
		srps_header->src_sid = taH2N16(message->src_sid);
		srps_header->spec_bits = TA_USER_MSG_TYPE;

		srps_header->dst_sid = taH2N16(message->dst_sid);
		srps_header->aid = taH2N32(area->aid);
		srps_header->msg_type = (TaULong) taH2N32(msg_type);
		srps_header->dst_oid = (TaUShort) taH2N16(dst_oid);
		srps_header->iface_code = (TaUShort) taH2N16(iface_code);

		if (buffer_length > 0)
			memcpy(&srps_header[1], buffer, buffer_length);
	}

	if (ta_trace_level & TA_TRAL_MESSAGE)
		fprintf(ta_file, 
	"set_msg mid:%d src_sid:%d dst_sid:%d n_sendto_sids:%d len:%d/%d\n",
			(int)message->mid, (int)message->src_sid, 
			(int)message->dst_sid, message->n_sendto_sids, 
			message->basic_length, message->buffer_length);
	if ((ta_trace_level & TA_TRAL_MESSAGE_DATA) && !SOAP_MSG(mid_resp))
		dumpMessage(message, '>', 30);

	area->n_send_used += message->buffer_length;
	if (area->n_send_used % 16)
		area->n_send_used += (16-area->n_send_used % 16);
	area->n_send_messages++;
	i = message->mid;
	if (with_lock)
		pthread_mutex_unlock(&area->send_station->ttd->send_mutex);
	return i;
}

/* ------------------------------------------------------- */
int taAddToMessage(TaArea *area, const void *buffer, int buffer_length,
	TaULong msg_type, TaUShort dst_oid, TaUShort iface_code)
{
	TaMessage *message;
	TaSRPSHeader *srps_header;
	TaBool with_lock;
	int rc;

	if (!area || !area->send_messages || !area->send_buffer 
	|| !area->host_station || area->n_send_messages <= 0)
		return TA_ERR_ARGUMENT;
	
	message = &area->send_messages[area->n_send_messages-1];
	if (area->n_send_used + buffer_length 
	>= area->qos.send_buffer_size) {
		if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file,
			"* add out of send_buffer_size:%d, n_used:%d+buflen:%d\n",
				area->qos.send_buffer_size,
				area->n_send_used, buffer_length);
		return TA_ERR_OUT_OF_MEMORY;
	}
	if (msg_type != message->msg_type
	||  dst_oid  != message->dst_oid
	|| iface_code!= message->iface_code)
		return TA_ERR_WRONG_DATA;
	with_lock = (TaBool)(area->send_station->ttd != NULL
		&& area == area->send_station->ttd->area);
	if (with_lock)
		pthread_mutex_lock(&area->send_station->ttd->send_mutex);
	memcpy(message->buffer+message->buffer_length,
		buffer, buffer_length);
	area->n_send_used = (int)(message->buffer+message->buffer_length
		+buffer_length - area->send_buffer);
	if (area->n_send_used % 16)
		area->n_send_used += (16-area->n_send_used % 16);
	srps_header = (TaSRPSHeader *) message->buffer;
	/* basic_length persists, buffer_length increments */
	message->buffer_length = message->buffer_length+buffer_length;
	srps_header->length = (TaULong) taH2N32(message->buffer_length);
	rc = area->n_send_messages-1;
	if (with_lock)
		pthread_mutex_unlock(&area->send_station->ttd->send_mutex);
	return rc;
}

/* ------------------------------------------------------- */
int taCopyFromThread(TaArea *area, TaThreadData *ttd)
{
	int rc = 0;
	TaArea *tarea = ttd->area;
	pthread_mutex_lock(&ttd->send_mutex);
	if (area->n_send_messages+tarea->n_send_messages >
	area->qos.messages_array_size) {
		if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file,
		"* copy out of messages_array_size:%d, n:%d+copy:%d\n",
				area->qos.messages_array_size,
				area->n_send_messages, 
				tarea->n_send_messages);
		rc = TA_ERR_OUT_OF_MEMORY;
	} else if (area->n_send_used+tarea->n_send_used >
	area->qos.send_buffer_size) {
		if (ta_trace_level & TA_TRAL_ERRORS)
			fprintf(ta_file,
		"* copy out of send_buffer_size:%d, n:%d+copy:%d\n",
				area->qos.send_buffer_size,
				area->n_send_used, 
				tarea->n_send_used);
		rc = TA_ERR_OUT_OF_MEMORY;
	} else {
		int i;
		TaMessage *msg, *tmsg;
		for (i = 0; i < tarea->n_send_messages; i++) {
			tmsg = &tarea->send_messages[i];
			msg = &area->send_messages[area->n_send_messages];
			memcpy(msg, tmsg, sizeof(TaMessage));
			msg->buffer = &area->send_buffer[area->n_send_used];
			memcpy(msg->buffer, tmsg->buffer, tmsg->buffer_length);
			++area->n_send_messages;
			area->n_send_used += tmsg->buffer_length;
		}
		tarea->n_send_messages = 0;
		tarea->n_send_used = 0;
	}
	pthread_mutex_unlock(&ttd->send_mutex);
	return rc;
}

/* ------------------------------------------------------- */
int taSendMessages(TaArea *area, TaBool is_second)
{
	int i, j, rc = 0, k;
	TaBool sids_all, multi_cast;
	if (!area || !area->host_station)
		return TA_ERR_ARGUMENT;
	if (area->t_last_msg >= area->t_crtu)
		return 0;	// not ready yet
	area->t_last_msg = area->t_crtu;

	// get messages from threads
	if (area->send_station->ttd)
		for (i = 0; i < area->n_stations; i++)
			if (area->stations[i] != area->host_station
			&& area->stations[i]->ttd
			&& area->stations[i]->ttd->area->n_send_messages > 0) {
				rc = 
				taCopyFromThread(area, area->stations[i]->ttd);
				if (rc < 0)
					return rc;
			}

	for (i = 0; i < area->n_send_messages; i++) {
		TaMessage *message, tmp_message;
		TaChannel *channel;
		TaULong msk;
		message = &area->send_messages[i];
		if (message->n_refs-- <= 0)
			continue;
		// to be clarified more
		sids_all = (message->sendto_sids[0] == TA_SIDS_ALL);
		multi_cast = (message->n_sendto_sids <= 0);
		// send all as multicast
		if (sids_all) {
			TaBool bdone = _False, bother = _False;
			for (k = 0; k < area->send_station->n_channels; k++) {
				channel = &area->send_station->channels[k];
				if (!channel->is_enabled)
					continue;
				if (multi_cast &&
				(channel->flags & TA_FLAGS_MCAST)) {
					rc |=
					taSendByChannel(area, message->buffer, 
						message->buffer_length, channel);
					//area->host_station->t_send = area->t_crtu;
					bdone = _True;
				} else
					bother = _True;
			}
			// to be clarified 
			if (bdone && !bother)
				continue;
			// othewise create message mask
			tmp_message = *message;
			message = &tmp_message;
			msk = 1;
			for (j = k = 0; j < area->n_stations; j++, msk = msk<<1) {
				TaStation *station;
				station = area->stations[j];
				if (!station || station == area->host_station
				/*|| !station->is_online */
				/*|| station->is_disconnected*/)
					continue;
				message->sendto_sids[k] = station->sid;
				message->mask |= msk;
				++k;
				if (k >= TA_MAX_MASK_ADDRS) {
					rc |= TA_ERR_OUT_OF_MEMORY;
					break;
				}
			}
			message->n_sendto_sids = k;
			if (ta_trace_level & TA_TRAL_MESSAGE)
				fprintf(ta_file, "send_msg_mask=%x mid:%d\n", 
					(int)message->mask, (int)message->mid);
		}
		rc |= taSendMessage(area, message);
	}
	return rc;
}

/* ------------------------------------------------------- */
int taSendMessage(TaArea *area, TaMessage *message)
{
	int j, k, rc = 0, irc, n_sendto_sids;
	TaBool is_mcast_done[TA_N_CHANNELS], first_done;
	TaBool sids_all, multi_cast;
	memset(is_mcast_done, 0, TA_N_CHANNELS*sizeof(TaBool));
	n_sendto_sids = message->n_sendto_sids;
	if (n_sendto_sids < 0)
		n_sendto_sids = -n_sendto_sids;
	// use the actual sendto_sids list
	for (j = 0; j < n_sendto_sids; j++) {
		TaStation *station;
		TaChannel *channel;
		if (message->sendto_sids[j] < 0
		|| message->sendto_sids[j] >= area->n_stations)
			continue;
		station = area->stations[message->sendto_sids[j]];
		if (!station || station->is_user_locked 
		|| station == area->host_station)
			continue;

		first_done = _False;
		for (k = irc = 0; k < station->n_channels; k++) {
			channel = &station->channels[k];
			if (!channel->is_enabled
			|| !(channel->trans_flags & TA_TRANS_SEND)
			|| channel->link_stat != TA_LINK_OK)
				continue;
			if (message->qos.send_once & first_done)
				continue;
			if (message->qos.send_mask) {
				if (!(message->qos.send_mask & 1<<channel->ctype))
					continue;
			}
			// filter by reliability
			if (message->qos.reliability == TA_PROT_RELIABLE) {
				if (channel->flags & TA_FLAGS_DGRAM)
					continue;
			} else if (message->qos.reliability == TA_PROT_UNRELIABLE) {
				if (!(channel->flags & TA_FLAGS_DGRAM))
					continue;
			} else if (message->qos.reliability == TA_PROT_MCAST) {
				if (!(channel->flags & TA_FLAGS_MCAST))
					continue;
			}
			// filter by send strategy
			if (channel->flags & TA_FLAGS_MCAST) {
				if (!is_mcast_done[k]) {
					irc |=
					taSendByChannel(area, message->buffer,
						message->buffer_length, channel);
					//area->host_station->t_send = area->t_crtu;
					is_mcast_done[k] = _True;
				}
			} else
				irc |=
				taSendByChannel(area, message->buffer,
					message->buffer_length, channel);
			if (irc > 0)
				first_done = _True;
		}
		if (irc) {
			station->t_send = area->t_crtu;
			if (message->mid != 0)
				station->t_mid_send = area->t_crtu;
		}
		rc |= irc;
	}
	return rc;
}

/* ------------------------------------------------------- */
int taSendPingMessage(TaArea *area, TaStation *station)
{
	TaSRPSHeader *srps_header = (TaSRPSHeader *) g_ping_message;
	TaMessage message;
	int rc;

	if (!station->qos.ping_mask)
		return 0;
	memset(&message, 0, sizeof(TaMessage));
	if (!srps_header->spec_bits) {
		memset(srps_header, 0, sizeof(TaSRPSHeader));
		strncpy(srps_header->protname, "SRPS", 4);
		srps_header->major_version = TA_MAJOR_VERSION;
		srps_header->spec_bits = TA_PING_MSG_TYPE;
		srps_header->length = (TaULong) taH2N32(sizeof(TaSRPSHeader));
	}
	srps_header->aid = taH2N32(area->aid);
	message.mid = srps_header->mid = 0;
	message.src_sid = area->host_station->sid;
	srps_header->src_sid = taH2N16(message.src_sid);
	message.dst_sid = ta_addr_all;
	srps_header->dst_sid = taH2N16(message.dst_sid);
	message.msg_type = srps_header->msg_type = 0;
	message.dst_oid = srps_header->dst_oid = 0;
	message.iface_code = srps_header->iface_code = 0;
	message.sendto_sids[0] = station->sid;
	message.n_sendto_sids = 1;
	if ((station->qos.enable_mask & station->qos.ping_mask)
	!= station->qos.enable_mask)
		message.qos.send_mask = station->qos.enable_mask 
			& station->qos.ping_mask;
	srps_header->spec_bits = station->recv_bits;
	message.buffer = (char *) srps_header;
	message.buffer_length = sizeof(TaSRPSHeader);
	rc = taSendMessage(area, &message);
	if (ta_trace_level & TA_TRAL_MESSAGE_DATA)
		fprintf(ta_file, "send_ping_msg tosid:%d rc:%d\n", 
			(int)station->sid, rc);
	return rc;

}

/* ------------------------------------------------------- */
int taSendPhaseMessage(TaArea *area, TaStation *station, TaUShort phase)
{
	TaSRPSHeader *srps_header = (TaSRPSHeader *) g_phase_message;
	TaMessage message;
	int rc;
	memset(&message, 0, sizeof(TaMessage));
	if (!srps_header->spec_bits) {
		memset(srps_header, 0, sizeof(TaSRPSHeader));
		strncpy(srps_header->protname, "SRPS", 4);
		srps_header->major_version = TA_MAJOR_VERSION;
		srps_header->length = (TaULong) taH2N32(sizeof(TaSRPSHeader));
		srps_header->spec_bits = TA_PHASE_MSG_TYPE;
	}
	srps_header->aid = taH2N32(area->aid);
	message.mid = srps_header->mid = 0;
	message.src_sid = area->host_station->sid;
	srps_header->src_sid = taH2N16(message.src_sid);
	message.dst_sid = ta_addr_all;
	srps_header->dst_sid = taH2N16(message.dst_sid);
	message.msg_type = srps_header->msg_type = 0;
	message.dst_oid = phase;
	srps_header->dst_oid = (TaUShort) taH2N16(message.dst_oid);
	message.iface_code = srps_header->iface_code = 0;
	
	message.sendto_sids[0] = station->sid;
	message.n_sendto_sids = 1;
	message.qos.send_once = _True;
	srps_header->spec_bits = 0;
	message.buffer = (char *) srps_header;
	message.buffer_length = sizeof(TaSRPSHeader);
	rc = taSendMessage(area, &message);
	if (ta_trace_level & TA_TRAL_MESSAGE_DATA)
		fprintf(ta_file, "send_phase_msg phase:%d rc:%d\n", 
			phase, rc);
	return rc;
}

/* ------------------------------------------------------- */
int taSendPingMessages(TaArea *area)
{
	TaStation *station;
	TaChannel *channel;
	int i, k, rc = 0;

	// accept from host station
	for (k = 0; k < area->host_station->n_channels; k++) {
		channel = &area->host_station->channels[k];
		if (!channel->is_enabled
		|| !(channel->trans_flags & TA_TRANS_RECV))
			continue;
		if (channel->flags & TA_FLAGS_STREAM
		&&  channel->flags & TA_FLAGS_LISTEN)
			taTryAcceptChannels(area, k);
	}
	// ping mcast
	station = area->send_station;
	for (k = 0; k < station->n_channels; k++) {
		channel = &station->channels[k];
		if (!channel->is_enabled
		//|| !(channel->flags & TA_FLAGS_DGRAM)
		//|| !(channel->flags & TA_FLAGS_MCAST)
		|| !(channel->trans_flags & TA_TRANS_SEND))
			continue;
		if (area->t_crtu > TA_TP(station->qos.tp_ping_timeout)
		+ station->t_send) {
			station->t_send = area->t_crtu;
			rc |= taSendPingMessage(area, station);
		}
	}

	for (i = 0; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!station)
			continue;
		if (station == area->host_station || station->is_user_locked)
			continue;
		// connect/disconnect if needed
		for (k = 0; k < station->n_channels; k++) {
			channel = &station->channels[k];
			if (!channel->is_enabled
			|| !(channel->trans_flags & TA_TRANS_SEND))
				continue;
			if (channel->flags & TA_FLAGS_STREAM) {
				if (channel->flags & TA_FLAGS_LISTEN) {
					// timed close accepted connection
					if (channel->link_stat == TA_LINK_OK 
					&& !channel->is_receiving2) {
						if (ta_trace_level & TA_TRAL_TRANSPORT)
							fprintf(ta_file,
					"? SERVER taShutdownChannelData sid:%d sd:%d\n",
								i, channel->sd);
						taShutdownChannelData(channel);
					}
				} else {
					// reconnect seems bad idea, do at server side
					if (channel->link_stat == TA_LINK_OK
					&& !channel->is_receiving2) {
						if (ta_trace_level & TA_TRAL_TRANSPORT)
							fprintf(ta_file,
					"? CLIENT taShutdownChannelData sid:%d sd:%d\n",
								i, channel->sd);
						taShutdownChannelData(channel);
					}
					if (channel->link_stat != TA_LINK_OK
					&& (station->qos.connect_on_request != CONNECT_WHEN_REQUEST)
					&& area->t_crtu > TA_TP(channel->qos.tp_connect_timeout)
					+ channel->t_crecv)
						taTryConnectChannel(area, station, k);
				}
			} else if (channel->link_stat != TA_LINK_OK
			&& channel->flags & TA_FLAGS_MESSAGE
			&& (station->qos.connect_on_request != CONNECT_WHEN_REQUEST)
			&& area->t_crtu > TA_TP(channel->qos.tp_connect_timeout)
			+ channel->t_crecv)
				taTryConnectChannel(area, station, k);
		}
		if (area->t_crtu > TA_TP(station->qos.tp_ping_timeout)
		+ station->t_send) {
			station->t_send = area->t_crtu;
			rc |= taSendPingMessage(area, station);
		}
	}

	return rc;
}

/* ------------------------------------------------------- */
int taSendPhaseMessages(TaArea *area, TaUShort phase)
{
	int rc = 0, i, k;
	TaStation *station;
	if (!area->host_station)
		return 0;

	for (k = 0; k < area->host_station->n_channels; k++) {
		TaChannel *channel;
		channel = &area->host_station->channels[k];
		if (!channel->is_enabled)
			continue;
		if (!(channel->flags & TA_FLAGS_MCAST))
			continue;
		return taSendPhaseMessage(area, area->send_station, phase);
	}

	for (i = 0; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!station)
			continue;
		if (station == area->host_station || station->is_user_locked)
			continue;
		rc |= taSendPhaseMessage(area, area->send_station, phase);
	}
	return rc;
}

/* ------------------------------------------------------- */
int taGetMessage(TaArea *area, int index, TaMessage *message)
{
	int rc = 1;
	if (!area->send_station || !area->send_station->ttd) {
		if (index < 0 || index >= area->n_recv_messages)
			rc = TA_NOT_FOUND;
		else
			*message = area->recv_messages[index];
	} else {
		TaThreadData *ttd = area->send_station->ttd;
		pthread_mutex_lock(&ttd->recv_mutex);
		if (index < 0 || index >= ttd->area->n_recv_messages)
			rc = TA_NOT_FOUND;
		else
			*message = ttd->area->recv_messages[index];
		pthread_mutex_unlock(&ttd->recv_mutex);
	}
	return rc;
}

/* ------------------------------------------------------- */
int taPutRecvToThread(TaMessage *message, TaStation *station)
{
	int rc = 0;
	TaArea *area = station->ttd->area;
	TaMessage *msg;
	pthread_mutex_lock(&station->ttd->recv_mutex);
	msg = &area->recv_messages[area->n_recv_messages];
	if (message->buffer_length >= area->qos.recv_buffer_size
	- area->n_recv_used)
		rc = -1;
	else {
		memcpy(msg, message, sizeof(TaMessage));
		msg->buffer = &area->recv_buffer[area->n_recv_used];
		memcpy(msg->buffer, message->buffer, message->buffer_length);
		++area->n_recv_messages;
		area->n_recv_used += message->buffer_length;
	}
	pthread_mutex_unlock(&station->ttd->recv_mutex);
	return 0;
}

/* ------------------------------------------------------- */
int taRecvMessages(TaArea *area, TaBool is_second)
{
	int rc = 0, i, k;
	TaBool all_disconnected, is_disconnected;
	TaStation *station;
	TaChannel *channel;

	if (!area && !area->host_station)
		return TA_ERR_ARGUMENT;
	// clear all previous messages and buffer
	if (!is_second) {
		area->n_recv_messages = 0;
		area->n_recv_used = 0;
		area->n_send_messages = 0;
		area->n_send_used = 0;
		area->with_second = _False;
	}
	// has some messages
	all_disconnected = _True;
	station = area->host_station;
	for (k = 0; k < station->n_channels; k++) {
		channel = &station->channels[k];
		if (!channel->is_enabled
		|| !(channel->trans_flags & TA_TRANS_RECV)
		|| channel->link_stat != TA_LINK_OK)
			continue;
		rc |= 
		taRecvChannelMessages(area, channel, station->sid);
	}
	if (rc > 0)
		all_disconnected = _False;
	for (i = 0; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!station || station->is_user_locked)
			continue;
		// self check for online
		if (station == area->host_station) {
			station->t_recv = area->t_crtu;
			station->is_online = _True;
			continue;
		}
		// get messages from other ports
		is_disconnected = _True;
		station->recv_bits = 0;
		for (k = 0; k < station->n_channels; k++) {
			channel = &station->channels[k];
			if (!channel->is_enabled
			|| !(channel->trans_flags & TA_TRANS_RECV))
				continue;
			if (channel->flags & TA_FLAGS_MEMORY) {
				if (channel->link_stat != TA_LINK_OK)
					taTryConnectChannel(area, station, k);
			}
			if (channel->link_stat != TA_LINK_OK)
				continue;
			//if (channel->flags & TA_FLAGS_STREAM
			//||  channel->flags & TA_FLAGS_MEMORY)
			rc |= taRecvChannelMessages(area, channel, station->sid);
			is_disconnected = _False;

			if (area->t_crtu > 
			TA_TP(channel->qos.tp_disconnect_timeout/2)
			+ channel->t_crecv)
				channel->is_receiving = _False;
			else
				channel->is_receiving = _True;
			if (area->t_crtu > 
			TA_TP(channel->qos.tp_disconnect_timeout)
			+ channel->t_crecv)
				channel->is_receiving2 = _False;
			else
				channel->is_receiving2 = _True;
			if (channel->ctype < 16) {
				if (channel->is_receiving)
					station->recv_bits |= 1<<channel->ctype;
			}
		}
		station->is_disconnected = all_disconnected && is_disconnected;
		// check for online
		if (area->t_crtu > TA_TP(station->qos.tp_discovery_timeout)
		+ station->t_recv)
			station->is_online = _False;
		else
			station->is_online = _True;
	}
	return rc;
}

/* ------------------------------------------------------- */
int taRecvParseSrpsMessage(TaArea *area, TaChannel *channel, 
	TaMessage *message, TaShort sid)
{
	int done = 0;
	TaSRPSHeader *srps_header, sh;
	TaStation *station;
	TaLong len, aid;

	srps_header = (TaSRPSHeader *) message->buffer;
	len = taN2H32(srps_header->length);
	aid = taN2H32(srps_header->aid);
	if (strncmp(srps_header->protname, "SRPS", 4)
	|| srps_header->major_version != TA_MAJOR_VERSION
	|| aid != area->aid || len < sizeof(TaSRPSHeader))
		return -1;
	sh.spec_bits = srps_header->spec_bits;
	sh.iface_code = (TaUShort)taN2H16(srps_header->iface_code);
	sh.msg_type = (TaULong)taN2H32(srps_header->msg_type);
	sh.dst_oid = (TaUShort)taN2H16(srps_header->dst_oid);
	sh.basic_length = (TaUShort)taN2H16(srps_header->basic_length);
	sh.length = len;
	sh.mid = (TaUShort)taN2H16(srps_header->mid);
	sh.mid_resp = (TaUShort)taN2H16(srps_header->mid_resp);
	sh.aid = aid;
	sh.src_sid = taN2H16(srps_header->src_sid);
	sh.dst_sid = taN2H16(srps_header->dst_sid);
	if (channel->qos.assign_sids) {
		sh.src_sid = sid;
		sh.dst_sid = area->host_station->sid;
	}

	// or any other reasons to discard the message
	// proper sid
	if (sh.src_sid < 0
	|| sh.src_sid >= area->n_stations)
		return -2;
	if (sh.dst_sid == ta_addr_all || sh.dst_sid == sh.src_sid)
		;
	else if (sh.dst_sid != area->host_station->sid)
		return -3;	// wrong dst
	if (channel->flags & TA_FLAGS_STREAM
	&& channel->flags & TA_FLAGS_LISTEN) {
		if (sh.src_sid == sid)
			channel->sid_acked = _True;
		else if (channel->sid_acked)
			return -4;
		else {
			// change accepted sd
			int k, sd = channel->sd;
			station = area->stations[sh.src_sid];
			if (station)
				for (k = 0; k < station->n_channels; k++) {
					if (station->channels[k].ctype 
					!= channel->ctype)
						continue;
					if (station->channels[k].sid_acked)
						continue;
					channel->sd = station->channels[k].sd;
					channel->sid_acked = _True;
					station->channels[k].sd = sd;
					break;
				}
			if (!channel->sid_acked)
				return -5;
		}
	}
	// avoid self-sending, exclude locked
	station = area->stations[sh.src_sid];
	if (!station || station == area->host_station
	|| station->is_user_locked)
		return -6;
	if (!taCheckChannels(channel, 
	&station->channels[channel->ctype]))
		return -7;	// other addr not applicable
	station->t_recv = area->t_crtu;
	done = 1;
 	if (TA_MSG_TYPE(sh.spec_bits) == TA_PHASE_MSG_TYPE) {
		// special phase correction logic
		taPhaseRecvCorrect(area, sh.dst_oid);
		if (ta_trace_level & TA_TRAL_MESSAGE)
			fprintf(ta_file,
			"recv_phase_msg type:%d len:%d\n", 
				(int)sh.msg_type, sh.length);
	} else if (TA_MSG_TYPE(sh.spec_bits)
	!= TA_USER_MSG_TYPE)
		return done;
	else if (sh.mid == 0 
	|| station->last_mid == sh.mid) // is it correct?
		return done;	// discard pings or repeats
	station->t_mid_recv = area->t_crtu;
	message->src_sid = sh.src_sid;
	message->dst_sid = sh.dst_sid;
	message->mid = station->last_mid = sh.mid;
	message->mid_resp = sh.mid_resp;
	message->msg_type = sh.msg_type;
	message->dst_oid = sh.dst_oid;
	message->iface_code = sh.iface_code;
	message->basic_length = sh.basic_length;
	message->buffer_length = (int)sh.length;
	return done+1;
}

/* ------------------------------------------------------- */
int taRecvParseSoapMessage(TaArea *area, TaChannel *channel,
	TaMessage *message, TaShort sid)
{
	char *p, interface_name[TA_NAME_LENGTH], item_name[TA_NAME_LENGTH];
	int i, l;
	message->mid = TA_SOAP_MID;
	p = strstr(message->buffer, "HTTP/1.1");
	if (!p)
		;
	else if (p == message->buffer) {
		if (strstr(message->buffer, "soap:Envelope "))
			message->mid = TA_SOAP_11_CLIENT;
		else if (strstr(message->buffer, "soap12:Envelope "))
			message->mid = TA_SOAP_12_CLIENT;
		else if (strstr(message->buffer, "text/xml"))
			message->mid = TA_SOAP_POST_CLIENT;
		else
			return TA_MSG_INCOMPATIBLE;
	} else if (strstr(message->buffer, "SOAPAction"))
		message->mid = TA_SOAP_11_MID;
	else if (strstr(message->buffer, "application/soap+xml"))
		message->mid = TA_SOAP_12_MID;
	else if (strstr(message->buffer, "GET ") == message->buffer)
		message->mid = TA_SOAP_GET_MID;
	else if (strstr(message->buffer, "POST") == message->buffer)
		message->mid = TA_SOAP_POST_MID;
	message->mid_resp = 0;
	message->src_sid = sid;
	message->dst_sid = area->host_station->sid;
	message->dst_oid = message->mid;
	interface_name[0] = item_name[0] = 0;
	switch (message->mid) {
	case TA_SOAP_11_MID:
		p = strchr(message->buffer, '/');
		if (!p) return TA_ERR_WRONG_DATA;
		++p;
		for (i = 0; p[i] != ' ' && p[i]; i++)
			interface_name[i] = p[i];
		interface_name[i] = 0;
		p = strstr(message->buffer, "SOAPAction");
		if (!p) return TA_ERR_WRONG_DATA;
		p = strchr(p, '\n');
		if (!p) return TA_ERR_WRONG_DATA;
		while (*p != '/') --p;
		++p;
		for (i = 0; p[i] && isalnum(p[i]) || p[i] == '_'; i++)
			item_name[i] = p[i];
		item_name[i] = 0;
		message->msg_type = (TaULong) taCrc32(interface_name, 0, 32);
		message->iface_code = (TaUShort) taCrc32(item_name, 0, 16);
		break;
	case TA_SOAP_12_MID:
		p = strchr(message->buffer, '/');
		if (!p) return TA_ERR_WRONG_DATA;
		++p;
		for (i = 0; p[i] != ' ' && p[i]; i++)
			interface_name[i] = p[i];
		interface_name[i] = 0;
		p = strstr(message->buffer, "soap12:Body");
		if (!p) return TA_ERR_WRONG_DATA;
		p = strchr(p, '<');
		if (!p) return TA_ERR_WRONG_DATA;
		++p;
		for (i = 0; p[i] != '\n' && p[i] != ' ' && p[i]; i++)
			item_name[i] = p[i];
		item_name[i] = 0;
		message->msg_type = (TaULong) taCrc32(interface_name, 0, 32);
		message->iface_code = (TaUShort) taCrc32(item_name, 0, 16);
		break;
	case TA_SOAP_GET_MID:
	case TA_SOAP_POST_MID:
		p = strchr(message->buffer, '/');
		if (!p) return TA_ERR_WRONG_DATA;
		++p;
		for (i = 0; p[i] != ' ' && p[i] != '/' && p[i]; i++)
			interface_name[i] = p[i];
		interface_name[i] = 0;
		p = strchr(&p[i], '/');
		++p;
		for (i = 0; p[i] != '?' && p[i] != ' ' && p[i]; i++)
			item_name[i] = p[i];
		item_name[i] = 0;
		message->msg_type = (TaULong) taCrc32(interface_name, 0, 32);
		message->iface_code = (TaUShort) taCrc32(item_name, 0, 16);
		break;
	case TA_SOAP_11_CLIENT:
	case TA_SOAP_12_CLIENT:
		p = strstr(message->buffer, "Result>");
		if (!p) return TA_ERR_WRONG_DATA;
		l = (int)p - (int)message->buffer;
		for (i = l-1; 
		i > 0 && message->buffer[i] != '<'; i--)
			;
		if (message->buffer[i] != '<')
			return TA_ERR_WRONG_DATA;
		l = l - (i+1);
		if (l >= TA_NAME_LENGTH)
			l = TA_NAME_LENGTH-1;
		strncpy(item_name, &message->buffer[i+1], l);
		item_name[l] = 0;
		message->msg_type = TA_SOAP_MID;
		message->iface_code = (TaUShort) taCrc32(item_name, 0, 16);
		message->mid_resp = TA_SOAP_MID;
		break;
	case TA_SOAP_GET_CLIENT:
	case TA_SOAP_POST_CLIENT:
		message->msg_type = TA_SOAP_MID;
		message->iface_code = TA_SOAP_POST_CLIENT;
		message->mid_resp = TA_SOAP_MID;
		break;
	case TA_SOAP_MID:
		message->msg_type = TA_SOAP_MID;
		message->iface_code = TA_SOAP_MID;
		message->mid_resp = TA_SOAP_MID;
		break;
	}
	p = strstr(message->buffer, "Content-Length:");
	if (p)
		sscanf(p+15, "%d", &message->basic_length);
	area->stations[sid]->t_recv = area->t_crtu;
	area->stations[sid]->t_mid_recv = area->t_crtu;
	if (ta_trace_level & TA_TRAL_MESSAGE_DATA) {
		fprintf(ta_file,"--- IN SOAP len:%d ---\n", message->buffer_length);
		fprintf(ta_file,"%s\n", message->buffer);
		fprintf(ta_file,"--- END OF SOAP '%s' '%s' ---\n", interface_name,
			item_name);
	}
	return 2;
}

/* ------------------------------------------------------- */
int taRecvChannelMessages(TaArea *area, TaChannel *channel, TaShort sid)
{
	int rc = 0, irc, remains_in_buf = 0, done = 0;
	// receive which can be received
	if (!channel->is_enabled) 
		return 0;
	do {
		TaMessage *message;
		TaStation *station;
		message = &area->recv_messages[area->n_recv_messages];
		message->buffer = &area->recv_buffer[area->n_recv_used];
		message->buffer_length = area->qos.recv_buffer_size
			- area->n_recv_used;
		if (message->buffer_length < TA_MIN_RECV_BUFFER_LENGTH) {
			if (ta_trace_level & TA_TRAL_ERRORS)
				fprintf(ta_file,
				"* out of recv_buffer_size:%d, used:%d\n",
					area->qos.recv_buffer_size,
					area->n_recv_used);
			return TA_ERR_OUT_OF_MEMORY;
		}
		if (remains_in_buf)
			rc = remains_in_buf;
		else
			rc = taRecvByChannel(area, message->buffer, 
				message->buffer_length, channel);
		if (rc <= 0)
			break;
		if (rc == 1)
			continue;

		irc = TA_ERR_WRONG_DATA;
		switch (area->stations[sid]->qos.prot_format) {
		case TA_PROT_FORMAT_SRPS:
			irc =
			taRecvParseSrpsMessage(area, channel, 
				message, sid);
			if (irc > 0) {
				if (rc > message->buffer_length) {
				//is sh.length! if (rc > (int)sh.length)
					remains_in_buf = rc - 
						message->buffer_length;
					rc = message->buffer_length;
				} else if (irc > 1) {
					int rc2, ctr;
					remains_in_buf = 0; ctr = 0;
					while ((rc < message->buffer_length) && (ctr < 10)) {
						int rc2 =
						taRecvByChannel(area, message->buffer+rc,
						message->buffer_length-rc, channel);
						if (ta_trace_level & TA_TRAL_MESSAGE)
							fprintf(ta_file,
							"large message, len:%d+%d expected:%d\n", 
								rc, rc2, message->buffer_length);
						if (rc2 == 0) {
							taSleep(500000);	// 50 msec
							++ctr;
						} else
							rc += rc2;
					}
					if (rc < message->buffer_length) {
						if (ta_trace_level & TA_TRAL_ERRORS)
							fprintf(ta_file,
							"* large message, received:%d expected:%d\n", 
								rc, message->buffer_length);
						rc = 0;
					}
				}
				//area->n_recv_used += rc;	// old place?
			} else
				irc = 0;
			break;
		case TA_PROT_FORMAT_SOAP:
			message->basic_length = 0;
			message->buffer[rc] = 0;
			++rc;
			irc =
			taRecvParseSoapMessage(area, channel,
				message, sid);
			if (ta_trace_level & TA_TRAL_MESSAGE_DATA && irc != 2) {
				fprintf(ta_file,"??? IN SOAP rc:%d ???\n", irc);
				fprintf(ta_file,"%s\n", message->buffer);
				fprintf(ta_file,"??? END OF SOAP ???\n");
			}
			/*irc |= taUnquoteBuffer(message->buffer, rc, message->basic_length);*/
			message->basic_length = rc;
			break;
		}
		if (rc == 0)
			break;
		if ((ta_trace_level & TA_TRAL_MESSAGE_DATA)
		&& area->stations[sid]->qos.prot_format == TA_PROT_FORMAT_SRPS)
			dumpMessage(message, '<', 30);
		if (irc <= 1) {
			if (ta_trace_level & TA_TRAL_MESSAGE)
				fprintf(ta_file,
				"?msg mid:%d rejected, irc:%d, length:%d\n",
					message->mid, irc, rc);
			if (irc < 0)
				break;
			continue;
		}
		++done;

		message->buffer_length = rc;
		message->t_got = area->t_crtu;
		station = area->stations[message->src_sid];
		if (!station->ttd) {
			area->n_recv_used += rc;
			++area->n_recv_messages;
		} else {
			if (taPutRecvToThread(message, station) < 0) {
				fprintf(ta_file,
				"* out of thread recv_buffer_size:%d, used:%d\n",
					station->ttd->area->qos.recv_buffer_size,
					station->ttd->area->n_recv_used);
				return TA_ERR_OUT_OF_MEMORY;
			}
		}
		if (ta_trace_level & TA_TRAL_MESSAGE)
			fprintf(ta_file, 
			"recv_msg mid:%d src_sid:%d len:%d/%d dst_sid:%d\n",
				(int)message->mid, (int)message->src_sid, 
				message->basic_length, message->buffer_length,
				(int)message->dst_sid);
	} while (rc > 0);
	return done;
}

/*
 * auxilary routines
 */
/* ------------------------------------------------------- */
static void dumpMessage(const TaMessage *message, char fs, int nbytes)
{
	TaSRPSHeader *srps_header;
	int i;
	srps_header = (TaSRPSHeader *) message->buffer;
	fprintf(ta_file, "%cSRPS ver:%d spec:%x iface:%d,%d type:%u,%u oid:%u,%u "
	"l:%u,%u bl:%d,%d mid:%u,%u/%u,%u aid:%d,%d src_sid:%d,%d dst_sid:%d,%d\n",
		fs, srps_header->major_version, (unsigned)srps_header->spec_bits,
		(int)srps_header->iface_code, (int)taN2H16(srps_header->iface_code), 
		(unsigned)srps_header->msg_type, (unsigned)taN2H32(srps_header->msg_type),
		(unsigned)srps_header->dst_oid, (unsigned)taN2H16(srps_header->dst_oid),
		(unsigned)srps_header->length, (unsigned)taN2H32(srps_header->length),
		(int)srps_header->basic_length, (int)taN2H16(srps_header->basic_length),
		(unsigned)srps_header->mid, (unsigned)taN2H16(srps_header->mid),
		(unsigned)srps_header->mid_resp, (unsigned)taN2H16(srps_header->mid_resp),
		(int)srps_header->aid, (int)taN2H32(srps_header->aid),
		(int)srps_header->src_sid, (int)taN2H16(srps_header->src_sid),
		(int)srps_header->dst_sid, (int)taN2H16(srps_header->dst_sid));
	for (i = 0; i < nbytes && i < message->buffer_length; i++) {
		char *buffer = taMessageBody((*message));
		fprintf(ta_file, "%02d%c ", buffer[i], 
			(isalnum(buffer[i]) ? buffer[i] : ' '));
	}
	fprintf(ta_file, "...\n");
}

