/*
 *                       ***** TA PING *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *   based on Mike Muuss, http://www.ping127001.com/pingpage/ping.text
 *   which is public domain
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#ifndef _WINDOWS
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/poll.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#else
#ifdef _MSC_VER
// The following two structures need to be packed tightly, but unlike
// Borland C++, Microsoft C++ does not do this by default.
#pragma pack(1)
#endif

struct icmp {
    BYTE icmp_type;          // ICMP packet type
    BYTE icmp_code;          // Type sub code
    USHORT icmp_cksum;
    USHORT icmp_id;
    USHORT icmp_seq;
};
struct ip {
    BYTE ip_hl:4;           // Length of the header in dwords
    BYTE ip_version:4;         // Version of IP
    BYTE tos;               // Type of service
    USHORT total_len;       // Length of the packet in dwords
    USHORT ident;           // unique identifier
    USHORT flags;           // Flags
    BYTE ttl;               // Time to live
    BYTE proto;             // Protocol number (TCP, UDP etc)
    USHORT checksum;        // IP checksum
    ULONG source_ip;
    ULONG dest_ip;
};

#ifdef _MSC_VER
#pragma pack()
#endif

#define ICMP_ECHO 8
#define ICMP_MINLEN 8

#endif

#define	MAXPACKET	4096	/* max packet size */
#define DATALEN		56

/* PING */
int taInitPing(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel);
int taSendToPing(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel);
int taRecvFromPing(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);
void taCloseSocket(TaChannel *channel);
const char * taPingReadableErrno(int ierrno);

static TaChannelProtocol g_pingprots[] = {
	{"PING", TA_CHANNEL_ICMP, TA_FLAGS_ICMP,
		taInitPing, taCloseSocket, NULL, NULL,
		taSendToPing, taRecvFromPing, taPingReadableErrno}
};

/* ------------------------------------------------------- */
int taInitPingProts()
{
	unsigned i;
	for (i = 0; i < sizeof(g_pingprots)/sizeof(TaChannelProtocol); i++)
	{
		if (taAddChannelProtocol(&g_pingprots[i]) < 0)
			return TA_ERR_ARGUMENT;
	}
	return i;
}

/* ------------------------------------------------------- */
int taInitPing(TaBool is_datahost, TaBool is_send_station, 
	const TaStation *station, TaChannel *channel)
{
	memset(&channel->sa_from, 0, sizeof(struct sockaddr_in));
	channel->sa_from.sin_family = AF_INET;
	channel->sa_from.sin_addr.s_addr = 
		inet_addr(station->qos.hostname);
	if (channel->sa_from.sin_addr.s_addr == (unsigned) -1) {
		struct hostent *hp;
		hp = gethostbyname(station->qos.hostname);
		if (hp)
			memcpy(&channel->sa_from.sin_addr, hp->h_addr,
				hp->h_length);
		else {
			channel->send_errno = TA_ERR_WRONG_DATA;
			return TA_ERR_WRONG_DATA;
		}
	}
	channel->sa_len = sizeof(struct sockaddr_in);
	if (is_send_station) {
		struct protoent *proto;
#ifndef _WINDOWS
        	int fl;
#else
        	u_long v;
#endif
		if ((proto = getprotobyname("icmp")) == NULL) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file, "* icmp: unknown protocol\n");
			return TA_ERR_WRONG_DATA;
		}
		if ((channel->sd = socket(AF_INET, SOCK_RAW, 
		proto->p_proto)) < 0) {
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file, "* icmp: socket: err:%d %s\n",
					errno, strerror(errno));
			return TA_ERR_WRONG_DATA;
		}
#ifndef _WINDOWS
		fl = O_NONBLOCK;
		if (fcntl(channel->sd, F_SETFL, fl) < 0) {
#else
		v = 1;
		if (ioctlsocket(channel->sd, FIONBIO, &v) != 0) {
#endif
			if (ta_trace_level & (TA_TRAL_TRANSPORT|TA_TRAL_ERRORS))
				fprintf(ta_file, "*** NONBLOCK socket  err:%d\n", errno);
			taCloseChannelData(channel);
			return TA_ERR_WRONG_DATA;
		}
#ifdef _WINDOWS
		/*
			if (setsockopt(channel->sd, IPPROTO_IP, IP_TTL, (const char*)&ttl, 
            sizeof(ttl)) == SOCKET_ERROR) {
				if (ta_trace_level & TA_TRAL_TRANSPORT)
					fprintf(ta_file, "*** TTL setsockopt failed: err:%d\n", errno);
				return TA_ERR_WRONG_DATA;
			}
		*/
#endif
	}

	channel->send_id = station->sid;
	channel->recv_id = getpid() & 0xFFFF;
	channel->recv_offset = 0;
	channel->link_stat = TA_LINK_OK;
	return 1;
}

/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
static int in_cksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 ) {
		u_short	u = 0;

		*(u_char *)(&u) = *(u_char *)w ;
		sum += u;
	}

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}

/*
 * 			P I N G E R
 * 
 * Compose and transmit an ICMP ECHO REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */
int taSendToPing(const TaArea *area, const char *buffer, int buffer_length, 
	TaChannel *channel)
{
	static u_char outpack[MAXPACKET];
	register struct icmp *icp = (struct icmp *) outpack;
        int i=0, cc;
	register struct timeval *tp = (struct timeval *) &outpack[8];
	register u_char *datap = &outpack[8+sizeof(struct timeval)];
	TaChannel *dstchannel;

	if (!area->send_station)
		return TA_ERR_ARGUMENT;
	dstchannel = &area->send_station->channels[channel->cid];
	if (!(dstchannel->poll_events & POLLOUT))
		return 0;

	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = channel->send_id;	// id of station
	icp->icmp_id = channel->recv_id;		/* ID */

	cc = DATALEN+8;			/* skips ICMP portion */

	//gettimeofday( tp, &tz );

	for( i=8; i<DATALEN; i++)	/* skip 8 for time */
		*datap++ = i;

	/* Compute ICMP checksum here */
	icp->icmp_cksum = in_cksum( (u_short *)icp, cc );

	/* cc = sendto(s, msg, len, flags, to, tolen) */
	i = sendto( dstchannel->sd, outpack, cc, 0, &channel->sa_from, 
		channel->sa_len );
	channel->send_errno = errno;
	if (i < 0 && channel->send_errno == TA_EC(EWOULDBLOCK))
		i = 0;

	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| i < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (i < 0)
			fprintf(ta_file, "* sendto %s %d bytes: err:%d %s\n",
				channel->hostname, cc, channel->send_errno, 
				strerror(channel->send_errno));
		else if (i > 0)
			fprintf(ta_file, "sendto %s %d bytes sid:%d\n",
				channel->hostname, cc, icp->icmp_seq);
	}
	return i;
}

/* ------------------------------------------------------- */
static int icmp_errors[] = {
	0,	// "Echo Reply",
	0,	// "ICMP 1",
	0,	// "ICMP 2",
	1,	// "Dest Unreachable",
	1,	// "Source Quench",
	0,	// "Redirect",
	0,	// "ICMP 6",
	0,	// "ICMP 7",
	0,	// "Echo",
	0,	// "ICMP 9",
	0,	// "ICMP 10",
	1,	// "Time Exceeded",
	1,	// "Parameter Problem",
	0,	// "Timestamp",
	0,	// "Timestamp Reply",
	0,	// "Info Request",
	0	// "Info Reply"
};

static int checkOutPacket(TaChannel *channel, char *buf, int cc)
{
	struct ip *ip;
	register struct icmp *icp;
	struct timeval tv;
	int hlen;

	ip = (struct ip *) buf;
	hlen = ip->ip_hl << 2;
	if (cc < hlen + ICMP_MINLEN) {
		//	printf("packet too short (%d bytes) from %sn", cc,
		//		inet_ntoa(ntohl(from->sin_addr))); /* DFM */
		return 0;
	}
	cc -= hlen;
	icp = (struct icmp *)(buf + hlen);
	if (icmp_errors[icp->icmp_type])
		return -icp->icmp_type;
	if(icp->icmp_id != channel->recv_id)
		return 0;			/* 'Twas not our ECHO */

	/*
	if (timing) {
		tp = (struct timeval *)&icp->icmp_data[0];
		tvsub( &tv, tp );
		triptime = tv.tv_sec*1000+(tv.tv_usec/1000);
		tsum += triptime;
		if( triptime < tmin )
			tmin = triptime;
		if( triptime > tmax )
			tmax = triptime;
	}
	*/

	return 1;
}

/* ------------------------------------------------------- */
int taRecvFromPing(TaArea *area, char *buffer, int buffer_length, TaChannel *channel)
{
	int rc, i, my_errno = 0;
	u_char	packet[MAXPACKET];
	TaChannel *cchannel = &area->send_station->channels[channel->cid], *dstchannel;
	char clienthost[NI_MAXHOST];
	char clientservice[NI_MAXSERV];

	if (cchannel == channel) {
		// got from send station
		if (!(channel->poll_events & POLLIN))
			return 0;
		channel->sa_msg_len = sizeof(channel->sa_msg);
		memset(&channel->sa_msg, 0, channel->sa_msg_len);
		rc = recvfrom(cchannel->sd, packet, sizeof (packet), 
			0, &channel->sa_msg, &channel->sa_msg_len);
		my_errno = channel->recv_errno = errno;
		if (rc < 0 && channel->recv_errno == TA_EC(EWOULDBLOCK))
			my_errno = channel->recv_errno = rc = 0;
		if (rc == 0)
			return 0;
		if (rc > 0) {
			i = checkOutPacket(channel, (char *)packet, rc);
			if (i < 0)
				rc = channel->recv_errno = i;
			if (i == 0)
				return 1;	// continue
		}
		clienthost[0] = clientservice[0] = 0;
		getnameinfo((struct sockaddr *)&channel->sa_msg,
			channel->sa_msg_len, clienthost,
			sizeof(clienthost), clientservice,
			sizeof(clientservice), NI_NUMERICHOST);
		// set to _False if can't determine dst address
		if (!channel->qos.ping_set_err) {
			if (i < 0) {
				channel->recv_errno = 0;
				return 1;
			}
		}

		channel->recv_offset = 0;
		for (i = 0; i < area->n_stations; i++) {
			TaStation *dstation = area->stations[i];
			TaChannel *dstchannel = 
				&dstation->channels[channel->cid];
			if (channel->sa_msg.sin_addr.s_addr
			!= dstchannel->sa_from.sin_addr.s_addr)
				continue;
			if (dstchannel == channel) {
				channel->recv_offset = rc;
				break;
			}
			dstchannel->recv_errno = channel->recv_errno;
			dstchannel->t_crecv = area->t_crtu;
			dstchannel->recv_offset = rc;
			memcpy(&dstchannel->sa_msg, &channel->sa_msg,
				sizeof (struct sockaddr_in));
			dstchannel->sa_msg_len = channel->sa_msg_len;
			return 1;
		}
		if (!channel->recv_offset) {
			if (area->n_stations > 2)
				return 1;	// unmatched host
		}
	} else {
		if (channel->recv_offset == 0)
			return 0;
		getnameinfo((struct sockaddr *)&channel->sa_msg,
			channel->sa_msg_len, clienthost,
			sizeof(clienthost), clientservice,
			sizeof(clientservice), NI_NUMERICHOST);
		rc = channel->recv_offset;
		channel->recv_offset = 0;
	}
	if (ta_trace_level & TA_TRAL_TRANSPORT
	|| rc < 0 && (ta_trace_level & TA_TRAL_ERRORS)) {
		if (rc < 0)
			fprintf(ta_file, "* recvfrom %s sd=%d rc=%d %d %s\n",
				clienthost, cchannel->sd,
				rc, my_errno, 
				strerror(my_errno));
		else if (rc > 0)
			fprintf(ta_file, "recvfrom %s sd=%d %d bytes\n", 
				clienthost, cchannel->sd, rc);
	}
	if (rc > 0) {
		if (buffer_length > rc)
			buffer_length = rc;
		memset(buffer, 0, buffer_length);
	}
	return rc;
}

/* ------------------------------------------------------- */
static char *g_ttab[] = {
        "Echo Reply",
        "ICMP 1",
        "ICMP 2",
        "Dest Unreachable",
        "Source Quench",
        "Redirect",
        "ICMP 6",
        "ICMP 7",
        "Echo",
        "ICMP 9",
        "ICMP 10",
        "Time Exceeded",
        "Parameter Problem",
        "Timestamp",
        "Timestamp Reply",
        "Info Request",
        "Info Reply"
};
const char * taPingReadableErrno(int ierrno)
{
	if (ierrno > 0)
		return strerror(ierrno);
	if (ierrno == 0)
		return "";
	if (-ierrno >= sizeof(g_ttab)/sizeof(const char *))
		return NULL;
	return g_ttab[-ierrno];
}

