/*
 *                       ***** TA Station *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1 Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following disclaimer.
 *
 * 2 Redistributions in binary form must reproduce the
 *   above copyright notice, this list of conditions
 *   and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * 3 Neither the name of the General Energy Technologies
 *   nor the names of its contributors may be used to
 *   endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <string.h>

static TaStationQos g_station_qos = {
	"", 	// hostname
	7700,		// first_port
	0,		// delta_port
	2, 		// enable_mask
	3, 		// ping_mask
	10000, 		// tp_cycle
	3000000, 	// tp_discovery_timeout
	2000000,	// tp_ping_timeout
	TA_SEND_STRAT_MIN_ZVAL,	// send_strat
	"239.255.0.1",	// group_addr
	TA_PROT_FORMAT_SRPS,	// prot_format
	_False};	// connect_on_request

/* ------------------------------------------------------- */
void taDefaultStationQos(TaStationQos *qos)
{
	*qos = g_station_qos;
}

/* ------------------------------------------------------- */
int taInitStation(TaShort sid, const TaStationQos *qos, 
	const TaChannelQos *cqos, const char *name,
	TaArea *area, TaStation *station)
{
	int i, rc, j, ctype;
	TaULong mask;

	if (!area || !station)
		return TA_ERR_ARGUMENT;
	memset(station, 0, sizeof(TaStation));
	if (sid < 0 || sid >= area->max_stations)
		return TA_ERR_OUT_OF_MEMORY;
	station->sid = sid;
	station->zval = sid;	// may be changed
	station->lib_cid = -1;
	area->stations[sid] = station;
	station->area = area;
	if (sid >= area->n_stations)
		area->n_stations = sid+1;
	if (qos)
		station->qos = *qos;
	else
		station->qos = g_station_qos;
	if (name)
		strncpy(station->name, name, TA_NAME_LENGTH);
	station->name[TA_NAME_LENGTH-1] = 0;
	if (!station->qos.hostname[0])
		strcpy(station->qos.hostname, station->name);
	if (station->qos.hostname[0] && 
	station->qos.first_port == g_station_qos.first_port) {
		char *p = strchr(station->qos.hostname, ':');
		if (p) {
			station->qos.first_port = atoi(p+1);
			*p = 0;
		}
	}

	rc = 0;
	mask = 1;
	for (i = 0; i < area->n_ctypes; i++) {
		TaChannel *channel = &station->channels[i];
		ctype = area->ctypes[i];
		mask = 1 << ctype;
		rc |= taInitChannel(ctype, cqos, channel);
		channel->cid = i;
		if (mask & station->qos.enable_mask) {
			station->n_channels = i+1;
			channel->is_enabled = _True;
			// the hostname can be set explicitly
			if (channel->flags & TA_FLAGS_UNIX_DOMAIN)
				sprintf(channel->hostname, "a%d-s%d",
					(int)area->aid, (int)station->sid);
			else if (channel->flags & TA_FLAGS_MCAST)
				strncpy(channel->hostname, 
					station->qos.group_addr, TA_NAME_LENGTH);
			else
				strncpy(channel->hostname, 
					station->qos.hostname, TA_NAME_LENGTH);
			channel->hostname[TA_NAME_LENGTH-1] = 0;
			if (channel->flags & (TA_FLAGS_DGRAM|TA_FLAGS_STREAM)) {
				if (station->qos.first_port < 0)
					// -8080 sets exact 8080
					channel->port = 
						-station->qos.first_port;
				else if (channel->flags & TA_FLAGS_STREAM
				&& !(channel->flags & TA_FLAGS_LISTEN))
					channel->port =
						station->qos.first_port+ctype;
				else
					channel->port =
						station->qos.first_port+ctype+1;
				if (station->qos.delta_port) {
					// find the same and increment
					for (j = i-1; j >= 0; j--) {
						if (area->ctypes[j] != ctype)
							continue;
						channel->port = channel->port
							+ station->qos.delta_port;
						break;
					}
				}
			}
		}
		// find partner server to client
		channel->partner_cid = -1;
//printf("i:%d flags:%d n_ctypes:%d\n", i, channel->flags, area->n_ctypes);
		if (channel->flags & TA_FLAGS_STREAM
		&& !(channel->flags & TA_FLAGS_LISTEN)) {
			int j;
			for (j = i-1; j >= 0; j--) {
//printf("j:%d flags:%d port:%d/%d ct:%d/%d\n", j, station->channels[j].flags,
//	channel->port, station->channels[j].port, area->ctypes[j], ctype);
				if (area->ctypes[j] != ctype-1)
					continue;
//				if (!(station->channels[j].flags 
//				& TA_FLAGS_STREAM))
//					continue;
				if (channel->port && station->channels[j].port
				&&  channel->port != station->channels[j].port)
					continue;
				channel->partner_cid = j;
				station->channels[j].partner_cid = i;
				break;
			}
		}
	}
	return rc;
}

/* ------------------------------------------------------- */
int taInitStationData(const TaArea *area, TaStation *station)
{
	int i, rc = 0, irc;
	TaOctet is_datahost = _False, is_send_station = _False;
	if (area->host_station == station)
		is_datahost = _True;
	else if (area->send_station == station)
		is_send_station = _True;
	
	for (i = 0; i < station->n_channels; i++) {
		TaChannel *channel = &station->channels[i];
		TaOctet trans_flags = 0;
		if (!channel->is_enabled)
			continue;
		taCloseChannelData(channel);
		if (channel->flags & TA_FLAGS_DGRAM) {
			if (is_datahost)
				trans_flags = TA_TRANS_RECV;
			else if (is_send_station)
				trans_flags = TA_TRANS_SEND;
			else if (channel->flags & TA_FLAGS_MCAST)
				continue;
			else {
				// indirect sending via send_station
				trans_flags = TA_TRANS_SEND | TA_TRANS_INDIRECT;
				channel->link_stat = TA_LINK_OK;
			}
		} else if (channel->flags & TA_FLAGS_STREAM) {
			if (channel->partner_cid < 0)
				continue;
			trans_flags = TA_TRANS_SEND | TA_TRANS_RECV;
			if (station->qos.send_strat == TA_SEND_STRAT_SERVER) {
				if (!(channel->flags & TA_FLAGS_LISTEN))
					continue;
			} else if (station->qos.send_strat == TA_SEND_STRAT_CLIENT) {
				if (channel->flags & TA_FLAGS_LISTEN)
					continue;
			} else {
				TaChannel *pchannel, *host_channel, *phost_channel;
				pchannel = &station->channels[channel->partner_cid];
				host_channel = &area->host_station->channels[i];
				phost_channel = &area->host_station->
					channels[channel->partner_cid];
				if (!phost_channel->is_enabled)
					continue;
				if (!pchannel->is_enabled || !host_channel->is_enabled)
					;
				else if (station->qos.send_strat 
				== TA_SEND_STRAT_MIN_ZVAL) {
					if ((area->host_station->zval > station->zval
					&& channel->flags & TA_FLAGS_LISTEN)
					||  (area->host_station->zval <= station->zval
					&& !(channel->flags & TA_FLAGS_LISTEN)))
						continue;
				} else if (station->qos.send_strat
				== TA_SEND_STRAT_CR_SW) {
					if (!(channel->flags & TA_FLAGS_LISTEN))
						trans_flags = TA_TRANS_RECV;
				}
			}
		} else if (channel->flags & (TA_FLAGS_LIB | TA_FLAGS_ICMP)) {
			if (is_datahost)
				continue;
			trans_flags = TA_TRANS_SEND|TA_TRANS_RECV;
		} else if (channel->flags & TA_FLAGS_MEMORY) {
			if (strcmp(area->host_station->qos.group_addr, 
			station->qos.group_addr))
				continue;	// mem connects grouped by hosts
			if (channel->flags & TA_FLAGS_MESSAGE) {
				if (is_datahost)
					trans_flags = TA_TRANS_RECV;
				else
					trans_flags = TA_TRANS_SEND;
			} else {
				if (is_datahost)
					continue;
				if (is_send_station)
					trans_flags = TA_TRANS_SEND|TA_TRANS_RECV;
				else
					trans_flags = TA_TRANS_RECV|TA_TRANS_INDIRECT;
			}
		} 
		channel->trans_flags = trans_flags;
		irc =
		taInitChannelData(is_datahost, is_send_station, station, channel);
		if (irc < 0)
			channel->link_stat = TA_LINK_ERR;
		rc |= irc;
	}

	if (station->on_init)
		rc = (*station->on_init)(area, station);
	return rc;
}

/* ------------------------------------------------------- */
int taReinitStation(const TaStationQos *qos, TaStation *station)
{
	int i, rc, j, ctype;
	TaArea *area = station->area;

	station->qos = *qos;
	if (!station->qos.hostname[0])
		strcpy(station->qos.hostname, station->name);
	if (station->qos.hostname[0] && 
	station->qos.first_port == g_station_qos.first_port) {
		char *p = strchr(station->qos.hostname, ':');
		if (p) {
			station->qos.first_port = atoi(p+1);
			*p = 0;
		}
	}

	rc = 0;
	for (i = 0; i < station->n_channels; i++) {
		TaChannel *channel = &station->channels[i];
		if (!channel->is_enabled)
			continue;
		ctype = channel->ctype;

		// the hostname can be set explicitly
		if (channel->flags & TA_FLAGS_UNIX_DOMAIN)
			sprintf(channel->hostname, "a%d-s%d",
				(int)area->aid, (int)station->sid);
		else if (channel->flags & TA_FLAGS_MCAST)
			strncpy(channel->hostname, 
				station->qos.group_addr, TA_NAME_LENGTH);
		else
			strncpy(channel->hostname, 
				station->qos.hostname, TA_NAME_LENGTH);
		channel->hostname[TA_NAME_LENGTH-1] = 0;
		if (channel->flags & (TA_FLAGS_DGRAM|TA_FLAGS_STREAM)) {
			if (station->qos.first_port < 0)
				// -8080 sets exact 8080
				channel->port = 
					-station->qos.first_port;
			else if (channel->flags & TA_FLAGS_STREAM
			&& !(channel->flags & TA_FLAGS_LISTEN))
				channel->port =
					station->qos.first_port+ctype;
			else
				channel->port =
					station->qos.first_port+ctype+1;
			if (station->qos.delta_port) {
				// find the same and increment
				for (j = i-1; j >= 0; j--) {
					if (area->ctypes[j] != ctype)
						continue;
					channel->port = channel->port
						+ station->qos.delta_port;
					break;
				}
			}
		}
		// find partner server to client
		channel->partner_cid = -1;
		if (channel->flags & TA_FLAGS_STREAM
		&& !(channel->flags & TA_FLAGS_LISTEN)) {
			int j;
			for (j = i-1; j >= 0; j--) {
				if (area->ctypes[j] != ctype-1)
					continue;
				if (channel->port && station->channels[j].port
				&&  channel->port != station->channels[j].port)
					continue;
				channel->partner_cid = j;
				station->channels[j].partner_cid = i;
				break;
			}
		}
	}
	return taInitStationData(station->area, station);
}

/* ------------------------------------------------------- */
void taCloseStationData(TaStation *station)
{
	int i;
	for (i = 0; i < station->n_channels; i++) {
		TaChannel *channel = &station->channels[i];
		if (!channel->is_enabled)
			continue;
		taCloseChannelData(channel);
	}
}

/* ------------------------------------------------------- */
int taUpdateStationData(const TaArea *area, TaStation *station, 
	void *client_data)
{
	int rc = 0;
	if (station->on_update)
		rc = (*station->on_update)(station, area->t_crtu, 
			client_data);
	return rc;
}

/* ------------------------------------------------------- */
void taDumpStationLinks(const TaArea *area)
{
	int i, k;
	for (i = 0; i < area->n_stations; i++) {
		const TaStation *station = area->stations[i];
		fprintf(ta_file, "#%d enable_mask:%x ping_mask:%x host:%c send:%c\n", 
			(int)station->sid, 
			(unsigned)station->qos.enable_mask,
			(unsigned)station->qos.ping_mask,
			(station == area->host_station ? 'Y' : 'N'),
			(station == area->send_station ? 'Y' : 'N'));
		for (k = 0; k < station->n_channels; k++) {
			const TaChannel *channel = &station->channels[k];
			const TaChannelProtocol *prot = 
				taChannelProtocol(channel->ctype);
			if (!channel->is_enabled || !prot)
				continue;
			fprintf(ta_file,
		"\tcid:%d %s link_stat:%d trans_flags:%d partner:%d sd:%d port:%d\n", 
				k, prot->name, (int)channel->link_stat, 
				(int)channel->trans_flags, channel->partner_cid,
				channel->sd, channel->port);
		}
	}
	fflush(ta_file);
}

