MODULE Ta_Files;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Files, RndFile, xFilePos, IOChan, SYSTEM;

	CONST
		(* ---------- File mode bits ---------- *)
		READ* = 0; WRITE* = 1; OLD* = 2; TEXT* = 3; RAW* = 4; 
		NOT_FOUND = 22;
		INVALID_VAL = 30; UNSUPPORTED = 31;
		BUFFER_LENGTH = 256;
		
	TYPE
		FileDesc = RECORD (Files.FileDesc)
			cid: RndFile.ChanId;
			shared, old, opened: BOOLEAN;
			crt_id, last_id: INTEGER;
		END;
		File = POINTER TO FileDesc;
		
		ReaderDesc = RECORD (Files.ReaderDesc)
			base: File;
			id: INTEGER;
			pos: RndFile.FilePos
		END;
		Reader = POINTER TO ReaderDesc;
		
		WriterDesc = RECORD (Files.WriterDesc)
			base: File;
			id: INTEGER;
			pos: RndFile.FilePos
		END;
		Writer = POINTER TO WriterDesc;
	
		LineReader = POINTER TO LineReaderDesc;
		LineReaderDesc = RECORD (Files.ReaderDesc)
			buffer: ARRAY BUFFER_LENGTH OF CHAR;
			pos, len: LONGINT
		END;

	(* ========== Files Access Implementation ========== *)
	PROCEDURE NewFile* ( name-: ARRAY OF CHAR; VAR file: Files.File);
		VAR cid: RndFile.ChanId; res: RndFile.OpenResults; 
		f: File;
	BEGIN
		RndFile.OpenClean(cid, name, RndFile.raw + RndFile.old, res);
		IF res = RndFile.opened THEN
			NEW(f); f.cid := cid; f.shared := FALSE; f.old := FALSE; 
			f.crt_id := 1; f.last_id := 0; f.opened := TRUE; file := f
		ELSE
			file := NIL
		END
	END NewFile;
	
	PROCEDURE OldFile* ( name-: ARRAY OF CHAR; shared: BOOLEAN;
	VAR file: Files.File);
		VAR cid: RndFile.ChanId; res: RndFile.OpenResults; fs: RndFile.FlagSet;
		f: File;
	BEGIN
		IF shared THEN
			fs := RndFile.read
		ELSE
			fs := RndFile.write
		END;
		fs := fs + RndFile.raw;
		RndFile.OpenOld(cid, name, fs, res);
		IF res = RndFile.opened THEN
			NEW(f); f.cid := cid; f.shared := shared; f.old := TRUE; 
			f.crt_id := 1; f.last_id := 0; f.opened := TRUE; file := f
		ELSE
			file := NIL
		END
	END OldFile;
	
	
	PROCEDURE P2L (pos: RndFile.FilePos): LONGINT;
		VAR j: LONGINT; b: BOOLEAN;
	BEGIN
		j := -1;
		b := xFilePos.PosToInt(j, pos);
		RETURN j
	END P2L;
	
	PROCEDURE L2P (lpos: LONGINT): RndFile.FilePos;
		VAR pos: RndFile.FilePos;
	BEGIN
		xFilePos.IntToPos(pos, lpos);
		RETURN pos
	END L2P;
	
	(* ========== File interface ========== *)
	PROCEDURE (f: File) Length (): LONGINT;
		VAR pos: RndFile.FilePos;
	BEGIN
		pos := RndFile.EndPos(f.cid);
		RETURN P2L(pos)
	END Length;
	
	PROCEDURE (f: File) NewReader (old: Files.Reader): Files.Reader;
		VAR rd: Reader;
	BEGIN
		ASSERT(f.old, INVALID_VAL);
		IF old # NIL THEN
			rd := old(Reader)
		ELSE
			NEW(rd)
		END; 
		rd.base := f; rd.id := f.crt_id;
		INC(f.crt_id);
		RETURN rd
	END NewReader;

	PROCEDURE (f: File) NewWriter (old: Files.Writer): Files.Writer;
		VAR wr: Writer;
	BEGIN
		ASSERT(~f.shared, INVALID_VAL);
		IF old # NIL THEN
			wr := old(Writer)
		ELSE
			NEW(wr)
		END; 
		wr.base := f; wr.id := f.crt_id;
		INC(f.crt_id);
		RETURN wr
	END NewWriter;

	PROCEDURE (f: File) Flush;
	BEGIN
		IOChan.Flush(f.cid)
	END Flush;
	
	PROCEDURE (f: File) Register (name: Files.Name; type: Files.Type; 
	ask: BOOLEAN; VAR res: INTEGER);
	BEGIN
		res := UNSUPPORTED
	END Register;
	
	PROCEDURE (f: File) Close;
	BEGIN
		IF f.opened THEN
			RndFile.Close(f.cid);
			f.opened := FALSE
		END
	END Close;

	(* ========== Reader interface ========== *)
	PROCEDURE (r: Reader) Base (): Files.File;
	BEGIN
		RETURN r.base
	END Base;
	
	PROCEDURE (r: Reader) Pos (): LONGINT;
		VAR pos: RndFile.FilePos;
	BEGIN
		RETURN P2L(r.pos)
	END Pos;

	PROCEDURE (r: Reader) SetPos (lpos: LONGINT);
	BEGIN
		r.pos := L2P(lpos);
		RndFile.SetPos(r.base.cid, r.pos);
		r.base.last_id := r.id;
		r.eof := FALSE
	END SetPos;

	PROCEDURE (r: Reader) ReadByte (VAR x: SHORTINT);
		VAR nread: SYSTEM.CARD32;
	BEGIN
		IF r.base.last_id # r.id THEN
			RndFile.SetPos(r.base.cid, r.pos);
			r.base.last_id := r.id
		END;
		IOChan.RawRead(r.base.cid, SYSTEM.ADR(x), 1, nread);
		r.pos := RndFile.CurrentPos(r.base.cid);
		r.eof := (nread = 0)
	END ReadByte;

	PROCEDURE (r: Reader) ReadBytes (VAR x: ARRAY OF SHORTINT;
	beg, len: LONGINT);
		VAR nread: SYSTEM.CARD32; 
	BEGIN
		ASSERT((beg >= 0) & (len >= 0), INVALID_VAL);
		IF r.base.last_id # r.id THEN
			RndFile.SetPos(r.base.cid, r.pos);
			r.base.last_id := r.id
		END;
		IF len > 0 THEN
			IOChan.RawRead(r.base.cid, SYSTEM.ADR(x)+beg, len, nread);
			r.pos := RndFile.CurrentPos(r.base.cid);
			r.eof := (SYSTEM.VAL(LONGINT, nread) # len)
		END
	END ReadBytes;

	(* ========== Writer interface ========== *)
	PROCEDURE (w: Writer) Base (): Files.File;
	BEGIN
		RETURN w.base
	END Base;
	
	PROCEDURE (w: Writer) Pos (): LONGINT;
		VAR pos: RndFile.FilePos;
	BEGIN
		RETURN P2L(w.pos)
	END Pos;

	PROCEDURE (w: Writer) SetPos (lpos: LONGINT);
	BEGIN
		w.pos := L2P(lpos);
		RndFile.SetPos(w.base.cid, w.pos);
		w.base.last_id := w.id
	END SetPos;

	PROCEDURE (w: Writer) WriteByte (x: SHORTINT);
	BEGIN
		IF w.base.last_id # w.id THEN
			RndFile.SetPos(w.base.cid, w.pos);
			w.base.last_id := w.id
		END;
		IOChan.RawWrite(w.base.cid, SYSTEM.ADR(x), 1);
		w.pos := RndFile.CurrentPos(w.base.cid)
	END WriteByte;

	PROCEDURE (w: Writer) WriteBytes (x-: ARRAY OF SHORTINT;
	beg, len: LONGINT);
	BEGIN
		ASSERT((beg >= 0) & (len >= 0), INVALID_VAL);
		IF w.base.last_id # w.id THEN
			RndFile.SetPos(w.base.cid, w.pos);
			w.base.last_id := w.id
		END;
		IF len > 0 THEN
			IOChan.RawWrite(w.base.cid, SYSTEM.ADR(x)+beg, len);
			w.pos := RndFile.CurrentPos(w.base.cid);
		END
	END WriteBytes;
	
	(* ========== Line Access Implementation ========== *)
	PROCEDURE NewLineReader* (old: Files.Reader; line: ARRAY OF CHAR; 
	ind: INTEGER): Files.Reader;
		VAR lrd: LineReader; j: INTEGER;
	BEGIN
		ASSERT(ind >= 0, INVALID_VAL);
		IF old = NIL THEN
			NEW(lrd)
		ELSE
			lrd := old(LineReader)
		END;
		j := 0;
		WHILE line[j+ind] # 0X DO
			lrd.buffer[j] := line[j+ind]; INC(j)
		END;
		lrd.buffer[j] := 0X; lrd.pos := 0; lrd.len := j;
		RETURN lrd
	END NewLineReader;
	
	(* ========== Reader interface ========== *)
	PROCEDURE (lrd: LineReader) Pos (): LONGINT;
	BEGIN
		RETURN lrd.pos
	END Pos;

	PROCEDURE (lrd: LineReader) ReadBytes (VAR x: ARRAY OF SHORTINT;
	beg, len: LONGINT);
		VAR j: INTEGER; ch: CHAR;
	BEGIN
		ASSERT((beg >= 0) & (len >= 0), INVALID_VAL);
		j := 0; ch := lrd.buffer[lrd.pos];
		WHILE (j < len) & (ch # 0X) DO
			x[beg+j] := SHORT(ORD(ch));
			INC(j); ch := lrd.buffer[lrd.pos+j]
		END;
		IF j < len THEN x[beg+j] := 0; lrd.eof := TRUE ELSE lrd.eof := FALSE END;
		INC(lrd.pos, j)
	END ReadBytes;
	

END Ta_Files.