MODULE Ta_Poll;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tl := Ta_Lib, Tc := Ta_Channels, SYSTEM, Log ;
	
	CONST
		MAX_FD_FOR_ONE = 4;
		
	TYPE
		PollStation = RECORD
			fd_array: ARRAY MAX_FD_FOR_ONE OF Tb.Int32;
			trans_flags: SET;
			e_ind, e_num: Tb.Int32;
		END;		
		Polling = POINTER TO PollingDesc;
		PollingDesc = RECORD (Tc.PollingDesc)
			n_stations: INTEGER;
			stations: POINTER TO ARRAY OF PollStation;
			nfds: Tb.Int32;
			ufds: ARRAY Tl.MAX_FD_SET OF Tl.pollfd;
			events: Tc.PollingEvents;
			fd_changed: BOOLEAN
		END;

		PollingFactoryDesc* = RECORD (Tc.PollingFactoryDesc)
		END;		
		PollingFactory* = POINTER TO PollingFactoryDesc;
	

	(* ========== Polling interface ========== *)
	PROCEDURE (polling: Polling) FdSetChannel (index: INTEGER; channel: Tc.Channel);
		VAR i: INTEGER;
	BEGIN
		ASSERT((index >= 0) & (index < polling.n_stations), Tb.CHECK_RANGE_ERROR);
		ASSERT(channel # NIL, Tb.NIL_POINTER);
		channel.GetFds(0, i, polling.stations[index].fd_array);
		polling.stations[index].e_num := i;
		polling.stations[index].trans_flags := channel.trans_flags;
		polling.fd_changed := TRUE
	END FdSetChannel;

	PROCEDURE Prepare (polling: Polling);
		VAR j, k: INTEGER;
	BEGIN
		polling.nfds := 0;

		FOR j := 0 TO polling.n_stations-1 DO
			polling.stations[j].e_ind := polling.nfds; 
			FOR k := 0 TO polling.stations[j].e_num-1 DO
				polling.ufds[polling.nfds+k].fd := polling.stations[j].fd_array[k];
				polling.ufds[polling.nfds+k].events := Tl.POLLERR;
				IF (Tc.TRANS_RECV_DIRECT IN polling.stations[j].trans_flags) 
				OR (Tc.TRANS_LISTEN IN polling.stations[j].trans_flags) THEN
					INC(polling.ufds[polling.nfds+k].events, Tl.POLLIN);
				END;
				IF Tc.TRANS_SEND_DIRECT IN polling.stations[j].trans_flags THEN
					INC(polling.ufds[polling.nfds+k].events, Tl.POLLOUT);
				END;
				IF (Tc.TRANS_CONNECTION IN polling.stations[j].trans_flags)
				OR (Tc.TRANS_LISTEN IN polling.stations[j].trans_flags) THEN
					INC(polling.ufds[polling.nfds+k].events, Tl.POLLHUP);
				END;
				INC(polling.nfds, polling.stations[j].e_num);
			END
		END
	END Prepare;
	
	PROCEDURE (polling: Polling) Poll ((*OUT*)VAR events: Tc.PollingEvents; VAR res: INTEGER);
		VAR j, k: INTEGER; tv: Tl.timeval; rc, my_errno: Tb.Int32;
		set: SET;
	BEGIN
		IF polling.fd_changed THEN
			Prepare(polling);
			polling.fd_changed := FALSE
		END;
		IF polling.nfds > 0 THEN
			FOR j := 0 TO polling.nfds-1 DO polling.ufds[j].revents := 0 END;
			rc := Tl.poll(polling.ufds[0], polling.nfds, 0);
			my_errno := Tb.Errno();
			IF Tb.TRAL_POLLING IN Tb.trace THEN
				Log.String("poll rc="); Log.Int(rc); Log.String(" n_stations="); Log.Int(polling.n_stations); 
				Log.String(" nfds="); Log.Int(polling.nfds); Log.String(" err="); Log.Int(my_errno); Log.Ln
			END;
			IF rc > 0 THEN
				FOR j := 0 TO polling.n_stations-1 DO
					polling.events[j] := {};
					IF polling.stations[j].e_num > 0 THEN
						FOR k := 0 TO polling.stations[j].e_num-1 DO
							set := SYSTEM.VAL(SET, polling.ufds[polling.stations[j].e_ind+k].revents);
							IF Tc.POLLIN IN set THEN
								INCL(polling.events[j], Tc.POLLIN+k*Tc.POLL_NEXT)
							END;
							IF Tc.POLLOUT IN set THEN
								INCL(polling.events[j], Tc.POLLOUT+k*Tc.POLL_NEXT)
							END;
							IF Tc.POLLERR IN set THEN
								INCL(polling.events[j], Tc.POLLERR+k*Tc.POLL_NEXT)
							END;
							IF Tc.POLLHUP IN set THEN
								INCL(polling.events[j], Tc.POLLHUP+k*Tc.POLL_NEXT)
							END
						END;
						IF Tb.TRAL_POLLING IN Tb.trace THEN
							Log.String("poll sid="); Log.Int(j);
							Log.String(" poll_events=");
							Log.Int(Tb.Set_Int32(polling.events[j])); Log.Ln;
						END
					END
				END
			END;
			res := SHORT(rc)
		ELSE
			res := -3
		END;
		events := polling.events
	END Poll;


	PROCEDURE (pf: PollingFactory) NewPolling* (n_stations: INTEGER; (*OUT*)VAR poll: Tc.Polling; 
	(*OUT*)VAR res: INTEGER);
		VAR polling: Polling;
	BEGIN
		IF (n_stations <= 0) OR (n_stations >= Tl.MAX_FD_SET) THEN
			res := Tb.CHECK_RANGE_ERROR
		ELSE
			NEW(polling);
			polling.n_stations := n_stations;
			NEW(polling.stations, n_stations);
			NEW(polling.events, n_stations);
			polling.nfds := 0;
			poll := polling;
			res := 0
		END;
	END NewPolling;
	
END Ta_Poll.