MODULE Ta_SOAPStations;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ti := Ta_TInfo, Ts := Ta_Stations, 
		Tm := Ta_TunedMappers, Tp := Ta_SimpleParsers, Ta_CfgTInfo, Files, Log, SYSTEM;
		
	CONST
		SOAP_MID* = 32010; SOAP_11_MID* = 32011; SOAP_12_MID* = 32012;
			SOAP_GET_MID* = 32013;  SOAP_POST_MID* = 32014;
			SOAP_MATCH_INTERFACE* = 32020;
			SOAP_11_CLIENT* = 32021; SOAP_12_CLIENT* = 32022;
			SOAP_GET_CLIENT* = 32023; SOAP_POST_CLIENT* = 32024;
		PARSE_STR* = -17; PARSE_NOT_FOUND* = -6; 
		PARSE_WRONG_MID* = -7; PARSE_WRONG_NAME* = -8;
		TRUNCATED* = -9; UNSUPPORTED* = -10;
		WRITE_OPTS = {Tm.addLf};

	TYPE
		TableQos* = RECORD (Ts.TableQosBase)
			match_mode*: INTEGER;
			xmlns*: Tb.Name64;
			nc_ignore*: BOOLEAN;
		END;

		Carrier = POINTER TO CarrierDesc;
		Mapper = POINTER TO MapperDesc;

		PtrData = POINTER TO PtrDataDesc;
		PtrDataDesc = RECORD
			nbytes, offset, item_type: Tb.Int32;
			dim: Tb.Int16;
			next: PtrData
		END;

		Reader = POINTER TO ReaderDesc;
		ReaderDesc = RECORD (Files.ReaderDesc)
			srd: Ts.Reader
		END;

		Writer = POINTER TO WriterDesc;
		WriterDesc = RECORD (Files.WriterDesc)
			swr: Ts.Writer
		END;

		CarrierDesc = RECORD (Ts.CarrierDesc)
			aid: INTEGER;
			rd: Reader;
			wr: Writer;
		END;

		CarrierFactoryDesc* = RECORD (Ts.CarrierFactoryDesc)
		END;		
		CarrierFactory* = POINTER TO CarrierFactoryDesc;

		MapperDesc = RECORD (Ts.MapperDesc)
			ti: Ti.TInfo;
			typ: Ti.Name;
			iface_name: Tb.Name64;
			tqos: TableQos;
			msg_type, basic_length, buffer_size: Tb.Int32;
			iface_code: Tb.Int16;
			mid_read: INTEGER;
			xmlns_in: Tb.Name64;
			waiting_for_response: BOOLEAN
		END;
		
		MapperFactoryDesc* = RECORD (Ts.MapperFactoryDesc)
		END;		
		MapperFactory* = POINTER TO MapperFactoryDesc;
		
		ParseData = RECORD (Tb.AnyRec)
			mapper: Mapper;
			item: Ti.Item;
			items: ARRAY Tp.MAX_STACK OF Ti.Item;
			dims: ARRAY Tp.MAX_STACK OF Tb.Int32;
			nlevel0: INTEGER;
			changed: BOOLEAN
		END;

		WriteData = RECORD (Tb.AnyRec)
			mapper: Mapper;
			f: Tm.Formatter;
			first: BOOLEAN
		END;

	VAR
		defaultTQos-: TableQos;


	(* ========== Carrier Factory ========== *)
	PROCEDURE (crf: CarrierFactory) NewCarrier* (pf: Tc.PollingFactory; 
	aid, n_stations: INTEGER; send_buffer_size, recv_buffer_size: Tb.Int32;  
	prot_format: ARRAY OF CHAR; (*OUT*)VAR cr: Ts.Carrier; (*OUT*)VAR res: INTEGER);
		VAR carrier: Carrier; polling: Tc.Polling;
	BEGIN
		ASSERT(n_stations > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(pf # NIL, Tb.NIL_POINTER);
		pf.NewPolling(n_stations, polling, res);
		IF res = 0 THEN
			NEW(carrier); carrier.aid := aid; NEW(carrier.rd); NEW(carrier.wr);
			carrier.SetNewCarrier(polling, send_buffer_size, recv_buffer_size, prot_format)
		END;
		cr := carrier
	END NewCarrier;

	(* ========== Reader Implementation ========== *)
	PROCEDURE (rd: Reader) Pos (): Tb.Int32;
	BEGIN
		RETURN rd.srd.pos
	END Pos;

	PROCEDURE (rd: Reader) ReadBytes (VAR x: ARRAY OF Tb.Int8;
	beg, len: Tb.Int32);
	BEGIN
		rd.srd.ReadBytes(x, beg, len);
		rd.eof := rd.srd.eof
	END ReadBytes;
	
	(* ========== Writer Implementation ========== *)
	PROCEDURE (wr: Writer) Pos (): Tb.Int32;
	BEGIN
		RETURN wr.swr.pos
	END Pos;
	
	PROCEDURE (wr: Writer) WriteBytes ( x-: ARRAY OF Tb.Int8;
	beg, len: Tb.Int32);
		VAR oldpos, j: Tb.Int32;
	BEGIN
		IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
			oldpos := wr.swr.pos;
			wr.swr.WriteBytes(x, beg, len);
			FOR j := 0 TO wr.swr.pos-oldpos-1 DO
				IF x[beg+j] # 0DH THEN
					Log.Char(SYSTEM.VAL(CHAR, x[beg+j]))
				ELSE
					Log.Ln
				END
			END
		ELSE
			wr.swr.WriteBytes(x, beg, len)
		END
	END WriteBytes;
	
	(* ========== Carrier Interface ========== *)
	PROCEDURE (cr: Carrier) InternalizeHeader (VAR rd: Ts.Reader; (*OUT*)VAR hdr: Ts.Header; 
	(*OUT*)VAR res: Tb.Int32);
		CONST HTTP11 = 11;
		VAR s: Tm.Scanner; j, k, l: INTEGER; name: Tb.Name64; 
		done, get_next, get_content: BOOLEAN;
	BEGIN
		cr.rd.srd := rd;
		s.ConnectTo(cr.rd);
		j := 0; rd.SetPos(0); hdr.mid_resp := 0; hdr.aid := cr.aid; 
		hdr.tlg_type := Ts.USER_TELEGRAM_TYPE;
		hdr.msg_type := 0; hdr.iface_code := 0; hdr.mid := 0;
		done := FALSE; get_next := FALSE; get_content := FALSE;
		s.SetOpts({Tm.withSpaces, Tm.withCrLf, Tm.separateStrings,
			Tm.specialAsString, Tm.bracesAsString, Tm.unquoteAmp, Tm.xmlMap});
		s.Scan;
		IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
			Log.String("--- IN SOAP ---"); Log.Int(rd.pos); Log.Ln
		END;
		WHILE ~done & (s.type # Tm.eot) DO
			IF s.type = Tm.string THEN
				IF j = 0 THEN
					IF s.string = "GET" THEN
						hdr.mid := SOAP_GET_MID
					ELSIF s.string = "POST" THEN
						hdr.mid := SOAP_POST_MID
					ELSIF s.string = "HTTP/1.1" THEN
						hdr.mid := HTTP11
					ELSE
						hdr.msg_type := SOAP_MID;
						hdr.iface_code := SOAP_MID;
						hdr.mid_resp := SOAP_MID;
						hdr.mid := SOAP_MID;
						done := TRUE
					END
				ELSE
					(* extract interface_name, msg_type *)
					IF (j = 1) & ((hdr.mid = SOAP_GET_MID) OR 
					(hdr.mid = SOAP_POST_MID)) THEN
						k := 0;
						WHILE (s.string[k+1] # 0X) & (s.string[k+1] # '/') DO
							name[k] := s.string[k+1]; INC(k)
						END;
						name[k] := 0X;
						hdr.msg_type := Tb.Crc32(name, -1, 0, 32);
						IF s.string[k+1] = '/' THEN
							l := k+2; k := 0;
							WHILE (s.string[k+l] # 0X) & (s.string[k+l] # '?') DO
								name[k] := s.string[k+l]; INC(k)
							END;
							name[k] := 0X;
							hdr.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(name, -1, 0, 16));
						END;
					END;
					(* get type of SOAP *)
					IF hdr.mid = HTTP11 THEN
						IF s.string = "<soap:Envelope" THEN
							hdr.mid := SOAP_11_CLIENT
						ELSIF s.string = "<soap12:Envelope" THEN
							hdr.mid := SOAP_12_CLIENT
						ELSIF s.string = "<string" THEN
							hdr.msg_type := SOAP_MID;
							hdr.iface_code := SOAP_MID;
							hdr.mid_resp := SOAP_MID;
							hdr.mid := SOAP_POST_CLIENT
						END
					ELSIF hdr.mid = SOAP_GET_MID THEN
						IF s.string = "Host:" THEN
							get_next := TRUE;
						END
					ELSIF hdr.mid = SOAP_POST_MID THEN
						IF s.string = "SOAPAction:" THEN
							get_next := TRUE;
							hdr.mid := SOAP_11_MID
						ELSIF s.string = "application/soap+xml;" THEN
							hdr.mid := SOAP_12_MID
						END
					END;
					(* extract item_name, iface_code *)
					IF hdr.mid = SOAP_GET_MID THEN
						IF get_next & (s.string # "Host:") THEN
							get_next := FALSE;
							done := TRUE
						END;
					END;
					IF hdr.mid = SOAP_11_MID THEN
						IF get_next & (s.string # "SOAPAction:") THEN
							get_next := FALSE;
							l := Tb.Len(s.string);
							WHILE (l >= 0) & (s.string[l] # '/') DO DEC(l) END;
							k := 0;
							WHILE s.string[k+l+1] # 0X DO 
								name[k] := s.string[k+l+1]; INC(k)
							END;
							name[k] := 0X;
							hdr.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(name, -1, 0, 16));
							done := TRUE
						END
					END;
					IF hdr.mid = SOAP_12_MID THEN
						IF s.string = "<soap12:Body>" THEN
							get_next := TRUE
						ELSIF get_next THEN
							get_next := FALSE;
							k := 0;
							WHILE s.string[k+1] # 0X DO 
								name[k] := s.string[k+1]; INC(k)
							END;
							name[k] := 0X;
							hdr.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(name, -1, 0, 16));
							done := TRUE
						END
					END;
					IF (hdr.mid = SOAP_11_CLIENT) OR (hdr.mid = SOAP_12_CLIENT) THEN
						IF s.string[0] = "<" THEN
							l := Tb.Len(s.string)-7;
							IF l > 0 THEN
								name := "Result>";
								k := 0;
								WHILE (s.string[k+l] # 0X) & (s.string[k+l] = name[k]) DO
									INC(k)
								END;
								IF s.string[k+l] = 0X THEN
									FOR k := 1 TO l-1 DO
										name[k-1] := s.string[k]
									END;
									name[k-1] := 0X;
									hdr.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(name, -1, 0, 16));
									hdr.msg_type := SOAP_MID;
									hdr.mid_resp := SOAP_MID;
									done := TRUE
								END
							END
						END
					END;
					(* extract length *)
					IF s.string = "Content-Length:" THEN
						get_content := TRUE;
					ELSIF get_content & (s.type = Tm.string) THEN
						get_content := FALSE;
						l := 0;
						Tb.StringToLInt(s.string, hdr.basic_length, l);
						IF l # 0 THEN hdr.basic_length := 0 END;
						hdr.buffer_length := rd.length;	(* to be changed ? *)
						IF (hdr.mid = SOAP_POST_MID) 
						& (hdr.mid = SOAP_POST_CLIENT) THEN
							done := TRUE
						END
					END;
				END
			END;
			IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
				IF s.type = Tm.crlf THEN
					Log.Ln
				ELSE
					Log.Char("[");Log.String(s.string);Log.Char("]")
				END
			END;
			s.Scan;
			IF s.type = Tm.string THEN INC(j) END
		END;
		IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
			Log.Ln; Log.String("--- END SOAP ---"); Log.Int(rd.pos); 
			Log.Int(hdr.mid); Log.Int(hdr.mid_resp); Log.Int(hdr.buffer_length);
			Log.String(" |"); Log.Int(hdr.msg_type); Log.Int(hdr.iface_code); Log.Ln
		END;
		res := 0;
	END InternalizeHeader;

	PROCEDURE (cr: Carrier) ExternalizeHeader (VAR wr: Ts.Writer;  hdr-: Ts.Header;
	(*OUT*)VAR res: Tb.Int32);
	BEGIN
		res := 0;
	END ExternalizeHeader;

	(* ========== Mapper interface ========== *)
	PROCEDURE (mapper: Mapper) ExploreObject (); 
		VAR mod: Ti.Name;
	BEGIN
		mapper.ti.item.GetTypeName(mod, mapper.typ);
		IF (mapper.tqos.match_mode = SOAP_GET_CLIENT)
		OR (mapper.tqos.match_mode = SOAP_POST_CLIENT) THEN
			mapper.msg_type := SOAP_MID;
			mapper.iface_code := SOAP_MID
		ELSIF (mapper.tqos.match_mode = SOAP_11_CLIENT)
		OR (mapper.tqos.match_mode = SOAP_12_CLIENT) THEN
			mapper.msg_type := SOAP_MID;
			mapper.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(mapper.typ, -1, 0, 16));
		ELSE
			mapper.msg_type := Tb.Crc32(mapper.iface_name, -1, 0, 32);
			mapper.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(mapper.typ, -1, 0, 16));
		END;
		mapper.cancelled := FALSE;
		IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("Table '"); Log.String(mapper.typ); Log.String("' [msg_type,iface_code]=");
			Log.Int(mapper.msg_type); 
			Log.Int(mapper.iface_code); Log.Ln;
		END;
	END ExploreObject;

	PROCEDURE (mapper: Mapper) MatchHeader ( hdr-: Ts.Header): BOOLEAN;
		VAR match: BOOLEAN;
	BEGIN
		CASE mapper.tqos.match_mode OF
		SOAP_MATCH_INTERFACE:
			match := (hdr.msg_type = mapper.msg_type) 
				& (hdr.iface_code = mapper.iface_code);
		| SOAP_11_CLIENT, SOAP_12_CLIENT:
			match := (hdr.msg_type = mapper.msg_type) & (hdr.iface_code = mapper.iface_code)
			& mapper.waiting_for_response;
		| SOAP_GET_CLIENT, SOAP_POST_CLIENT:
			match := (hdr.msg_type = mapper.msg_type) & mapper.waiting_for_response;
		END;
		IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("SOAP Match="); Log.Bool(match); Log.Int(hdr.mid);
			Log.String("# l="); Log.Int(hdr.basic_length); 
			Log.String(" m="); Log.Int(hdr.msg_type); Log.Int(mapper.msg_type); 
			Log.String(" i="); Log.Int(hdr.iface_code); Log.Int(mapper.iface_code); 
			Log.String(" s="); Log.Int(hdr.src_sid);
			Log.String(" o="); Log.Int(hdr.dst_oid); 
			Log.String(" w="); Log.Bool(mapper.waiting_for_response); Log.Ln;
		END;
		RETURN match
	END MatchHeader;
	
	PROCEDURE OnReadSoap (parse_flag: INTEGER; 
	VAR nstack: Tp.NamesStackArray; nlevel: INTEGER; value: ARRAY OF CHAR;
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		VAR ritem, base, ri: Ti.Item; ok: BOOLEAN; nl: INTEGER; 
		oldstr: Tb.Name; len: Tb.Int32;
	BEGIN
		WITH userdata: ParseData DO
			nl := nlevel - userdata.nlevel0;
			IF nl >= 0 THEN
				IF (parse_flag = Tp.COUPLE_BEGIN) 
				OR (parse_flag = Tp.COUPLE_BOTH) THEN
					IF nl = 0 THEN
						ritem := userdata.item
					ELSE
						ritem := userdata.items[nl-1];
					END;
					CASE ritem.typ OF
					0:
						userdata.items[nl] := Ti.zeroitem;
					| Tb.ptrTyp:
						IF nstack[nlevel-1].index = 0 THEN
							userdata.dims[nl] := Ti.PtrLen(ritem)
						END;
						len := nstack[nlevel-1].index+1;
						CASE userdata.mapper.tqos.realloc_mode OF
						| Ts.DONT_REALLOC:
							IF len > userdata.dims[nl] THEN
								res := TRUNCATED
							END
						| Ts.REALLOC_LARGEST:
							res := UNSUPPORTED;
						| Ts.REALLOC_EXACT:
							res := UNSUPPORTED;
						END;
						IF (res >= 0) & (len > 0) THEN
							ritem.Deref(base);
							IF base.typ = Tb.arrTyp THEN
								base.Index(nstack[nlevel-1].index, ri);
								IF ri.typ = Tb.recTyp THEN
									userdata.mapper.ti.Find(nstack[nlevel].name, 
										ri, userdata.items[nl]);
								ELSIF (ri.typ = Tb.char8Typ) OR (ri.typ = Tb.char16Typ) THEN
									userdata.items[nl] := base
								ELSE
									userdata.items[nl] := ri
								END
							END
						END
					| Tb.arrTyp:
						len := ritem.Len();
						IF nstack[nlevel-1].index >= len THEN
							res := TRUNCATED
						ELSE
							ritem.Index(nstack[nlevel-1].index, ri);
							IF ri.typ = Tb.recTyp THEN
								userdata.mapper.ti.Find(nstack[nlevel].name, 
									ri, userdata.items[nl])
							ELSIF (ri.typ = Tb.char8Typ) OR (ri.typ = Tb.char16Typ) THEN
								userdata.items[nl] := ritem
							ELSE
								userdata.items[nl] := ri
							END
						END
					ELSE
						userdata.mapper.ti.Find(nstack[nlevel].name, ritem, userdata.items[nl])
					END;
				END;
				IF (parse_flag = Tp.COUPLE_END) OR (parse_flag = Tp.COUPLE_BOTH) THEN
					IF (userdata.items[nl].typ = 0) OR (userdata.items[nl].typ = Tb.ptrTyp) 
					OR (userdata.items[nl].typ = Tb.recTyp) THEN
					ELSE
						Ti.ItemVal_String(userdata.items[nl], FALSE, oldstr, ok);
						IF ~ok OR ~Tb.Equal(value, oldstr) THEN
							userdata.changed := TRUE;
							Ti.String_ItemVal(userdata.items[nl], value, ok);
						END;
						IF ~ok THEN
							IF ~userdata.mapper.tqos.nc_ignore THEN
								res := PARSE_STR
							END
						END
					END
				END;
				IF parse_flag = Tp.COUPLE_DIM THEN
				END
			END;
			(*IF Tb.TRAL_MESSAGE IN Tb.trace THEN
				IF res # 0 THEN Log.Char('*'); Log.Int(res) END;
				Log.String("onread"); Log.Int(parse_flag); Log.Int(nlevel); Log.String('# ');
				Log.String(nstack[nlevel].name); Log.Char(' ');
				Log.String(value); Log.Int(nstack[nlevel].index); Log.Ln;
			END*)
		END
	END OnReadSoap;
	
	PROCEDURE UnquoteAndParseXML (mapper: Mapper; VAR item: Ti.Item; 
	VAR s: Tm.Scanner; VAR res: INTEGER; VAR changed: BOOLEAN);
		VAR pld: ParseData;
	BEGIN
		pld.mapper := mapper; pld.item := item; pld.changed := FALSE;
		CASE mapper.tqos.match_mode OF
		| SOAP_MATCH_INTERFACE:
			pld.nlevel0 := 3;
		| SOAP_11_CLIENT, SOAP_12_CLIENT:
			pld.nlevel0 := 0;
		| SOAP_GET_CLIENT, SOAP_POST_CLIENT:
			pld.nlevel0 := 0;
		END;
		Tp.ParseXML(s, OnReadSoap, pld, res);
		changed := pld.changed
	END UnquoteAndParseXML;
	
	PROCEDURE UnquoteAndParseLine (mapper: Mapper; VAR item: Ti.Item; 
	VAR s: Tm.Scanner; VAR res: INTEGER; VAR changed: BOOLEAN);
		VAR pld: ParseData;
	BEGIN
		pld.mapper := mapper; pld.item := item; pld.changed := FALSE; pld.nlevel0 := 0;
		Tp.ParseLine(s, OnReadSoap, pld, res);
		changed := pld.changed
	END UnquoteAndParseLine;
	
	PROCEDURE (mapper: Mapper) ReadObject (VAR item: Ti.Item; VAR rd: Ts.Reader;
	VAR changed: BOOLEAN);
		CONST OPTS = {Tm.withCrLf, Tm.separateStrings, Tm.bracesAsString, 
				Tm.unquoteAmp};
			STR_OPTS = {Tm.withCrLf, Tm.separateStrings, Tm.bracesAsString, 
				Tm.unquoteAmp, Tm.specialAsString};
		VAR cr: Carrier; s: Tm.Scanner; oldtype, res: INTEGER; name: Tb.Name;
	BEGIN
		ASSERT(rd.cr IS Carrier, Tb.INVALID_EXTENSION);
		cr := rd.cr(Carrier);
		cr.rd.srd := rd;
		rd.SetPos(0); res := PARSE_NOT_FOUND;
		s.ConnectTo(cr.rd);
		s.SetOpts(OPTS);
		s.Scan;
		mapper.mid_read := rd.hdr.mid; mapper.xmlns_in[0] := 0X;
		CASE rd.hdr.mid OF
		| SOAP_11_MID, SOAP_12_MID:
			WHILE (s.type # Tm.eot) & (s.string # "?>") DO
				IF s.string = "Host" THEN
					s.Scan; s.SetOpts(STR_OPTS); s.Scan;
					Tb.Copy(s.string, mapper.xmlns_in);
					s.SetOpts(OPTS)
				END;
				s.Scan
			END;
			IF s.type # Tm.eot THEN
				UnquoteAndParseXML(mapper, item, s, res, changed)
			END;
		| SOAP_GET_MID:
			WHILE s.type # Tm.eot DO
				IF s.string = "?" THEN
					UnquoteAndParseLine(mapper, item, s, res, changed)
				ELSIF s.string = "Host" THEN
					s.Scan; s.SetOpts(STR_OPTS); s.Scan;
					Tb.Copy(s.string, mapper.xmlns_in);
					s.SetOpts(OPTS)
				END;
				s.Scan
			END;
		| SOAP_POST_MID:
			oldtype := s.type;
			WHILE (s.type # Tm.eot) & ((s.type # Tm.crlf) OR (oldtype # Tm.crlf)) DO 
				IF s.string = "Host" THEN
					s.Scan; s.SetOpts(STR_OPTS); s.Scan;
					Tb.Copy(s.string, mapper.xmlns_in);
					s.SetOpts(OPTS)
				END;
				oldtype := s.type; s.Scan
			END;
			IF s.type # Tm.eot THEN
				UnquoteAndParseLine(mapper, item, s, res, changed)
			END;
		| SOAP_11_CLIENT, SOAP_12_CLIENT:
			Tb.Copy(mapper.typ, name);
			Tb.Append("Result", name);
			WHILE (s.type # Tm.eot) & ~Tb.Equal(s.string, name) DO s.Scan END;
			IF s.type # Tm.eot THEN
				UnquoteAndParseXML(mapper, item, s, res, changed)
			END;
		| SOAP_POST_CLIENT:
			WHILE (s.type # Tm.eot) & (s.string # "string") DO s.Scan END;
			WHILE (s.type # Tm.eot) & (s.string # "?>") DO s.Scan END;
			IF s.type # Tm.eot THEN
				UnquoteAndParseXML(mapper, item, s, res, changed)
			END;
		| SOAP_MID:
			name := "append to the previous buffer";
		ELSE
			res := PARSE_WRONG_MID
		END;
		IF mapper.tqos.match_mode = SOAP_MATCH_INTERFACE THEN
			IF (res = 0) & (mapper.xmlns_in[0] # 0X) & (mapper.tqos.xmlns[0] # 0X) THEN
				IF ~Tb.CapEqual(mapper.xmlns_in, mapper.tqos.xmlns) THEN
					res := PARSE_WRONG_NAME
				END
			END
		END;
		IF res # 0 THEN
			IF (Tb.TRAL_MESSAGE IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* Read Object "); Log.String(mapper.xmlns_in); 
				Log.Int(res); Log.Int(rd.hdr.mid); Log.Ln
			END
		ELSE
			IF Tb.TRAL_MESSAGE IN Tb.trace THEN
				Log.String("Read Object "); Log.String(mapper.xmlns_in); 
				Log.Int(res); Log.Int(rd.hdr.mid); Log.Ln
			END
		END;
		mapper.waiting_for_response := FALSE
	END ReadObject;

	PROCEDURE (mapper: Mapper) WriteHeader (VAR hdr: Ts.Header;  wr-: Ts.Writer);
	BEGIN
		CASE mapper.tqos.match_mode OF
		SOAP_11_CLIENT, SOAP_12_CLIENT, SOAP_GET_CLIENT, SOAP_POST_CLIENT:
			hdr.mid := SOAP_MID;
		ELSE
		END
	END WriteHeader;
	
	PROCEDURE WriteXMLItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR wrd: Tb.AnyRec): INTEGER;
		VAR rc, res: INTEGER; str: Tb.Name; ok: BOOLEAN; iit: Ti.Item;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		(*IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("xmlwrite "); Log.String(name); Log.Int(item.typ); Log.Int(item.adr);
			Log.Bool(begin); Log.Ln
		END;*)
		WITH wrd: WriteData DO
			CASE item.typ OF
			Tb.recTyp:
				IF begin THEN
					rc := Ti.ITERATE_INTO
				END;
				IF Tb.Equal(name, wrd.mapper.typ) THEN
				ELSIF begin THEN 
					wrd.f.WriteChar('<'); wrd.f.WriteString(name); wrd.f.WriteChar(">"); wrd.f.WriteLn
				ELSE
					wrd.f.WriteString('</'); wrd.f.WriteString(name); wrd.f.WriteChar(">"); wrd.f.WriteLn
				END
			| Tb.arrTyp:
				item.Index(0, iit);
				IF (iit.typ = Tb.char8Typ) OR (iit.typ = Tb.char16Typ) THEN
					Ti.ItemVal_String(item, TRUE, str, ok);
					IF ~ok THEN
						rc := Ti.STOP_ITERATE
					ELSIF str[0] # 0X THEN
						wrd.f.WriteChar('<'); wrd.f.WriteString(name); wrd.f.WriteChar(">");
						wrd.f.SetOpts(WRITE_OPTS+{Tm.unquoteAmp}); wrd.f.WriteString(str);
						wrd.f.SetOpts(WRITE_OPTS);
						wrd.f.WriteString('</'); wrd.f.WriteString(name); wrd.f.WriteChar(">"); 
						wrd.f.WriteLn
					END
				ELSE
					IF begin THEN
						rc := Ti.ITERATE_FOREACH
					END
				END;
			| Tb.ptrTyp:
				IF begin THEN 
					rc := Ti.ITERATE_INTO
				END;
			ELSE
				Ti.ItemVal_String(item, TRUE, str, ok);
				IF ~ok THEN
					IF (Tb.TRAL_MESSAGE IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* writeXml from "); Log.String(name);
						Log.Int(item.typ); Log.Ln
					END;
					rc := Ti.STOP_ITERATE
				ELSIF str[0] # 0X THEN
					wrd.f.WriteChar('<'); wrd.f.WriteString(name); wrd.f.WriteChar(">");
					wrd.f.SetOpts(WRITE_OPTS+{Tm.unquoteAmp}); wrd.f.WriteString(str);
					wrd.f.SetOpts(WRITE_OPTS);
					wrd.f.WriteString('</'); wrd.f.WriteString(name); wrd.f.WriteChar(">"); 
					wrd.f.WriteLn
				END
			END;
		END;
		RETURN rc
	END WriteXMLItem;
	
	PROCEDURE WriteXMLRec (mapper: Mapper;  item-: Ti.Item; VAR f: Tm.Formatter);
		VAR wrd: WriteData; res: INTEGER;
	BEGIN
		wrd.mapper := mapper; wrd.f := f;
		mapper.ti.Iterate(WriteXMLItem, mapper.typ, item, wrd, res);
		f := wrd.f;
		mapper.cancelled := (res = Ti.STOP_ITERATE);
	END WriteXMLRec;
	
	PROCEDURE WriteLineItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR wrd: Tb.AnyRec): INTEGER;
		VAR rc: INTEGER; str: Tb.Name; ok: BOOLEAN; iit: Ti.Item;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		(*IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("linewrite "); Log.String(name); Log.Int(item.typ); Log.Bool(begin); Log.Ln
		END;*)
		WITH wrd: WriteData DO
			CASE item.typ OF
			Tb.recTyp:
				IF begin THEN
					rc := Ti.ITERATE_INTO
				END;
			| Tb.arrTyp:
				item.Index(0, iit);
				IF (iit.typ = Tb.char8Typ) OR (iit.typ = Tb.char16Typ) THEN
					Ti.ItemVal_String(item, TRUE, str, ok);
					IF ~ok THEN
						rc := Ti.STOP_ITERATE
					ELSIF str[0] # 0X THEN
						IF ~wrd.first THEN wrd.f.WriteChar('&') END; wrd.first := FALSE;
						wrd.f.WriteString(name); wrd.f.WriteChar("="); 
						wrd.f.SetOpts(WRITE_OPTS+{Tm.unquotePercent}); wrd.f.WriteString(str); 
						wrd.f.SetOpts(WRITE_OPTS)
					END;
				ELSE
					IF begin THEN
						(* rc := Ti.ITERATE_FOREACH # No good idea yet *)
					END
				END;
			| Tb.ptrTyp:
				IF begin THEN 
					rc := Ti.ITERATE_INTO
				END;
			ELSE
				Ti.ItemVal_String(item, TRUE, str, ok);
				IF ~ok THEN
					rc := Ti.STOP_ITERATE
				ELSIF str[0] # 0X THEN
					IF ~wrd.first THEN wrd.f.WriteChar('&') END; wrd.first := FALSE;
					wrd.f.WriteString(name); wrd.f.WriteChar("="); 
					wrd.f.SetOpts(WRITE_OPTS+{Tm.unquotePercent}); wrd.f.WriteString(str); 
					wrd.f.SetOpts(WRITE_OPTS)
				END;
			END;
		END;
		RETURN rc
	END WriteLineItem;
	
	PROCEDURE WriteLineRec (mapper: Mapper;  item-: Ti.Item; VAR f: Tm.Formatter);
		VAR wrd: WriteData; res: INTEGER;
	BEGIN
		wrd.mapper := mapper; wrd.f := f; wrd.first := TRUE;
		mapper.ti.Iterate(WriteLineItem, mapper.typ, item, wrd, res);
		f := wrd.f;
		mapper.cancelled := (res = Ti.STOP_ITERATE);
	END WriteLineRec;
	
	PROCEDURE (mapper: Mapper) WriteObject ( item-: Ti.Item; 
	VAR wr: Ts.Writer);
		CONST RCODE_OK = 200;
		VAR cr: Carrier; f: Tm.Formatter; rcode, nlen, n, n0: Tb.Int32; xmlns: Tb.Name64;
	BEGIN
		ASSERT(wr.cr IS Carrier, Tb.INVALID_EXTENSION);
		cr := wr.cr(Carrier);
		cr.wr.swr := wr;
		cr.wr.swr.SetPos(0);
		f.ConnectTo(cr.wr); f.SetOpts({Tm.withoutLf});
		rcode := RCODE_OK; n0 := 0; nlen := 0;
		IF mapper.tqos.xmlns[0] # 0X THEN
			xmlns := mapper.tqos.xmlns
		ELSE
			xmlns := mapper.xmlns_in
		END;
		CASE mapper.tqos.match_mode OF
		| SOAP_MATCH_INTERFACE:
			CASE mapper.mid_read OF
			| SOAP_11_MID:
				f.WriteString("HTTP/1.1 "); f.WriteInt(rcode); f.WriteString(" OK"); f.WriteLn;
				f.WriteString("Content-Type: text/xml; charset=utf-8"); f.WriteLn;
				f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6; 
				f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
				f.WriteString("<?xml version='1.0' encoding='utf-8'?>"); f.WriteLn;
				f.WriteString("<soap:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' ");
				f.WriteString("xmlns:xsd='http://www.w3.org/2001/XMLSchema' ");
				f.WriteString("xmlns:soap='http://schemas.xmlsoap.org/soap/envelope/'>"); f.WriteLn;
				f.WriteString("  <soap:Body>"); f.WriteLn; f.WriteString("    <"); f.WriteString(mapper.typ);
				f.WriteString("Response xmlns='http://"); f.WriteString(xmlns); f.WriteString("'>");
				f.WriteLn; f.WriteString("      <"); f.WriteString(mapper.typ); f.WriteString("Result>"); f.WriteLn;
				WriteXMLRec(mapper, item, f);
				f.WriteString("      </"); f.WriteString(mapper.typ); f.WriteString("Result>"); f.WriteLn;
				f.WriteString("    </"); f.WriteString(mapper.typ); f.WriteString("Response>"); f.WriteLn;
				f.WriteString("  </soap:Body>"); f.WriteLn; f.WriteString("</soap:Envelope>"); f.WriteLn; 
			| SOAP_12_MID:
				f.WriteString("HTTP/1.1 "); f.WriteInt(rcode); f.WriteString(" OK"); f.WriteLn;
				f.WriteString("Content-Type: application/soap+xml; charset=utf-8"); f.WriteLn;
				f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6; 
				f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
				f.WriteString("<?xml version='1.0' encoding='utf-8'?>"); f.WriteLn;
				f.WriteString("<soap12:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' ");
				f.WriteString("xmlns:xsd='http://www.w3.org/2001/XMLSchema' ");
				f.WriteString("xmlns:soap12='http://www.w3.org/2003/05/soap-envelope'>"); f.WriteLn;
				f.WriteString("  <soap12:Body>"); f.WriteLn; f.WriteString("    <"); f.WriteString(mapper.typ);
				f.WriteString("Response xmlns='http://"); f.WriteString(xmlns); f.WriteString("'>");
				f.WriteLn; f.WriteString("      <"); f.WriteString(mapper.typ); f.WriteString("Result>"); f.WriteLn;
				WriteXMLRec(mapper, item, f);
				f.WriteString("      </"); f.WriteString(mapper.typ); f.WriteString("Result>"); f.WriteLn;
				f.WriteString("    </"); f.WriteString(mapper.typ); f.WriteString("Response>"); f.WriteLn;
				f.WriteString("  </soap12:Body>"); f.WriteLn; f.WriteString("</soap12:Envelope>"); f.WriteLn; 
			| SOAP_GET_MID, SOAP_POST_MID:
				f.WriteString("HTTP/1.1 "); f.WriteInt(rcode); f.WriteString(" OK"); f.WriteLn;
				f.WriteString("Content-Type: text/xml; charset=utf-8"); f.WriteLn;
				f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6; 
				f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
				f.WriteString("<?xml version='1.0' encoding='utf-8'?>"); f.WriteLn;
				f.WriteString("<string xmlns='http://"); f.WriteString(xmlns); f.WriteString("'>"); f.WriteLn;
				WriteXMLRec(mapper, item, f);
				f.WriteString("</string>"); f.WriteLn; 
			ELSE
			END
		| SOAP_11_CLIENT:
			f.WriteString("POST /"); f.WriteString(mapper.iface_name); f.WriteString(" HTTP/1.1"); 
			f.WriteLn; f.WriteString("Host: "); f.WriteString(mapper.tqos.xmlns); f.WriteLn;
			f.WriteString("Content-Type: text/xml; charset=utf-8"); f.WriteLn;
			f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6;  f.WriteLn; 
			f.WriteString('SOAPAction: "http://'); f.WriteString(mapper.tqos.xmlns); f.WriteChar('/'); 
			f.WriteString(mapper.typ); f.WriteChar('"'); f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
			f.WriteString("<?xml version='1.0' encoding='utf-8'?>"); f.WriteLn;
			f.WriteString("<soap:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' ");
			f.WriteString("xmlns:xsd='http://www.w3.org/2001/XMLSchema' ");
			f.WriteString("xmlns:soap='http://schemas.xmlsoap.org/soap/envelope/'>"); f.WriteLn;
			f.WriteString("  <soap:Body>"); f.WriteLn; f.WriteString("    <"); f.WriteString(mapper.typ);
			f.WriteString(" xmlns='http://"); f.WriteString(mapper.tqos.xmlns); f.WriteString("'>"); f.WriteLn;
			WriteXMLRec(mapper, item, f);
			f.WriteString("    </"); f.WriteString(mapper.typ); f.WriteChar(">"); f.WriteLn;
			f.WriteString("  </soap:Body>"); f.WriteLn; f.WriteString("</soap:Envelope>"); f.WriteLn; 
			mapper.waiting_for_response := TRUE;
		| SOAP_12_CLIENT:
			f.WriteString("POST /"); f.WriteString(mapper.iface_name); f.WriteString(" HTTP/1.1"); 
			f.WriteLn; f.WriteString("Host: "); f.WriteString(mapper.tqos.xmlns); f.WriteLn;
			f.WriteString("Content-Type: application/soap+xml; charset=utf-8"); f.WriteLn;
			f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6;  f.WriteLn; 
			f.WriteString('SOAPAction: "http://'); f.WriteString(mapper.tqos.xmlns); f.WriteChar('/'); 
			f.WriteString(mapper.typ); f.WriteChar('"'); f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
			f.WriteString("<?xml version='1.0' encoding='utf-8'?>"); f.WriteLn;
			f.WriteString("<soap12:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' ");
			f.WriteString("xmlns:xsd='http://www.w3.org/2001/XMLSchema' ");
			f.WriteString("xmlns:soap12='http://www.w3.org/2003/05/soap-envelope'>"); f.WriteLn;
			f.WriteString("  <soap12:Body>"); f.WriteLn; f.WriteString("    <"); f.WriteString(mapper.typ);
			f.WriteString(" xmlns='http://"); f.WriteString(mapper.tqos.xmlns); f.WriteString("'>"); f.WriteLn;
			WriteXMLRec(mapper, item, f);
			f.WriteString("    </"); f.WriteString(mapper.typ); f.WriteChar(">"); f.WriteLn;
			f.WriteString("  </soap12:Body>"); f.WriteLn; f.WriteString("</soap12:Envelope>"); f.WriteLn; 
			mapper.waiting_for_response := TRUE;
		| SOAP_POST_CLIENT:
			f.WriteString("POST /"); f.WriteString(mapper.iface_name); f.WriteChar("/");
			f.WriteString(mapper.typ); f.WriteString(" HTTP/1.1"); 
			f.WriteLn; f.WriteString("Host: "); f.WriteString(mapper.tqos.xmlns); f.WriteLn;
			f.WriteString("Accept: text/html, application/xhtml+xml, */*"); f.WriteLn;
			f.WriteString("Cache-Control: no-cache"); f.WriteLn;
			f.WriteString("Content-Type: application/x-www-form-urlencoded"); f.WriteLn;
			f.WriteString("Content-Length:       "); f.FlushBuf; nlen := cr.wr.swr.pos-6;  
			f.WriteLn; f.WriteLn; n0 := cr.wr.swr.pos;
			WriteLineRec(mapper, item, f);
			mapper.waiting_for_response := TRUE;
		| SOAP_GET_CLIENT:
			f.WriteString("GET /"); f.WriteString(mapper.iface_name); f.WriteChar("/");
			f.WriteString(mapper.typ); f.WriteChar("?"); 
			WriteLineRec(mapper, item, f);
			f.WriteString(" HTTP/1.1"); f.WriteLn;
			f.WriteString("Accept: text/html, application/xhtml+xml, */*"); f.WriteLn;
			f.WriteString("Host: "); f.WriteString(mapper.tqos.xmlns); f.WriteLn; 
			f.WriteLn;
			mapper.waiting_for_response := TRUE;
		END;
		IF ~mapper.cancelled THEN
			f.FlushBuf;
			n := cr.wr.swr.pos;
			IF (nlen > 0) & (n0 > 0) THEN
				cr.wr.swr.SetPos(nlen);
				f.WriteInt(n-n0); f.FlushBuf
			END
		END;
		wr := cr.wr.swr;
	END WriteObject;

	PROCEDURE (mapper: Mapper) HandleMsg (VAR msg: Tb.Msg);
	BEGIN
		WITH msg: TableQos DO
			ASSERT((msg.match_mode >= SOAP_MATCH_INTERFACE)
				& (msg.match_mode <= SOAP_POST_CLIENT), Tb.CHECK_RANGE_ERROR);
			mapper.tqos.xmlns := msg.xmlns;
			IF mapper.tqos.match_mode # msg.match_mode THEN
				mapper.tqos.match_mode := msg.match_mode;
				mapper.ExploreObject
			END;
			mapper.tqos.nc_ignore := msg.nc_ignore
		ELSE
		END
	END HandleMsg; 

	(* ========== Initialize ========== *)
	PROCEDURE (mf: MapperFactory) NewMapper* (ti: Ti.TInfo;
	 iface_name-: ARRAY OF CHAR; VAR tqos: Ts.TableQosBase;
	sid, oid: INTEGER; (*OUT*)VAR mapr: Ts.Mapper; (*OUT*)VAR res: INTEGER);
		VAR mapper: Mapper;
	BEGIN
		NEW(mapper);
		mapper.ti := ti; Tb.Copy(iface_name, mapper.iface_name);
		WITH tqos: TableQos DO
			mapper.tqos := tqos
		ELSE
			mapper.tqos := defaultTQos;
			Ts.CopyTableQos(tqos, mapper.tqos)
		END;
		ASSERT((mapper.tqos.match_mode >= SOAP_MATCH_INTERFACE)
			& (mapper.tqos.match_mode <= SOAP_POST_CLIENT), Tb.CHECK_RANGE_ERROR);
		mapper.xmlns_in[0] := 0X;
		mapr := mapper; res := 0
	END NewMapper;

BEGIN
	Ts.DefaultTableQos(defaultTQos);
	defaultTQos.match_mode := SOAP_MATCH_INTERFACE;
	defaultTQos.xmlns := "127.0.0.1";
	defaultTQos.nc_ignore := FALSE;
END Ta_SOAPStations.