MODULE Ta_SRPSStations;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ti := Ta_TInfo, Ts := Ta_Stations, 
		SYSTEM, Log;
	
	CONST 
		HEADER_SIZE = 32; MAX_DEPTH = 16; PTR_SIZE = 16;
		
	TYPE
		Carrier = POINTER TO CarrierDesc;
		Mapper = POINTER TO MapperDesc;

		SRPSHeader = RECORD
			protname: ARRAY 4 OF Tb.Int8;
			major_version, spec_bits: Tb.Int8;
			iface_code: Tb.Int16;
			msg_type: Tb.Int32;
			dst_oid, basic_length: Tb.Int16;
			length: Tb.Int32;
			mid, mid_resp: Tb.Int16;
			aid: Tb.Int32;
			src_sid, dst_sid: Tb.Int16;
		END;
		
		PtrData = POINTER TO PtrDataDesc;
		PtrDataDesc = RECORD
			nbytes, offset, msg_type, item_type: Tb.Int32;
			dim, iface_code: Tb.Int16;
			next: PtrData
		END;

		CarrierDesc = RECORD (Ts.CarrierDesc)
			aid: INTEGER;
		END;
		
		CarrierFactoryDesc* = RECORD (Ts.CarrierFactoryDesc)
		END;		
		CarrierFactory* = POINTER TO CarrierFactoryDesc;

		MapperDesc = RECORD (Ts.MapperDesc)
			ti: Ti.TInfo;
			iface_name: Tb.Name64;
			sid, oid, realloc_mode: INTEGER;
			msg_type, basic_length, buffer_size, dyn_buffer_size: Tb.Int32;
			iface_code: Tb.Int16;
			ptr_init, tail_init, ptr_data, tail_data: PtrData;
		END;
		
		MapperFactoryDesc* = RECORD (Ts.MapperFactoryDesc)
		END;		
		MapperFactory* = POINTER TO MapperFactoryDesc;
		
		ExploreData = RECORD (Tb.AnyRec)
			mapper: Mapper;
			msg_type: Tb.Int32;
			iface_code: Tb.Int16;
			ptr_init, tail_init: PtrData;
			packed_size, packed_size0: Tb.Int32;
		END;
		
		ReadData = RECORD (Tb.AnyRec)
			mapper: Mapper;
			rd: Ts.Reader;
			dyn_pos, basic_length: Tb.Int32;
			changed: BOOLEAN;
		END;
		
		PackedSizeData = RECORD (Tb.AnyRec)
			size: Tb.Int32;
			presizes: ARRAY MAX_DEPTH OF Tb.Int32;
			ind: INTEGER
		END;
		
		WriteData = RECORD (Tb.AnyRec)
			mapper: Mapper;
			wr: Ts.Writer;
			dyn_buffer_size, dyn_pos: Tb.Int32;
		END;
		

	VAR
		sheader: SRPSHeader;


	
	(* ========== Carrier Factory ========== *)
	PROCEDURE (crf: CarrierFactory) NewCarrier* (pf: Tc.PollingFactory; 
	aid, n_stations: INTEGER; send_buffer_size, recv_buffer_size: Tb.Int32;  
	prot_format: ARRAY OF CHAR; (*OUT*)VAR cr: Ts.Carrier; (*OUT*)VAR res: INTEGER);
		VAR carrier: Carrier; polling: Tc.Polling;
	BEGIN
		ASSERT(n_stations > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(pf # NIL, Tb.NIL_POINTER);
		pf.NewPolling(n_stations, polling, res);
		IF res = 0 THEN
			NEW(carrier); carrier.aid := aid;
			carrier.SetNewCarrier(polling, send_buffer_size, recv_buffer_size, prot_format)
		END;
		cr := carrier
	END NewCarrier;

	(* ========== Carrier Interface ========== *)
	PROCEDURE (cr: Carrier) InternalizeHeader (VAR rd: Ts.Reader; (*OUT*)VAR hdr: Ts.Header; 
	(*OUT*)VAR res: Tb.Int32);
		VAR sh: SRPSHeader; x8: Tb.Bits8; x16: Tb.Bits16; x32: Tb.Bits32;
	BEGIN
		res := 0;
		rd.ReadBytes(sh.protname, 0, 4);
		rd.Read8(x8); sh.major_version := SYSTEM.VAL(Tb.Int8, x8);
		IF (sh.protname[0] # sheader.protname[0]) OR (sh.protname[1] # sheader.protname[1])
		OR (sh.protname[2] # sheader.protname[2]) OR (sh.protname[3] # sheader.protname[3])
		OR (sh.major_version # sheader.major_version) THEN
			res := Ts.WRONG_HDR_DATA;
		ELSE
			rd.Read8(x8); sh.spec_bits := SYSTEM.VAL(Tb.Int8, x8);
			rd.Read16(x16); sh.iface_code := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read32(x32); sh.msg_type := SYSTEM.VAL(Tb.Int32, x32);
			rd.Read16(x16); sh.dst_oid := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read16(x16); sh.basic_length := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read32(x32); sh.length := SYSTEM.VAL(Tb.Int32, x32);
			rd.Read16(x16); sh.mid := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read16(x16); sh.mid_resp := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read32(x32); sh.aid := SYSTEM.VAL(Tb.Int32, x32);
			rd.Read16(x16); sh.src_sid := SYSTEM.VAL(Tb.Int16, x16);
			rd.Read16(x16); sh.dst_sid := SYSTEM.VAL(Tb.Int16, x16);
		END;
		IF rd.eof THEN
			res := Ts.CANT_READ_HDR
		END;
		IF res = 0 THEN
			hdr.mid := sh.mid; hdr.mid_resp := sh.mid_resp; hdr.src_sid := sh.src_sid;
			hdr.dst_sid := sh.dst_sid; hdr.dst_oid := sh.dst_oid; hdr.iface_code := sh.iface_code;
			hdr.msg_type := sh.msg_type; hdr.buffer := 0; hdr.buffer_length := sh.length;
			hdr.basic_length := sh.basic_length;
			hdr.tlg_type := sh.spec_bits;
			hdr.aid := SYSTEM.VAL(Tb.Int16, sh.aid);
		END
	END InternalizeHeader;

	PROCEDURE (cr: Carrier) ExternalizeHeader (VAR wr: Ts.Writer;  hdr-: Ts.Header;
	(*OUT*)VAR res: Tb.Int32);
		VAR sh: SRPSHeader;
	BEGIN
		sh := sheader;
		sh.iface_code := hdr.iface_code; sh.msg_type := hdr.msg_type; sh.dst_oid := hdr.dst_oid;
		sh.basic_length := SYSTEM.VAL(Tb.Int16, hdr.basic_length);
		sh.length := hdr.buffer_length;
		sh.mid := hdr.mid; sh.mid_resp := hdr.mid_resp;
		sh.aid := hdr.aid; sh.src_sid := Tb.Int_Int16(cr.host_sid); sh.dst_sid := hdr.dst_sid;
		IF hdr.tlg_type = Ts.PING_TELEGRAM_TYPE THEN
			sh.spec_bits := Ts.PING_TELEGRAM_TYPE
		END;
		wr.WriteBytes(sh.protname, 0, 4);
		wr.Write8(SYSTEM.VAL(Tb.Bits8, sh.major_version));
		wr.Write8(SYSTEM.VAL(Tb.Bits8, sh.spec_bits));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.iface_code));
		wr.Write32(SYSTEM.VAL(Tb.Bits32, sh.msg_type));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.dst_oid));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.basic_length));
		wr.Write32(SYSTEM.VAL(Tb.Bits32, sh.length));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.mid));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.mid_resp));
		wr.Write32(SYSTEM.VAL(Tb.Bits32, sh.aid));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.src_sid));
		wr.Write16(SYSTEM.VAL(Tb.Bits16, sh.dst_sid));
	END ExternalizeHeader;
	
	(* ========== Mapper interface ========== *)
	PROCEDURE ExploreItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR ed: Tb.AnyRec): INTEGER;
		VAR rc, res: INTEGER; do: BOOLEAN; mod, typ: Ti.Name;
		ptr: PtrData; base: Ti.Item; edp: ExploreData; o_msg: Tb.Int32; o_iface: Tb.Int16;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		WITH ed: ExploreData DO
			do := TRUE;
			CASE item.typ OF
			Tb.recTyp:
				IF begin THEN rc := Ti.ITERATE_INTO; do := FALSE END
			| Tb.arrTyp:
				IF begin THEN 
					rc := Ti.ITERATE_INTO; ed.packed_size0 := ed.packed_size
				ELSE
					ed.packed_size := ed.packed_size0 + 
						(ed.packed_size - ed.packed_size0) * item.Len();
				END;
				do := FALSE
			| Tb.ptrTyp:
				IF begin THEN 
					IF Ti.PtrLen(item) = 0 THEN
						Ti.MakeArr(item, 1)
					END;
					NEW(ptr); ptr.item_type := Ti.ItemTypOf(item); 
					IF ed.ptr_init = NIL THEN
						ed.ptr_init := ptr; ed.tail_init := ptr
					ELSE
						ed.tail_init.next := ptr; ed.tail_init := ptr
					END;
					edp := ed;
					edp.msg_type := 0; edp.iface_code := 0;
					item.Deref(base); res := 0;
					ed.mapper.ti.Iterate(ExploreItem, name, base, edp, res);
					ed.tail_init := edp.tail_init;
					ptr.msg_type := edp.msg_type; ptr.iface_code := edp.iface_code;
					IF Tb.TRAL_MESSAGE IN Tb.trace THEN
						item.GetTypeName(mod, typ);
						Log.String("ptr "); Log.String(name); Log.Int(item.typ); Log.Char(' '); Log.String(typ);
						Log.Char(' '); Log.Int(edp.msg_type); Log.Int(edp.iface_code); Log.Ln
					END
				END;
				INC(ed.packed_size, 16)
			ELSE
				INC(ed.packed_size, item.Size())
			END;
			IF do THEN
				IF item.typ = Tb.recTyp THEN
					item.GetTypeName(mod, typ)
				ELSE
					Tb.Copy(Tb.compatible_types[item.typ], typ)
				END;
				o_msg := ed.msg_type; o_iface := ed.iface_code;
				ed.msg_type := Tb.Crc32(name, -1, ed.msg_type, 32);
				ed.iface_code := SYSTEM.VAL(Tb.Int16, Tb.Crc32(typ, -1, 
					ed.iface_code, 16));
				IF Tb.TRAL_MESSAGE IN Tb.trace THEN
					Log.String(name); Log.Int(item.typ); Log.Char(' '); Log.String(typ);
					Log.Char(' '); Log.Int(item.adr-ed.mapper.ti.item.adr);
					Log.Int(ed.msg_type); Log.Int(ed.iface_code); 
					Log.String(" old"); Log.Int(o_msg); Log.Int(o_iface); Log.Ln
				END
			END
		END;
		RETURN rc
	END ExploreItem;
	
	PROCEDURE (mapper: Mapper) ExploreObject (); 
		VAR res: INTEGER; ed: ExploreData;
	BEGIN
		ed.mapper := mapper; ed.ptr_init := NIL; ed.tail_init := NIL;
		ed.msg_type := 0; ed.iface_code := 0; 
		ed.packed_size := 0; ed.packed_size0 := 0;
		mapper.ti.Iterate(ExploreItem, mapper.iface_name, mapper.ti.item, ed, res);
		mapper.buffer_size := ed.packed_size;
		mapper.ptr_init := ed.ptr_init; mapper.tail_init := ed.tail_init;
		mapper.ptr_data := NIL; mapper.tail_data := NIL;
		mapper.msg_type := ed.msg_type; mapper.iface_code := ed.iface_code; 
		mapper.cancelled := (res = Ti.STOP_ITERATE);
		IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("Table [size,msg_type,iface_code]="); Log.Int(mapper.buffer_size);
			Log.Int(mapper.msg_type); Log.Int(mapper.iface_code); Log.Ln;
		END;
	END ExploreObject;

	PROCEDURE (mapper: Mapper) MatchHeader ( hdr-: Ts.Header): BOOLEAN;
		VAR match: BOOLEAN;
	BEGIN
		match :=  (hdr.basic_length - HEADER_SIZE = mapper.buffer_size)
			& (hdr.msg_type = mapper.msg_type) & (hdr.iface_code = mapper.iface_code);
		IF mapper.sid >= 0 THEN
			IF (hdr.src_sid >= 0) & (hdr.src_sid # mapper.sid) THEN
				match := FALSE
			END
		END;
		IF mapper.oid >= 0 THEN
			IF (hdr.dst_oid >= 0) & (hdr.dst_oid # mapper.oid) THEN
				match := FALSE
			END
		END;
		IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("SRPS Match="); Log.Bool(match); Log.Int(hdr.mid);
			Log.String("# l="); Log.Int(hdr.basic_length-HEADER_SIZE); Log.Int(mapper.buffer_size); 
			Log.String(" m="); Log.Int(hdr.msg_type); Log.Int(mapper.msg_type); 
			Log.String(" i="); Log.Int(hdr.iface_code); Log.Int(mapper.iface_code); 
			Log.String(" s="); Log.Int(mapper.sid); Log.Int(hdr.src_sid);
			Log.String(" o="); Log.Int(mapper.oid); Log.Int(hdr.dst_oid); Log.Ln;
		END;
		RETURN match
	END MatchHeader;
	
	PROCEDURE ReadBytes ( x-: ARRAY OF Tb.Int8; adr: Tb.Address; 
	len: INTEGER; VAR changed: BOOLEAN);
		VAR j: INTEGER; byte: Tb.Int8;
	BEGIN
		FOR j := 0 TO len-1 DO
			SYSTEM.GET(adr+j, byte);
			IF byte # x[j] THEN
				SYSTEM.PUT(adr+j, x[j]);
				changed := TRUE
			END
		END
	END ReadBytes;

	PROCEDURE GetPtrData (mapper: Mapper;  item-: Ti.Item): PtrData;
		VAR ptr: PtrData; found: BOOLEAN; item_type: Tb.Int32;
	BEGIN
		ptr := mapper.ptr_init; found := FALSE;
		item_type := Ti.ItemTypOf(item);
		WHILE ~found & (ptr # NIL) DO
			IF item_type = ptr.item_type THEN
				found := TRUE
			ELSE
				ptr := ptr.next
			END
		END;
		RETURN ptr
	END GetPtrData;

	PROCEDURE ReadItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR rda: Tb.AnyRec): INTEGER;
		VAR rc: INTEGER; x8: Tb.Bits8; x16: Tb.Bits16; x32: Tb.Bits32; x64: Tb.Bits64;
		size, offset, msg_type: Tb.Int32; iface_code, l, len: Tb.Int16; ptr: PtrData;
		base: Ti.Item; rdp: ReadData;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		WITH rda: ReadData DO
			CASE item.typ OF
			| Tb.boolTyp, Tb.int8Typ, Tb.char8Typ:
				rda.rd.Read8(x8);
				ReadBytes(x8, item.adr, 1, rda.changed);
			| Tb.int16Typ, Tb.char16Typ:
				rda.rd.Read16(x16);
				ReadBytes(x16, item.adr, 2, rda.changed);
			| Tb.int32Typ, Tb.real32Typ:
				rda.rd.Read32(x32);
				ReadBytes(x32, item.adr, 4, rda.changed);
			| Tb.int64Typ, Tb.real64Typ:
				rda.rd.Read64(x64);
				ReadBytes(x64, item.adr, 8, rda.changed);
			| Tb.recTyp:
				IF begin THEN rc := Ti.ITERATE_INTO END
			| Tb.arrTyp:
				IF begin THEN rc := Ti.ITERATE_FOREACH END
			| Tb.ptrTyp:
				rda.rd.Read32(x32); size := SYSTEM.VAL(Tb.Int32, x32);
				rda.rd.Read32(x32); offset := SYSTEM.VAL(Tb.Int32, x32);
				rda.rd.Read32(x32); msg_type := SYSTEM.VAL(Tb.Int32, x32);
				rda.rd.Read16(x16); iface_code := SYSTEM.VAL(Tb.Int16, x16);
				rda.rd.Read16(x16); len := SYSTEM.VAL(Tb.Int16, x16);
				ptr := GetPtrData(rda.mapper, item);
				IF ptr = NIL THEN
					rc := Ti.STOP_ITERATE
				ELSIF (ptr.msg_type # msg_type) OR (ptr.iface_code # iface_code) THEN
					rc := Ti.STOP_ITERATE
				ELSE
					l := SYSTEM.VAL(Tb.Int16, Ti.PtrLen(item));
					CASE rda.mapper.realloc_mode OF
					| Ts.DONT_REALLOC:
						IF len > l THEN
							rc := Ti.STOP_ITERATE
						END;
					| Ts.REALLOC_LARGEST:
						IF len > l THEN
							Ti.MakeArr(item, len)
						END;
					| Ts.REALLOC_EXACT:
						IF len # l THEN
							Ti.MakeArr(item, len)
						END
					END;
					IF (rc # Ti.STOP_ITERATE) & (len > 0) THEN
						item.Deref(base);
						rdp := rda;
						rdp.dyn_pos := rdp.basic_length;
						rdp.rd.SetPos(rdp.dyn_pos+offset); 
						rda.mapper.ti.Iterate(ReadItem, "", base, rdp, rc);
						IF rc = Ti.STOP_ITERATE THEN
							rc := Ti.STOP_ITERATE
						ELSE
							rc := Ti.ITERATE_NEXT
						END;
						rda.changed := rdp.changed
					END
				END
			ELSE
				ASSERT(item.typ = Tb.boolTyp, Tb.WRONG_TYPE)
			END
		END;
		RETURN rc
	END ReadItem;

	PROCEDURE (mapper: Mapper) ReadObject (VAR item: Ti.Item; VAR rd: Ts.Reader;
	VAR changed: BOOLEAN);
		VAR res: INTEGER; rda: ReadData;
	BEGIN
		ASSERT(rd.pos = HEADER_SIZE, Tb.ALREADY_INITIALIZED);
		ASSERT(rd.hdr.basic_length - HEADER_SIZE = mapper.buffer_size,
			Tb.NOT_FOUND);
		rda.mapper := mapper; rda.rd := rd; rda.changed := FALSE;
		rda.dyn_pos := HEADER_SIZE; 
		rda.basic_length := mapper.buffer_size + HEADER_SIZE;
		mapper.ti.Iterate(ReadItem, "", item, rda, res);
		changed := rda.changed;
		mapper.cancelled := (res = Ti.STOP_ITERATE);
		mapper.eof := rd.eof
	END ReadObject;
	
	PROCEDURE (mapper: Mapper) WriteHeader (VAR hdr: Ts.Header;  wr-: Ts.Writer);
		VAR res: Tb.Int32;
	BEGIN
		hdr.basic_length := HEADER_SIZE+mapper.buffer_size;
		hdr.buffer_length := hdr.basic_length+mapper.dyn_buffer_size; 
		hdr.iface_code := mapper.iface_code; hdr.msg_type := mapper.msg_type;
		hdr.dst_oid := Tb.Int_Int16(mapper.oid);
		wr.WriteHeader(hdr, res);
		IF Tb.TRAL_MESSAGE IN Tb.trace THEN
			Log.String("SrpsWriteHeader"); Log.Int(hdr.mid); Log.Char('#'); Log.Int(hdr.dst_oid);
			Log.String(" l="); Log.Int(hdr.basic_length); Log.Int(hdr.buffer_length); Log.Ln
		END;
		mapper.eof := (res # 0)
	END WriteHeader;
	
	PROCEDURE ItemPackedSize ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR psd: Tb.AnyRec): INTEGER;
		VAR rc: INTEGER; l: Tb.Int32;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		WITH psd: PackedSizeData DO
			CASE item.typ OF
			| Tb.arrTyp:
				IF begin THEN 
					psd.presizes[psd.ind] := psd.size; INC(psd.ind);
					rc := Ti.ITERATE_INTO
				ELSE
					l := item.Len();
					DEC(psd.ind);
					psd.size := psd.presizes[psd.ind] + l * (psd.size - psd.presizes[psd.ind]);
				END;
			| Tb.ptrTyp:
				INC(psd.size, PTR_SIZE);
			| Tb.recTyp:
				IF begin THEN rc := Ti.ITERATE_INTO END;
			ELSE
				INC(psd.size, item.Size());
			END
		END;
		RETURN rc
	END ItemPackedSize;
	
	PROCEDURE WriteItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR wd: Tb.AnyRec): INTEGER;
		VAR rc, res: INTEGER; x8: Tb.Bits8; x16: Tb.Bits16; x32: Tb.Bits32; 
		x64: Tb.Bits64; len, iface_code: Tb.Int16; size, offset, msg_type: Tb.Int32; 
		ptr: PtrData; base: Ti.Item; pos, pos1: Tb.Int32; wdp: WriteData;
		psd: PackedSizeData;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		WITH wd: WriteData DO
			CASE item.typ OF
			| Tb.boolTyp, Tb.int8Typ, Tb.char8Typ:
				SYSTEM.MOVE(item.adr, SYSTEM.ADR(x8), 1);
				wd.wr.Write8(x8)
			| Tb.int16Typ, Tb.char16Typ:
				SYSTEM.MOVE(item.adr, SYSTEM.ADR(x16), 2);
				wd.wr.Write16(x16)
			| Tb.int32Typ, Tb.real32Typ:
				SYSTEM.MOVE(item.adr, SYSTEM.ADR(x32), 4);
				wd.wr.Write32(x32)
			| Tb.int64Typ, Tb.real64Typ:
				SYSTEM.MOVE(item.adr, SYSTEM.ADR(x64), 8);
				wd.wr.Write64(x64)
			| Tb.recTyp:
				IF begin THEN rc := Ti.ITERATE_INTO END
			| Tb.arrTyp:
				IF begin THEN rc := Ti.ITERATE_FOREACH END
			| Tb.ptrTyp:
				ptr := GetPtrData(wd.mapper, item);
				IF ptr # NIL THEN
					msg_type := ptr.msg_type; iface_code := ptr.iface_code;
					len := SYSTEM.VAL(Tb.Int16, Ti.PtrLen(item))
				ELSE
					msg_type := 0; iface_code := 0; len := 0
				END;
				IF len <= 0 THEN
					size := 0; offset := 0
				ELSE
					item.Deref(base);
					psd.size := 0; psd.ind := 0;
					wd.mapper.ti.Iterate(ItemPackedSize, "", base, psd, res);
					size := psd.size; 
					offset := wd.dyn_buffer_size; INC(wd.dyn_buffer_size, size);
					(* write data allocated in ptr *)
					wdp := wd;
					wdp.wr.SetPos(wdp.dyn_pos+offset); 
					wd.mapper.ti.Iterate(WriteItem, name, base, wdp, res);
					wd.dyn_buffer_size := wdp.dyn_buffer_size
				END;
				wd.wr.Write32(SYSTEM.VAL(Tb.Bits32, size));	(* size:4 *)
				wd.wr.Write32(SYSTEM.VAL(Tb.Bits32, offset));	(* offset:4 *)
				wd.wr.Write32(SYSTEM.VAL(Tb.Bits32, msg_type));	(* msg_type:4 *)
				wd.wr.Write16(SYSTEM.VAL(Tb.Bits16, iface_code));	(* iface_code:2 *)
				wd.wr.Write16(SYSTEM.VAL(Tb.Bits16, len))	(*dim:2 *)
			ELSE
				ASSERT(item.typ = Tb.boolTyp, Tb.WRONG_TYPE)
			END
		END;
		RETURN rc
	END WriteItem;
	
	PROCEDURE (mapper: Mapper) WriteObject ( item-: Ti.Item; 
	VAR wr: Ts.Writer);
		VAR res: INTEGER; wd: WriteData; pos: Tb.Int32;
	BEGIN
		IF wr.length = 0 THEN
			(* keep a place for header *)
			wr.SetPos(HEADER_SIZE);
		END;
		pos := wr.pos;
		ASSERT(pos = HEADER_SIZE, Tb.ALREADY_INITIALIZED);
		(* keep a place for basic buffer *)
		wr.SetPos(wr.pos+mapper.buffer_size);
		IF wr.eof THEN
			mapper.cancelled := TRUE
		ELSE
			wd.dyn_pos := wr.pos;
			wr.SetPos(pos);
			wd.mapper := mapper;
			wd.wr := wr;
			wd.dyn_buffer_size := 0;
			mapper.ti.Iterate(WriteItem, "", item, wd, res);
			mapper.dyn_buffer_size := wd.dyn_buffer_size;
			IF wd.dyn_buffer_size > 0 THEN
				wr.SetPos(HEADER_SIZE
					+ mapper.buffer_size + mapper.dyn_buffer_size);
			END;
			mapper.cancelled := (res = Ti.STOP_ITERATE)
		END;
		IF wr.eof THEN
			IF (Tb.TRAL_MESSAGE IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* EOF SrpsWriteObject at pos"); Log.Int(wr.pos); Log.Ln
			END
		END		
	END WriteObject;

	(* ========== Initialize ========== *)
	PROCEDURE (mf: MapperFactory) NewMapper* (ti: Ti.TInfo;
	 iface_name-: ARRAY OF CHAR; VAR tqos: Ts.TableQosBase;
	sid, oid: INTEGER; (*OUT*)VAR mapr: Ts.Mapper; (*OUT*)VAR res: INTEGER);
		VAR mapper: Mapper;
	BEGIN
		NEW(mapper);
		mapper.ti := ti; Tb.Copy(iface_name, mapper.iface_name);
		mapper.sid := sid; mapper.oid := oid; 
		mapper.realloc_mode := tqos.realloc_mode;
		mapr := mapper; res := 0
	END NewMapper;
	

BEGIN
	sheader.protname[0] := SYSTEM.VAL(Tb.Int8, 'S'); 
	sheader.protname[1] := SYSTEM.VAL(Tb.Int8, 'R'); 
	sheader.protname[2] := SYSTEM.VAL(Tb.Int8, 'P'); 
	sheader.protname[3] := SYSTEM.VAL(Tb.Int8, 'S');
	sheader.major_version := 1;	(* sh.major_version := 2 *)
	sheader.spec_bits := Ts.USER_TELEGRAM_TYPE;
END Ta_SRPSStations.