MODULE Ta_STAreas;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Log, Tb := Ta_Basics, Tc := Ta_Channels, Ts := Ta_Stations, Ta := Ta_Areas, 
		Ta_FactManager, Tt := Ta_Tables, SYSTEM,
	(* ---------- Single Threaded implementation for XDS ---------- *)
		oberonRTS, ProgramArgs, TextIO;
	

	
	TYPE
		Area = POINTER TO AreaDesc;
		Directory = POINTER TO DirectoryDesc;

	
		UpdData = POINTER TO UpdDataDesc;
		UpdDataDesc = RECORD
			on_update: Ta.AreaUpdater;
			par: Tb.AnyPtr;
			period, counter: LONGINT;
			t_period, t_setpoint: Tb.Time;
			next: UpdData
		END;
		
		AreaDesc = RECORD (Ta.AreaDesc)
			d: Directory;
			finish_flag: INTEGER;
			aid : INTEGER;		(* > 0 unique area identifier *)
			name : Tb.Name64;
			aqos : Ta.AreaQos;
			n_stations, max_stations, host_sid, send_sid: INTEGER;
			stations: Ts.Stations;
			tbl: Tt.Tablist;
			cr: Ts.Carrier;
			cfactory: Tc.ChannelFactory;
			mfactory: Ts.MapperFactory;
			(* ---------- Runtime Data ---------- *)
			t_cycle, t_setpoint, t_crt, t_crtu, tp_cdt, t_expired, 
			t_collect, tp_collect: Tb.Time;
	
			upd, tail: UpdData	
		END;
			
		DirectoryDesc = RECORD (Ta.DirectoryDesc)
			loaded_as: INTEGER;
			dqos: Ta.DirectoryQos;
			
			startarea: Area
		END;
		
		
	VAR
		dir: Directory;


	(* ========== Area Implementation ========== *)
	PROCEDURE (area: Area) StartNoUpdate ();
		VAR ok: BOOLEAN;
	BEGIN
		ASSERT(area.tbl # NIL, Tb.NIL_POINTER);
		area.tbl.Start(ok);
		area.finish_flag := 0; 
		Tb.GetTime(area.t_crtu); area.t_setpoint := area.t_crtu; 
		IF area.d.dqos.collect_cycle < 0 THEN
			area.tp_collect := Tb.INFINITE_TIME
		ELSE
			Tb.Long_Time(area.d.dqos.collect_cycle, area.tp_collect);
			Tb.AddTime(area.t_crtu, area.tp_collect, area.t_collect)
		END;
	END StartNoUpdate;

	PROCEDURE DoBeforeRun (area: Area);
		VAR cst: Tb.Time; s: Tb.Name; ok: BOOLEAN;
	BEGIN
		area.StartNoUpdate;
		Tb.Long_Time(area.d.dqos.cycle, area.t_cycle);
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Tb.Time_String(area.t_crtu, {Tb.tDD, Tb.tMM, Tb.tYY, 
				Tb.thh, Tb.tmm, Tb.tss, Tb.tms, Tb.tloc, Tb.tdwh, Tb.tdwl}, s);
			Log.String(s); Log.String(" Run cycle=");
			Tb.Time_String(area.t_cycle, {Tb.thh,Tb.tmm,Tb.tss,Tb.tms}, s);
			Log.String(s); Log.String(" loaded_as="); Log.Int(area.d.loaded_as); 
			Log.String(" tables_started="); Log.Bool(ok); Log.Ln
		END;
		IF area.d.dqos.expired # 0 THEN
			Tb.Long_Time(area.d.dqos.expired, cst);
			Tb.AddTime(area.t_setpoint, cst, area.t_expired);
			IF Tb.TRAL_RUNTIME IN Tb.trace THEN
				Tb.Time_String(area.t_expired, {Tb.tDD, Tb.tMM, Tb.tYY, 
					Tb.thh, Tb.tmm, Tb.tss, Tb.tms, Tb.tloc, Tb.tdwh, Tb.tdwl}, s);
				Log.String("expired at "); Log.String(s); 
				Log.Ln
			END
		END;
	END DoBeforeRun;
	
	PROCEDURE (area: Area) Start ();
	BEGIN
		DoBeforeRun(area);
		area.d.startarea := area
	END Start;
	
	PROCEDURE CloseStationsData (area: Area);
		VAR j: INTEGER;
	BEGIN
		FOR j := 0 TO area.n_stations-1 DO
			IF area.stations[j] # NIL THEN
				area.stations[j].Close
			END
		END
	END CloseStationsData;
	
	PROCEDURE (area: Area) Finish (finish_flag: INTEGER);
		VAR cst: Tb.Time; ldelay: LONGINT;
	BEGIN
		ASSERT(finish_flag # 0, Tb.ZERO_VAL);
		IF finish_flag > 0 THEN
			CASE area.d.loaded_as OF
			| Ta.AS_GUI_APP:
			ELSE
				area.finish_flag := finish_flag;
			END;
			IF Tb.TRAL_RUNTIME IN Tb.trace THEN
				Log.String("Finish"); Log.Int(area.finish_flag); Log.Ln
			END;
			IF area.d.startarea = area THEN
				area.d.startarea := NIL
			END;
			CloseStationsData(area)
		ELSE
			area.d.dqos.expired := -finish_flag;
			ldelay := -1000000;
			ldelay := ldelay * finish_flag;
			Tb.Long_Time(ldelay, cst);
			Tb.AddTime(area.t_setpoint, cst, area.t_expired)
		END
	END Finish;
	

	PROCEDURE RecvMessages (area: Area);
		VAR j, sid: INTEGER; done: BOOLEAN; rd: Ts.Reader;
	BEGIN
		area.cr.Poll;
		area.tbl.is_updating := TRUE;
		FOR j := 0 TO area.n_stations-1 DO
			area.stations[j].CheckPoll()
		END;
		FOR j := 0 TO area.n_stations-1 DO
			IF Tc.TRANS_RECV IN area.stations[j].chs.trans_flags THEN
				area.stations[j].Recv;
				WHILE area.stations[j].rd.res > 0 DO
					area.stations[j].Recv;
				END
			END
		END;
		j := 0; 
		REPEAT
			area.cr.GetRecv(j, rd, sid);
			IF ~rd.eof THEN
				area.tbl.Process(rd, sid, done);
				INC(j)
			END
		UNTIL rd.eof;
		area.tbl.ProcessExpired(done);
	END RecvMessages;
	
	PROCEDURE UpdateData (area: Area);
		VAR upd: UpdData;
	BEGIN
		upd := area.upd;
		WHILE upd # NIL DO
			IF upd.period >= 0 THEN
				INC(upd.counter);
				IF upd.counter >= upd.period THEN
					upd.on_update(area, upd.par);
					upd.counter := 0
				END;
			ELSE
				IF Tb.CompareTime(upd.t_setpoint, area.t_crtu) <= 0 THEN
					upd.on_update(area, upd.par);
					Tb.AddTime(area.t_crtu, upd.t_period, upd.t_setpoint)
				END
			END;
			upd := upd.next
		END;
	END UpdateData;
	
	PROCEDURE SendMessages (area: Area);
		VAR j: INTEGER; t_crt: Tb.Time;
	BEGIN
		FOR j := 0 TO area.n_stations-1 DO
			IF Tc.TRANS_SEND IN area.stations[j].chs.trans_flags THEN
				IF area.stations[j].cqos.enable_ping THEN
					IF Tb.CompareTime(area.t_crtu, area.stations[j].wr.t_send) 
					>= area.stations[j].cqos.tp_ping_timeout THEN
						area.stations[j].SendPing(area.host_sid)
					END
				END
			END;
			area.tbl.is_updating := FALSE;
		END;
	END SendMessages;

	PROCEDURE (area: Area) Update ();
		VAR s: Tb.Name;
	BEGIN
		Tb.GetTime(area.t_crtu);
		RecvMessages(area);
		UpdateData(area);
		SendMessages(area);
		Tb.GetTime(area.t_crt);
		Tb.SubtractTime(area.t_crt, area.t_crtu, area.tp_cdt);
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Tb.Time_String(area.t_crtu, 
				{Tb.tDD, Tb.tMM, Tb.tYY, Tb.thh, Tb.tmm, Tb.tss, 
				Tb.tms, Tb.tloc}, s); Log.String(s); 
			Log.String(" updated proctime=");
			Tb.Time_String(area.tp_cdt, {Tb.thh, Tb.tmm, Tb.tss, Tb.tms}, s);
			Log.String(s); 
			Log.Ln
		END;
		IF area.d.dqos.collect_cycle >= 0 THEN
			IF Tb.CompareTime(area.t_crtu, area.t_collect) >= 0 THEN
				oberonRTS.Collect;
				IF Tb.TRAL_RUNTIME IN Tb.trace THEN
					Log.String("Garbage Collector called"); Log.Ln;
				END;
				Tb.AddTime(area.t_crtu, area.tp_collect, area.t_collect)
			END
		END
	END Update;
	
	PROCEDURE AddCarrier(area: Area;  cqos-: Tc.ChannelQosBase; 
	VAR res: INTEGER): Ts.Carrier;
		VAR pfactory: Tc.PollingFactory; crfactory: Ts.CarrierFactory; cr: Ts.Carrier;
	BEGIN
		res := 0;
		Ta_FactManager.CreatePollingFactory(area.aqos.polling, pfactory, res);
		IF res = 0 THEN
			Ta_FactManager.CreateCarrierFactory(cqos.prot_format, crfactory, res)
		END;
		IF res = 0 THEN
			crfactory.NewCarrier(pfactory, area.aid, area.max_stations, 
				area.aqos.send_buffer_size, area.aqos.recv_buffer_size, 
				cqos.prot_format, cr, res);
		END;
		IF res # 0 THEN cr := NIL END;
		RETURN cr
	END AddCarrier;
	
	PROCEDURE AddStation(area: Area; sid, host_sid: INTEGER; 
	 name-: ARRAY OF CHAR; VAR cqos: Tc.ChannelQosBase; VAR res: INTEGER);
	BEGIN
		res := 0;
		IF (area.host_sid < 0) & (sid = host_sid) THEN
			area.host_sid := sid
		ELSIF area.send_sid < 0 THEN
			area.send_sid := sid
		END;
		IF area.cr = NIL THEN
			area.cr := AddCarrier(area, cqos, res);
		ELSE
			ASSERT(area.cr.prot_format = cqos.prot_format, Tb.WRONG_TYPE)
		END;
		IF (area.cfactory = NIL) & (res = 0) THEN
			Ta_FactManager.CreateChannelFactory(cqos.protocol, area.cfactory, res)
		END;
		IF (area.mfactory = NIL) & (res = 0) THEN
			Ta_FactManager.CreateMapperFactory(cqos.prot_format, area.mfactory, res)
		END;
		IF res = 0 THEN
			area.stations[sid] := area.cr.NewStation(area.stations[sid], area.cfactory,
				area.aid, sid, name, cqos, (area.host_sid = sid), (area.send_sid = sid));
			IF area.stations[sid] = NIL THEN
				res := Tb.INIT_FAILED
			END;
		END;
	END AddStation;
	
	PROCEDURE (area: Area) AddStations ( name-: ARRAY OF CHAR; 
	n_appended, host_sid: INTEGER; VAR cqos: Tc.ChannelQosBase;
	(*OUT*)VAR res: INTEGER);
		VAR i, j, len: INTEGER; name1: Tb.Name64;
	BEGIN
		res := 0;
		IF n_appended = Ta.APPEND_AS_LIST THEN
			i := 0; len := Tb.Len(name);
			WHILE (i < len) & (res = 0) DO
				j := 0; name1[0] := 0X;
				WHILE (name[i+j] # ' ') & (name[i+j] # 09X) & (name[i+j] # 0X) DO
					IF (name[i+j] = '*') & (j = 0) THEN
						IF host_sid < 0 THEN
							host_sid := area.n_stations
						END;
						INC(i)
					ELSE
						name1[j] := name[i+j];
						INC(j)
					END;
				END;
				name1[j] := 0X;
				AddStation(area, area.n_stations, host_sid, name1, cqos, res);
				INC(i,j+1);
				INC(area.n_stations);
				IF (i < len) & (area.n_stations >= area.max_stations) THEN
					res := Tb.TRUNCATED
				END
			END
		ELSIF (n_appended >= 0) & 
		(area.n_stations+n_appended <= area.max_stations) THEN
			i := 0;
			WHILE (i < n_appended) & (res = 0) DO
				AddStation(area, area.n_stations, host_sid, name, cqos, res);
				INC(area.n_stations); INC(i)
			END
		ELSE
			res := Tb.TRUNCATED
		END
	END AddStations;	

	PROCEDURE (area: Area) NStations (): INTEGER; 
	BEGIN
		RETURN area.n_stations
	END NStations;

	PROCEDURE (area: Area) HostSid (): INTEGER; 
	BEGIN
		RETURN area.host_sid
	END HostSid;

	PROCEDURE (area: Area) Station (sid: INTEGER): Ts.Station; 
	BEGIN
		ASSERT((sid >= 0) & (sid < area.n_stations), Tb.CHECK_RANGE_ERROR);
		RETURN area.stations[sid]
	END Station;

	PROCEDURE (area: Area) AddUpdater (on_update: Ta.AreaUpdater; par: Tb.AnyPtr;
	period: LONGINT);
		VAR upd: UpdData;
	BEGIN
		NEW(upd); upd.on_update := on_update; upd.par := par; upd.period := period;
		IF period < 0 THEN
			Tb.Long_Time(-period, upd.t_period)
		END;
		IF area.upd = NIL THEN
			area.upd := upd; area.tail := upd;
		ELSE
			area.tail.next := upd; area.tail := upd
		END
	END AddUpdater;

	PROCEDURE (area: Area) NewTable (VAR rec: Tb.AnyRec; 
	 iface_name-: ARRAY OF CHAR;  tname-: ARRAY OF CHAR;
	sid, oid: INTEGER; VAR tqos: Ts.TableQosBase; 
	on_data: Ta.OnDataHandler): Ta.Table;
		VAR res: INTEGER;
	BEGIN
		IF area.tbl = NIL THEN
			Tt.NewTablist(area.mfactory, area.stations, 
				area.aid, area.host_sid, area.send_sid, area.tbl, res);
			ASSERT(res = 0, Tb.INIT_FAILED)
		END;
		RETURN area.tbl.NewTable(rec, iface_name, tname, 
			area.n_stations, sid, oid, tqos, on_data);
	END NewTable;

	PROCEDURE (area: Area) Message (sid, src_oid: INTEGER; VAR msg: Tb.Msg);
		VAR j: INTEGER;
	BEGIN
		IF sid = Ts.sendSID THEN sid := area.send_sid END;
		IF sid = Ts.hostSID THEN sid := area.host_sid END;
		IF (src_oid >= 0) OR (src_oid = Ts.toALL) THEN
			area.tbl.Message(src_oid, msg)
		ELSIF sid >= 0 THEN
			area.stations[sid].HandleMsg(msg)
		ELSIF sid = Ts.toALL THEN
			FOR j := 0 TO area.n_stations-1 DO
				area.stations[j].HandleMsg(msg)
			END
		ELSE
			WITH msg: Ta.GetDirectoryQos DO
				msg.dqos := area.d.dqos
			| msg: Ta.GetAreaRunTimes DO
				msg.t_setpoint := area.t_setpoint; msg.t_crt := area.t_crt;
				msg.t_crtu := area.t_crtu; msg.tp_cdt := area.tp_cdt;
			ELSE
			END
		END
	END Message; 

	(* ---------- Single Startarea Running ---------- *)
	PROCEDURE Run (area: Area);
		VAR j: INTEGER; b: BOOLEAN; t_setpoint, cst: Tb.Time; 
	BEGIN
		WHILE area.finish_flag = 0 DO
			j := 0; b := FALSE;
			IF area.d.dqos.with_compensation THEN
				t_setpoint := area.t_setpoint;
				WHILE ~b & (j < 5) DO
					Tb.AddTime(t_setpoint, area.t_cycle, t_setpoint);
					IF Tb.CompareTime(t_setpoint, area.t_crtu) >= 0 THEN
						b := TRUE
					END;
					INC(j)
				END
			ELSE
				Tb.AddTime(area.t_crtu, area.t_cycle, t_setpoint); b := TRUE
			END;
			IF b THEN
				Tb.SubtractTime(t_setpoint, area.t_crtu, cst);
				Tb.Sleep(cst);
				area.t_setpoint := t_setpoint;
			ELSE
				IF Tb.TRAL_RUNTIME IN Tb.trace THEN
					Log.String("? proctime skip="); Log.Ln;
				END;
				Tb.Sleep(area.t_cycle);
				area.t_setpoint := area.t_crtu
			END;
			area.Update;
			IF (area.d.dqos.expired # 0) & (Tb.CompareTime(area.t_crtu, 
			area.t_expired) > 0) THEN
				area.Finish(Ta.FINISH_EXPIRED)
			END
		END;
	END Run; 
	


	(* ========== Directory Implementation ========== *)
	PROCEDURE (d: Directory) NewArea (aid: INTEGER; name: ARRAY OF CHAR;
	max_stations: INTEGER;  aqos-: Ta.AreaQos) 
	: Ta.Area;
		VAR area: Area;
	BEGIN
		ASSERT(max_stations > 0, Tb.ZERO_VAL);
		NEW(area); area.d := d; area.aid := aid; Tb.Copy(name, area.name); 
		IF d.loaded_as = Ta.AS_INPROC_LIB THEN
			area.max_stations := 2
		ELSE
			area.max_stations := max_stations
		END;
		NEW(area.stations, area.max_stations);
		area.aqos := aqos; area.host_sid := -1; area.send_sid := -1;
		RETURN area
	END NewArea;
	

	PROCEDURE (d: Directory) InitHostModule ( dqos-: Ta.DirectoryQos);
	BEGIN
		ASSERT(oberonRTS.Search(d.default_mod.name) # oberonRTS.nullModule, 
			Tb.NOT_FOUND);
		d.dqos := dqos;
		d.default_mod.ti.ConnectTo(d.dqos, "", "", NIL);
		d.default_mod.ti.Insert(0, "cycle", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(1, "expired", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(2, "collect_cycle", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(3, "with_compensation", Tb.boolTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.InitConfig(d.command_line, d.default_mod.name, 
			d.current_dir, d.ind_command_line);
		d.default_mod.initproc;
		IF d.startarea # NIL THEN
			Run(d.startarea)
		END
	END InitHostModule;		
	

	PROCEDURE (d: Directory) LoadedAs () : INTEGER;
	BEGIN 
		RETURN d.loaded_as
	END LoadedAs;


	(* ========== Initialize and Install ========== *)
	PROCEDURE Initialize;
		VAR i: INTEGER; str, command_line: Tb.Name; cid: ProgramArgs.ChanId;
	BEGIN
		NEW(dir);

		i := 0;
		Tb.AppendAt(".", command_line, i);
		cid := ProgramArgs.ArgChan();
		WHILE ProgramArgs.IsArgPresent() DO
			TextIO.ReadToken(cid, str);
			Tb.AppendAt(" ", command_line, i);
			Tb.AppendAt(str, command_line, i)
		END;
		dir.loaded_as := Ta.AS_CONSOLE_APP;

		Ta.SetDir(dir, "", "", command_line, 1);

	END Initialize;
	
BEGIN
	Initialize
END Ta_STAreas.